/*
 * Author:      William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990-1999, William Chia-Wei Cheng.
 *
 * Permission limited to the use, copy, display, distribute without
 * charging for a fee, and produce derivative works of "tgif" and
 * its documentation for not-for-profit purpose is hereby granted by
 * the Author, provided that the above copyright notice appears in
 * all copies made of "tgif" and that both the copyright notice
 * and this permission notice appear in supporting documentation,
 * and that the name of the Author not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission.  The Author makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.  All other rights (including, but not limited to, the
 * right to sell "tgif", the right to sell or distribute derivative
 * works of "tgif", the right to distribute "tgif" for a fee, and
 * the right to include "tgif" or derivative works of "tgif" in a
 * for-sale product) are reserved by the Author.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
 * OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * @(#)$Header: /mm/src/tgif/v4/RCS/cutpaste.c,v 4.7 1999/01/22 00:06:43 william Exp $
 */

#define _INCLUDE_FROM_CUTPASTE_C_

#include "tgifdefs.h"

#include "arc.e"
#include "auxtext.e"
#include "box.e"
#include "choice.e"
#include "choose.e"
#include "cmd.e"
#include "color.e"
#ifndef _NO_KINPUT
#include "convkinput.e"
#endif /* ~_NO_KINPUT */
#include "cutpaste.e"
#include "cursor.e"
#include "dialog.e"
#include "drawing.e"
#include "dup.e"
#include "file.e"
#include "font.e"
#include "grid.e"
#include "ini.e"
#include "mark.e"
#include "miniline.e"
#include "move.e"
#include "msg.e"
#include "names.e"
#ifndef _NO_NKF
#include "nkf.e"
#endif /* ~_NO_NKF */
#include "obj.e"
#include "oval.e"
#include "page.e"
#include "pattern.e"
#include "poly.e"
#include "polygon.e"
#include "rcbox.e"
#include "remote.e"
#include "select.e"
#include "setup.e"
#include "special.e"
#include "stretch.e"
#include "text.e"
#include "util.e"
#include "xbitmap.e"
#include "xpixmap.e"

int	copyingToCutBuffer=FALSE;
int	pastingFile=FALSE;

static char *cutBuffer=NULL;

static
int WriteBufToIni(buf, ini_fname)
   char *buf, *ini_fname;
{
   char *pszSec=buf, *pszKey=NULL, *pszKeyStart=NULL;
   int len=strlen(pszSec);

   pszKeyStart = (&buf[len+1]);
   if (*pszSec == '[' && pszSec[len-1] == ']') {
      pszSec[len-1] = '\0';
      pszSec++;
   }
   tgWriteProfileString(pszSec, NULL, NULL, ini_fname);

   for (pszKey=pszKeyStart; *pszKey != '\0'; pszKey++) {
      char *psz=strchr(pszKey, '=');

      if (psz == NULL) {
         tgWriteProfileString(pszSec, NULL, NULL, ini_fname);
         tgWriteProfileString(NULL, NULL, NULL, ini_fname);
         return FALSE;
      }
      *psz = '\0';
      tgWriteProfileString(pszSec, pszKey, &psz[1], ini_fname);
      *psz = '=';

      pszKey += strlen(pszKey);
   }
   tgWriteProfileString(NULL, NULL, NULL, ini_fname);

   return TRUE;
}

static
int WriteBufToCutBuffer(buf, bytes_to_write)
   char *buf;
   int bytes_to_write;
{
   int copy_failed=FALSE;

   copyingToCutBuffer = TRUE;
   XStoreBytes(mainDisplay, buf, bytes_to_write);
   XSync(mainDisplay, False);
   if (copyingToCutBuffer == INVALID) {
      copy_failed = TRUE;
   }
   copyingToCutBuffer = FALSE;
   if (copy_failed) {
      char ch='\0';

      XStoreBytes(mainDisplay, &ch, 1);
   }
   return (!copy_failed);
}

static
int CopyObjectToCutBuffer()
   /* returns FALSE if copying in text mode -- this is */
   /*    interpreted as an attempt to copy highlighted text */
{
   FILE *fp=NULL;
   char tmpfile[MAXSTRING];
   struct stat stat;
   unsigned char header=TGIF_HEADER;
   struct ObjRec *partial_text_obj_ptr=NULL;
    
   if (curChoice == DRAWTEXT) {
      if (!textHighlight) {
         return FALSE;
      } else {
         /* create a partial text object with only the minilines */
         partial_text_obj_ptr = CreateTextObjFromHighLightedText();
         if (partial_text_obj_ptr == NULL) {
            return FALSE;
         }
      }
   } else if (topSel == NULL) {
      MsgBox("No object selected for the COPY operation.", TOOL_NAME, INFO_MB);
      return TRUE;
   }
   sprintf(tmpfile, "%sTgifXXXXXX", TMP_DIR);
   mktemp(tmpfile);
   if ((fp=fopen(tmpfile, "w+")) == NULL) {
      sprintf(gszMsgBox, "Cannot open %s.", tmpfile);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return TRUE;
   }
   writeFileFailed = FALSE;
   if (write(fileno(fp), (char*)(&header), 1) < 1) writeFileFailed = TRUE;

   if (curChoice == DRAWTEXT) {
      /* create a partial text object with only the minilines */
      copyInDrawTextMode = TRUE;
      Save(fp, partial_text_obj_ptr, 0, 1);
      copyInDrawTextMode = FALSE;

      FreeObj(partial_text_obj_ptr);
   } else {
      struct SelRec *sel_ptr=NULL;
      struct ObjRec *top_obj=NULL, *bot_obj=NULL;

      for (sel_ptr=botSel; sel_ptr != NULL; sel_ptr=sel_ptr->prev) {
         struct ObjRec *obj_ptr=DupObj(sel_ptr->obj);

         obj_ptr->prev = NULL;
         obj_ptr->next = top_obj;

         if (top_obj == NULL) {
            bot_obj = obj_ptr;
         } else {
            top_obj->prev = obj_ptr;
         }
         top_obj = obj_ptr;
      }
      Save(fp, bot_obj, 0, 1);
      while (top_obj != NULL) {
         struct ObjRec *obj_ptr=top_obj->next;

         FreeObj(top_obj);
         top_obj = obj_ptr;
      }
   }
   if (writeFileFailed) {
      sprintf(gszMsgBox, "Fail to write to '%s'.  File system may be full.",
            tmpfile);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      fclose(fp);
      unlink(tmpfile);
      writeFileFailed = FALSE;
      return TRUE;
   }
   fflush(fp);
   if (fstat(fileno(fp), &stat) < 0) {
      fclose(fp);
      unlink(tmpfile);
      sprintf(gszMsgBox, "FSTAT error in %s.  Copy aborted!", tmpfile);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return TRUE;
   }

   if (cutBuffer != NULL) free(cutBuffer);
   cutBuffer = (char*)malloc((stat.st_size+1)*sizeof(char));
   if (cutBuffer == NULL) FailAllocMessage();

   rewind(fp);
   if (read(fileno(fp), cutBuffer, stat.st_size) < stat.st_size) {
      sprintf(gszMsgBox, "READ error in %s.  Copy aborted!", tmpfile);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
   } else {
      if (!WriteBufToCutBuffer(cutBuffer, stat.st_size)) {
         sprintf(gszMsgBox, "%s  %s", "Copy to cut buffer failed.",
               "Selected object(s) may be too big.");
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      } else {
         sprintf(gszMsgBox, "Copy buffer updated.");
         Msg(gszMsgBox);
      }
   }
   fclose(fp);
   unlink(tmpfile);

   return (curChoice != DRAWTEXT);
}

int CopyToCutBuffer()
   /* returns FALSE if copying in text mode -- this is */
   /*    interpreted as an attempt to copy highlighted text */
{
   if (curChoice == DRAWTEXT) {
      if (!textHighlight) {
         return FALSE;
      } else {
         if (escPressed) {
            escPressed = FALSE;
            Msg("An <ESC> key press is ignored.");
         }
         if (CanCopyHighLightedTextAsStrings()) {
            XEvent ev;

            copyInDrawTextMode = TRUE;
            ev.type = KeyPress;
            DrawText(&ev);
            return FALSE;
         } else {
            /*
             * create a partial text object with only the minilines, this
             *       will be done in CopyObjectToCutBuffer() below
             */
         }
      }
   } else if (topSel == NULL) {
      MsgBox("No object selected for the COPY operation.", TOOL_NAME, INFO_MB);
      return TRUE;
   }
   return CopyObjectToCutBuffer();
}

int CopyPlainTextAsObject()
   /* returns FALSE if copying in text mode -- this is */
   /*    interpreted as an attempt to copy highlighted text */
{
   if (!(curChoice == DRAWTEXT && textHighlight &&
         CanCopyHighLightedTextAsStrings())) {
      return (curChoice != DRAWTEXT);
   }
   return CopyObjectToCutBuffer();
}

void CutToCutBuffer()
{
   if (curChoice == DRAWTEXT && textCursorShown) {
      CopyToCutBuffer();
      DeleteHighlightedText();
   }
   if (curChoice == NOTHING && topSel != NULL) {
      CopyToCutBuffer();
      DelAllSelObj();
   }
}

void PasteString(CutBuffer)
   char *CutBuffer;
{
   char *c_ptr=NULL, *dest_c_ptr=NULL;
   int x, y, num_lines=0, char_count, root_x, root_y, grid_x, grid_y;
   unsigned int status;
   struct ObjRec *obj_ptr=NULL;
   struct TextRec *text_ptr=NULL;
   Window root_win=None, child_win=None;
   MiniLineInfo *pFirstMiniLine=NULL, *pLastMiniLine=NULL;

   if (*CutBuffer == '\0') {
      MsgBox("Cut buffer is empty", TOOL_NAME, INFO_MB);
      return;
   }
   TieLooseEnds();
   SetCurChoice(NOTHING);
   if (topSel!=NULL) { HighLightReverse(); RemoveAllSel(); }

   XQueryPointer(mainDisplay, drawWindow, &root_win, &child_win,
         &root_x, &root_y, &x, &y, &status);
   GridXY(x, y, &grid_x, &grid_y);

   obj_ptr = (struct ObjRec *)malloc(sizeof(struct ObjRec));
   if (obj_ptr == NULL) FailAllocMessage();
   memset(obj_ptr, 0, sizeof(struct ObjRec));

   text_ptr = (struct TextRec *)malloc(sizeof(struct TextRec));
   if (text_ptr == NULL) FailAllocMessage();
   memset(text_ptr, 0, sizeof(struct TextRec));

   for (c_ptr=CutBuffer, num_lines = 0; *c_ptr != '\0'; num_lines++) {
      MiniLineInfo *pMiniLine=NULL;
      char s[MAXSTRING+1];

      char_count = 0;
      dest_c_ptr = s;
      while (*c_ptr != '\0' && *c_ptr != '\n' && *c_ptr != '\r') {
         *dest_c_ptr++ = *c_ptr++;
         if (++char_count == MAXSTRING) {
            sprintf(gszMsgBox, "String length exceeds %1d.  %s.",
                  MAXSTRING, "String truncated");
            Msg(gszMsgBox);
            while (*c_ptr != '\0' && *c_ptr != '\n' && *c_ptr != '\r') c_ptr++;
            break;
         }
      }
      *dest_c_ptr = '\0';
      pMiniLine = CreateMiniLineFromString(s, &pFirstMiniLine, &pLastMiniLine);

      while (*c_ptr == '\n' || *c_ptr == '\r') c_ptr++;
   }
   text_ptr->lines = num_lines;
   text_ptr->minilines.first = pFirstMiniLine;
   text_ptr->minilines.last = pLastMiniLine;
   text_ptr->baseline_y = grid_y+pFirstMiniLine->asc;

   CopyCurInfoIntoTextPtr(obj_ptr, text_ptr);

   obj_ptr->x = grid_x;
   obj_ptr->y = grid_y;
   obj_ptr->type = OBJ_TEXT;
   obj_ptr->color = colorIndex;
   obj_ptr->id = objId++;;
   obj_ptr->dirty = FALSE;
   obj_ptr->rotation = 0;
   obj_ptr->detail.t = text_ptr;
   obj_ptr->fattr = obj_ptr->lattr = NULL;
   obj_ptr->ctm = NULL;

   RecalcTextMetrics(text_ptr, grid_x, text_ptr->baseline_y);
   UpdTextBBox(obj_ptr);
   AddObj(NULL, topObj, obj_ptr);
   PlaceTopObj(obj_ptr);

   SelectTopObj();
   RecordNewObjCmd();
   SetFileModified(TRUE);
   justDupped = FALSE;
}

static
struct ObjRec *CreateTmpBoxObj(LtX, LtY, RbX, RbY)
   int LtX, LtY, RbX, RbY;
{
   register struct BoxRec *box_ptr;
   register struct ObjRec *obj_ptr;

   box_ptr = (struct BoxRec *)malloc(sizeof(struct BoxRec));
   if (box_ptr == NULL) FailAllocMessage();
   memset(box_ptr, 0, sizeof(struct BoxRec));
   box_ptr->fill = NONEPAT;
   box_ptr->width = 0;
   box_ptr->pen = NONEPAT;
   box_ptr->dash = 0;

   obj_ptr = (struct ObjRec *)malloc(sizeof(struct ObjRec));
   if (obj_ptr == NULL) FailAllocMessage();
   memset(obj_ptr, 0, sizeof(struct ObjRec));

   obj_ptr->bbox.ltx = obj_ptr->obbox.ltx = obj_ptr->x = LtX;
   obj_ptr->bbox.lty = obj_ptr->obbox.lty = obj_ptr->y = LtY;
   obj_ptr->bbox.rbx = obj_ptr->obbox.rbx = RbX;
   obj_ptr->bbox.rby = obj_ptr->obbox.rby = RbY;
   obj_ptr->type = OBJ_BOX;
   obj_ptr->color = colorIndex;
   obj_ptr->id = 0;
   obj_ptr->dirty = FALSE;
   obj_ptr->rotation = 0;
   obj_ptr->detail.b = box_ptr;
   obj_ptr->fattr = obj_ptr->lattr = NULL;
   obj_ptr->ctm = NULL;
   obj_ptr->invisible = FALSE;
   obj_ptr->trans_pat = FALSE;

   return obj_ptr;
}

void AssignNewObjIds(ObjPtr)
   struct ObjRec *ObjPtr;
{
   register struct ObjRec *obj_ptr;
   register struct AttrRec *attr_ptr;

   ObjPtr->id = objId++;
   switch (ObjPtr->type) {
   case OBJ_GROUP:
   case OBJ_SYM:
   case OBJ_ICON:
   case OBJ_PIN:
      for (obj_ptr=ObjPtr->detail.r->first; obj_ptr != NULL;
            obj_ptr=obj_ptr->next) {
         AssignNewObjIds(obj_ptr);
      }
      break;
   default: break;
   }
   for (attr_ptr=ObjPtr->fattr; attr_ptr != NULL; attr_ptr=attr_ptr->next) {
      AssignNewObjIds(attr_ptr->obj);
   }
}

char *FetchCutBuffer(pn_buf_sz)
   int *pn_buf_sz;
{
   char *cut_buffer=(char*)XFetchBytes(mainDisplay, pn_buf_sz);

   return cut_buffer;
}

int PasteFromCutBuffer()
   /* returns FALSE if pasting in text mode and non-tgif bytes are */
   /*    in the cut buffer -- this is interpreted as an attempt to */
   /*    paste into the current text */
{
   FILE *fp=NULL;
   int len, ltx, lty, rbx, rby, dx, dy, read_status, paste_string=FALSE;
   char tmpfile[MAXSTRING], *cut_buffer=NULL, *orig_cut_buffer=NULL;
   unsigned char header=TGIF_HEADER;
   struct ObjRec *obj_ptr=NULL, *tmp_obj=NULL, *partial_text_obj_ptr=NULL;
   struct ObjRec *saved_top_obj=NULL, *saved_bot_obj=NULL;
   struct ObjRec *tmp_top_obj=NULL, *tmp_bot_obj=NULL;

   orig_cut_buffer = cut_buffer = (char*)FetchCutBuffer(&len);
   if (len == 0) {
      MsgBox("Cut buffer is empty", TOOL_NAME, INFO_MB);
      return TRUE;
   }
   if (((unsigned char)(*cut_buffer)) != header) {
      paste_string = TRUE;
   } else if (strncmp(&cut_buffer[1], "%TGIF", 5) == 0) {
      paste_string = FALSE;
   } else if (strncmp(&cut_buffer[1], "state(", 6) == 0) {
      /* very old tgif format (even before tgif-2.12) */
      paste_string = FALSE;
   } else {
      paste_string = TRUE;
   }
   if (paste_string) {
      if (curChoice == DRAWTEXT) {
         XEvent ev;

         XFree(cut_buffer);
         pasteInDrawTextMode = TRUE;
         ev.type = KeyPress;
         DrawText(&ev);
         return FALSE;
      }
#ifndef _NO_KINPUT
      if (copyAndPasteJIS) {
         CvtJisToEuc(cut_buffer, cut_buffer);
      }
#endif /* ~_NO_KINPUT */
      Msg("Paste from a non-tgif tool.");
      PasteString(cut_buffer);
      XFree(cut_buffer);
      return TRUE;
   }
   cut_buffer++;
   len--;

   sprintf(tmpfile, "%sTgifXXXXXX", TMP_DIR);
   mktemp(tmpfile);
   if ((fp=fopen(tmpfile, "w+")) == NULL) {
      sprintf(gszMsgBox, "Cannot open %s for write.", tmpfile);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      XFree(orig_cut_buffer);
      return TRUE;
   }
   writeFileFailed = FALSE;
   if (write(fileno(fp), cut_buffer, len) < len) {
      fclose(fp);
      unlink(tmpfile);
      sprintf(gszMsgBox, "FWRITE error in writing to '%s'.  Paste aborted!",
            tmpfile);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      XFree(orig_cut_buffer);
      return TRUE;
   }
   fflush(fp);
   rewind(fp);
   XFree(orig_cut_buffer);

   SetWatchCursor(drawWindow);
   SetWatchCursor(mainWindow);

   if (curChoice == DRAWTEXT) {
      partial_text_obj_ptr = NULL;
      SaveCursorPositionInCurText();
   } else {
      TieLooseEnds();
      SetCurChoice(NOTHING);
      if (topSel != NULL) { HighLightReverse(); RemoveAllSel(); }
   }
   saved_top_obj = topObj;
   saved_bot_obj = botObj;
   curPage->top = topObj = NULL;
   curPage->bot = botObj = NULL;
    
   importingFile = TRUE;
   pastingFile = TRUE;
   readingPageNum = 0;
   while ((read_status=ReadObj(fp, &obj_ptr)) == TRUE) {
      if (obj_ptr != NULL) {
         AdjForOldVersion(obj_ptr);
         UnlockAnObj(obj_ptr);
         AssignNewObjIds(obj_ptr);
         AddObj(NULL, topObj, obj_ptr);
      }
   }
   fclose(fp);
   if (!PRTGIF && colorLayers && needToRedrawColorWindow) {
      RedrawColorWindow();
   }
   importingFile = FALSE;
   pastingFile = FALSE;
   unlink(tmpfile);
   SetDefaultCursor(mainWindow);
   ShowCursor();

   if (read_status == INVALID) {
      sprintf(gszMsgBox, "File version too large (=%1d).  Paste aborted!",
            fileVersion);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return TRUE;
   }
   if (curChoice == DRAWTEXT && textCursorShown && topObj != NULL &&
         topObj == botObj && topObj->type == OBJ_TEXT) {
      /* we are pasting minilines and not object */
      partial_text_obj_ptr = topObj;
      RestoreCursorPositionInCurText();
      curPage->top = topObj = saved_top_obj;
      curPage->bot = botObj = saved_bot_obj;

      if (curStrBlock->type == SB_SUPSUB_CENTER) {
         sprintf(gszMsgBox, "%s.",
               "Cannot paste mixed text at the current cursor position");
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      } else {
         Msg("String blocks pasted from tgif.");
         PasteMiniLinesAtCursor(partial_text_obj_ptr);
      }
      FreeObj(partial_text_obj_ptr);

      return FALSE;
   }
   if (topObj != NULL) SetFileModified(TRUE);

   ltx = topObj->obbox.ltx;
   lty = topObj->obbox.lty;
   rbx = topObj->obbox.rbx;
   rby = topObj->obbox.rby;
   for (obj_ptr = topObj->next; obj_ptr != NULL; obj_ptr = obj_ptr->next) {
      if (obj_ptr->obbox.ltx < ltx) ltx = obj_ptr->obbox.ltx;
      if (obj_ptr->obbox.lty < lty) lty = obj_ptr->obbox.lty;
      if (obj_ptr->obbox.rbx > rbx) rbx = obj_ptr->obbox.rbx;
      if (obj_ptr->obbox.rby > rby) rby = obj_ptr->obbox.rby;
   }
   tmp_obj = CreateTmpBoxObj(ltx, lty, rbx, rby);

   tmp_top_obj = topObj;
   tmp_bot_obj = botObj;
   curPage->top = topObj = NULL;
   curPage->bot = botObj = NULL;
   PlaceTopObj(tmp_obj);
   curPage->top = topObj = tmp_top_obj;
   curPage->bot = botObj = tmp_bot_obj;

   dx = tmp_obj->obbox.ltx - ltx;
   dy = tmp_obj->obbox.lty - lty;
   FreeBoxObj(tmp_obj);

   for (obj_ptr=topObj; obj_ptr != NULL; obj_ptr=obj_ptr->next) {
      MoveObj(obj_ptr, dx, dy);
   }
    
   SelAllObj(FALSE);
    
   if (botObj != NULL) {
      botObj->next = saved_top_obj;
   } else {
      curPage->top = topObj = saved_top_obj;
   }

   if (saved_top_obj != NULL) {
      saved_top_obj->prev = botObj;
      curPage->bot = botObj = saved_bot_obj;
   }
   RedrawDrawWindow(botObj);
   PrepareToRecord(CMD_NEW, NULL, NULL, 0);
   RecordCmd(CMD_NEW, NULL, topSel, botSel, numObjSelected);
   HighLightForward();

   Msg("Objects pasted from tgif.");
   return TRUE;
}        

int PasteFromFile()
   /* returns FALSE if pasting in text mode and non-tgif bytes are */
   /*    in the cut buffer -- this is interpreted as an attempt to */
   /*    paste into the current text */
{
   char file_name[MAXPATHLENGTH+1];
   FILE *fp;
   char inbuf[MAXSTRING+1], * cut_buffer=NULL;
   int size=0, pos=0;

   if (SelectFileNameToPaste("Please select a file to PASTE ...", file_name) ==
         INVALID) {
      return TRUE;
   } else if (FileIsRemote(file_name)) {
      MsgBox("Pasting remote file not supported.", TOOL_NAME, INFO_MB);
      return TRUE;
   }
   if (curChoice == DRAWTEXT) {
      XEvent ev;

      pasteInDrawTextMode = TRUE;
      pasteFromFileInDrawTextMode = TRUE;
      strcpy(pasteFromFileName, file_name);
      ev.type = KeyPress;
      DrawText(&ev);
      return FALSE;
   }
   if ((fp=fopen(file_name, "r")) == NULL) {
      sprintf(gszMsgBox, "Cannot open '%s' for read.", file_name);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return TRUE;
   }
   while (fgets(inbuf, MAXSTRING, fp) != NULL) size += strlen(inbuf);
   fclose(fp);
   if (size == 0) {
      sprintf(gszMsgBox, "File '%s' is empty.", file_name);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return TRUE;
   }
   cut_buffer = (char*)malloc((size+2)*sizeof(char));
   if (cut_buffer == NULL) {
      sprintf(gszMsgBox, "Cannot malloc %1d bytes.", size+2);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return TRUE;
   }
   if ((fp=fopen(file_name, "r")) == NULL) {
      sprintf(gszMsgBox, "Cannot open '%s' for read.", file_name);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return TRUE;
   }
   while (fgets(&cut_buffer[pos], MAXSTRING, fp) != NULL) {
      pos += strlen(&cut_buffer[pos]);
   }
   fclose(fp);
#ifndef _NO_NKF
   if (useNKF) {
      char *out_buffer=(char*)malloc(size+2);

      if (out_buffer == NULL) FailAllocMessage();
      do_kconv(cut_buffer, out_buffer, size+2, _EUC, _AUTO );
      PasteString(out_buffer);
      free(out_buffer);
   } else {
#ifndef _NO_KINPUT
      if (copyAndPasteJIS) {
         CvtJisToEuc(cut_buffer, cut_buffer);
      }
#endif /* ~_NO_KINPUT */
      PasteString(cut_buffer);
   }
#else /* _NO_NKF */
   PasteString(cut_buffer);
#endif /* ~_NO_NKF */

   return TRUE;
}

char *GetCompoundTextBytesFromSelection(pulLen)
   unsigned long *pulLen;
   /* the returned string should be freed by calling UtilFree() */
{
   int done=FALSE, error=FALSE;
   unsigned long ulCurSize=0L, ulTotalLeft=0x10000L;
   char *cut_buffer=NULL;
   time_t tloc, endtime;
   Atom compound_text_atom=(Atom)0;
   Window selection_owner_win=None;

   if (pulLen != NULL) *pulLen = 0L;

   selection_owner_win = XGetSelectionOwner(mainDisplay, XA_PRIMARY);
   if (selection_owner_win == None) {
      sprintf(gszMsgBox, "There is no selection owner of PRIMARY.");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return NULL;
   }
   compound_text_atom = XInternAtom(mainDisplay, "COMPOUND_TEXT", False);
   XConvertSelection(mainDisplay, XA_PRIMARY, compound_text_atom, XA_PRIMARY,
         mainWindow, CurrentTime);
   XFlush(mainDisplay);

   time(&tloc);
   endtime = tloc + 15;

   SaveStatusStrings();
   SetStringStatus("Converting COMPOUND_TEXT...");
   while (!done) {
      XEvent ev;

      if (XCheckTypedEvent(mainDisplay, SelectionNotify, &ev)) {
         done = TRUE;
         break;
      }
      time(&tloc);
      if (tloc >= endtime) break;
   }
   RestoreStatusStrings();

   if (!done) {
      sprintf(gszMsgBox, "%s.",
            "Timeout occurred when attempting to convert COMPOUND_TEXT");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return NULL;
   }
   done = FALSE;
   for (;;) {
      Atom type_atom=(Atom)0;
      unsigned long ulBytesLeft=0L, ulLen=0L;
      char *psz=NULL;
      int nFormat=0;
      int nStatus=XGetWindowProperty(mainDisplay, mainWindow, XA_PRIMARY,
            ulCurSize, ulTotalLeft, False, AnyPropertyType, &type_atom,
            &nFormat, &ulLen, &ulBytesLeft, (unsigned char **)(&psz));

      XSync(mainDisplay, False);
      if (nStatus != Success || type_atom == None) {
         sprintf(gszMsgBox, "%s.",
               "Problem encountered when converting COMPOUND_TEXT");
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         error = TRUE;
         break;
      }
      if (psz != NULL && ulLen > 0) {
         if (cut_buffer == NULL) {
            cut_buffer = (char*)malloc(ulLen+1);
         } else {
            cut_buffer = (char*)realloc(cut_buffer, ulCurSize+ulLen+1);
         }
         if (cut_buffer == NULL) FailAllocMessage();
         strncpy(&cut_buffer[ulCurSize], psz, ulLen);
         ulCurSize += ulLen;
         cut_buffer[ulCurSize] = '\0';
      }
      if (psz != NULL) XFree(psz);

      if (ulBytesLeft == 0L) {
         break;
      } else {
         ulTotalLeft = ulBytesLeft;
      }
   }
   if (error || cut_buffer == NULL || ulCurSize == 0L) {
      if (cut_buffer != NULL) UtilFree(cut_buffer);
      if (pulLen != NULL) *pulLen = 0L;
      return NULL;
   }
   *pulLen = ulCurSize;

#ifndef _NO_KINPUT
   CvtCompoundTextToEuc(cut_buffer, cut_buffer);
#endif /* ~_NO_KINPUT */
   return cut_buffer;
}

int PasteCompoundText()
{
   unsigned long uLen=0L;
   char *cut_buffer=NULL;

   if (curChoice == DRAWTEXT) {
      XEvent ev;

      pasteInDrawTextMode = TRUE;
      pasteCompoundTextInDrawTextMode = TRUE;
      ev.type = KeyPress;
      DrawText(&ev);
      return FALSE;
   }
   cut_buffer = GetCompoundTextBytesFromSelection(&uLen);
   if (cut_buffer == NULL || uLen == 0L) {
      return TRUE;
   }
   Msg("Paste compound-text from a non-tgif tool.");
   PasteString(cut_buffer);
   XFree(cut_buffer);
   return TRUE;
}

void CleanUpCutBuffer()
{
   if (cutBuffer != NULL) {
      *cutBuffer = '\0';
      free(cutBuffer);
      cutBuffer = NULL;
   }
   copyingToCutBuffer = FALSE;
}

/* ----------------------- Properties ----------------------- */

static
void SkipIntProp(lWhich, plSkip)
   long lWhich, *plSkip;
{
   (*plSkip) |= lWhich;
}

static
void TestStringProp(n1, n2, psz1, psz2, lWhich, plSkip)
   int n1, n2;
   char *psz1, *psz2; /* psz2 cannot be NULL */
   long lWhich, *plSkip;
{
   if (n1 != n2 ||
         ((psz1 == NULL || *psz1 == '\0') && *psz2 != '\0') ||
         ((psz1 != NULL && *psz1 != '\0') && strcmp(psz1, psz2) != 0)) {
      SkipIntProp(lWhich, plSkip);
   }
}

static
void SetStringProp(n1, pn2, psz1, psz2)
   int n1, *pn2;
   char *psz1, *psz2; /* psz2 cannot be NULL */
{
   *pn2 = n1;
   if (psz1 == NULL) {
      *psz2 = '\0';
   } else {
      strcpy(psz2, psz1);
   }
}

static
void TestIntProp(n1, n2, lWhich, plSkip)
   int n1, n2;
   long lWhich, *plSkip;
{
   if (n1 != n2) SkipIntProp(lWhich, plSkip);
}

static
void TestFontProp(nDoubleByte1, nFont1, nStyle1, nDoubleByte2, nFont2,
      nStyle2, lWhich, plSkip)
   int nDoubleByte1, nFont1, nStyle1;
   int nDoubleByte2, nFont2, nStyle2;
   long lWhich, *plSkip;
{
   if (nStyle1 != nStyle2) {
      SkipIntProp(lWhich, plSkip);
   } else if (nDoubleByte1 != nDoubleByte2) {
      SkipIntProp(lWhich, plSkip);
   } else if (nDoubleByte1) {
      if (nFont1 != nFont2) {
         SkipIntProp(lWhich, plSkip);
      }
   } else {
      if (nFont1 != nFont2) {
         SkipIntProp(lWhich, plSkip);
      }
   }
}

static
void TestCTMProp(ctm1, transformed, ctm2, lWhich, plSkip)
   struct XfrmMtrxRec *ctm1, *ctm2;
   int transformed;
   long lWhich, *plSkip;
{
   if ((ctm1 == NULL && transformed) ||
         (ctm1 != NULL && (!transformed ||
         ctm1->m[0] != ctm2->m[0] ||
         ctm1->m[1] != ctm2->m[1] ||
         ctm1->m[2] != ctm2->m[2] ||
         ctm1->m[3] != ctm2->m[3] ||
         ctm1->t[0] != ctm2->t[0] ||
         ctm1->t[1] != ctm2->t[1]))) {
      SkipIntProp(lWhich, plSkip);
   }
}

void SetIntPropertyMask(lWhich, nValue, pszValue, plMask, plSkip, pProp)
   long lWhich, *plMask, *plSkip;
   int nValue;
   char *pszValue;
   struct PropertiesRec *pProp;
{
   if (((*plSkip) & lWhich) == lWhich) {
      /*
       * this property is inconsistent, continue skipping it
       */
   } else if (((*plMask) & lWhich) == lWhich) {
      /*
       * this property was set before, if it's not the same, skip/invalidate it
       */
      switch (lWhich) {
      case PROP_MASK_COLOR:
         TestStringProp(nValue, pProp->color, pszValue, pProp->color_str,
               lWhich, plSkip);
         break;
      case PROP_MASK_WIDTH:
         TestStringProp(nValue, pProp->width, pszValue, pProp->width_spec,
               lWhich, plSkip);
         break;
      case PROP_MASK_AW:
         TestStringProp(nValue, pProp->aw, pszValue, pProp->aw_spec,
               lWhich, plSkip);
         break;
      case PROP_MASK_AH:
         TestStringProp(nValue, pProp->ah, pszValue, pProp->ah_spec,
               lWhich, plSkip);
         break;
      case PROP_MASK_TRANSPAT:
         TestIntProp(nValue, pProp->trans_pat, lWhich, plSkip);
         break;
      case PROP_MASK_FILL:
         TestIntProp(nValue, pProp->fill, lWhich, plSkip);
         break;
      case PROP_MASK_PEN:
         TestIntProp(nValue, pProp->pen, lWhich, plSkip);
         break;
      case PROP_MASK_DASH:
         TestIntProp(nValue, pProp->dash, lWhich, plSkip);
         break;
      case PROP_MASK_ARROW_STYLE:
         TestIntProp(nValue, pProp->arrow_style, lWhich, plSkip);
         break;
      case PROP_MASK_CURVED:
         TestIntProp(nValue, pProp->curved, lWhich, plSkip);
         break;
      case PROP_MASK_RCB_RADIUS:
         TestIntProp(nValue, pProp->rcb_radius, lWhich, plSkip);
         break;
      case PROP_MASK_TEXT_JUST:
         TestIntProp(nValue, pProp->text_just, lWhich, plSkip);
         break;
      case PROP_MASK_TEXT_SZ_UNIT:
         TestIntProp(nValue, pProp->text_sz_unit, lWhich, plSkip);
         break;
      }
   } else {
      /* this property is never set, go set it */
      (*plMask) |= lWhich;
      switch (lWhich) {
      case PROP_MASK_COLOR:
         SetStringProp(nValue, &pProp->color, pszValue, pProp->color_str);
         break;
      case PROP_MASK_WIDTH:
         SetStringProp(nValue, &pProp->width, pszValue, pProp->width_spec);
         break;
      case PROP_MASK_AW:
         SetStringProp(nValue, &pProp->aw, pszValue, pProp->aw_spec);
         break;
      case PROP_MASK_AH:
         SetStringProp(nValue, &pProp->ah, pszValue, pProp->ah_spec);
         break;
      case PROP_MASK_TRANSPAT: pProp->trans_pat = nValue; break;
      case PROP_MASK_FILL: pProp->fill = nValue; break;
      case PROP_MASK_PEN: pProp->pen = nValue; break;
      case PROP_MASK_DASH: pProp->dash = nValue; break;
      case PROP_MASK_ARROW_STYLE: pProp->arrow_style = nValue; break;
      case PROP_MASK_CURVED: pProp->curved = nValue; break;
      case PROP_MASK_RCB_RADIUS: pProp->rcb_radius = nValue; break;
      case PROP_MASK_TEXT_JUST: pProp->text_just = nValue; break;
      case PROP_MASK_TEXT_SZ_UNIT: pProp->text_sz_unit = nValue; break;
      case PROP_MASK_VSPACE: pProp->v_space = nValue; break;
      }
   }
}

void SetFontPropertyMask(nDoubleByte, nFont, nStyle, plMask, plSkip, pProp)
   int nDoubleByte, nFont, nStyle;
   long *plMask, *plSkip;
   struct PropertiesRec *pProp;
{
   if (((*plSkip) & PROP_MASK_TEXT_FONT) == PROP_MASK_TEXT_FONT) {
      /*
       * this property is inconsistent, continue skipping it
       */
   } else if (((*plMask) & PROP_MASK_TEXT_FONT) == PROP_MASK_TEXT_FONT) {
      /*
       * this property was set before, if it's not the same, skip/invalidate it
       */
      TestFontProp(nDoubleByte, nFont, nStyle, pProp->double_byte,
            pProp->text_font, pProp->text_style, PROP_MASK_TEXT_FONT, plSkip);
   } else {
      /* this property is never set, go set it */
      (*plMask) |= PROP_MASK_TEXT_FONT;
      pProp->double_byte = nDoubleByte;
      pProp->text_font = nFont;
      pProp->text_style = nStyle;
   }
}

void SetCTMPropertyMask(ctm, plMask, plSkip, pProp)
   struct XfrmMtrxRec *ctm;
   long *plMask, *plSkip;
   struct PropertiesRec *pProp;
{
   if (((*plSkip) & PROP_MASK_CTM) == PROP_MASK_CTM) {
      /*
       * this property is inconsistent, continue skipping it
       */
   } else if (((*plMask) & PROP_MASK_CTM) == PROP_MASK_CTM) {
      /*
       * this property was set before, if it's not the same, skip/invalidate it
       */
      TestCTMProp(ctm, pProp->transformed, &pProp->ctm, PROP_MASK_CTM, plSkip);
   } else {
      /* this property is never set, go set it */
      (*plMask) |= PROP_MASK_CTM;
      if (ctm == NULL) {
         pProp->transformed = FALSE;
      } else {
         pProp->transformed = TRUE;
         memcpy(&pProp->ctm, ctm, sizeof(struct XfrmMtrxRec));
      }
   }
}

static
void SetPropMask(ObjPtr, plMask, plSkip, pProp)
   struct ObjRec *ObjPtr;
   long *plMask, *plSkip;
   struct PropertiesRec *pProp;
{
   if (curChoice == DRAWTEXT) {
      if (textCursorShown && textHighlight) {
         SetIntPropertyMask(PROP_MASK_TRANSPAT, transPat, NULL,
               plMask, plSkip, pProp);
         SetIntPropertyMask(PROP_MASK_FILL, objFill, NULL,
               plMask, plSkip, pProp);
         SetIntPropertyMask(PROP_MASK_PEN, penPat, NULL,
               plMask, plSkip, pProp);

         SetIntPropertyMask(PROP_MASK_TEXT_JUST, textJust, NULL,
               plMask, plSkip, pProp);
         SetIntPropertyMask(PROP_MASK_VSPACE, textVSpace, NULL,
               plMask, plSkip, pProp);

         if (HighlightedTextHasSameProperty(PROP_MASK_COLOR,
               curStrBlock->seg->color, TRUE)) {
            SetIntPropertyMask(PROP_MASK_COLOR, colorIndex,
                  colorMenuItems[colorIndex], plMask, plSkip, pProp);
         }
         if (HighlightedTextHasSameProperty(PROP_MASK_TEXT_SZ_UNIT,
               curStrBlock->seg->sz_unit, TRUE)) {
            SetIntPropertyMask(PROP_MASK_TEXT_SZ_UNIT, curSzUnit, NULL,
                  plMask, plSkip, pProp);
         }
         if (HighlightedTextHasSameProperty(PROP_MASK_TEXT_FONT,
               curStrBlock->seg->font, TRUE) &&
               HighlightedTextHasSameProperty(PROP_MASK_TEXT_STYLE,
               curStrBlock->seg->style, TRUE)) {
            SetFontPropertyMask(canvasFontDoubleByte, curFont, curStyle, plMask,
                  plSkip, pProp);
         }
      } else {
         SetIntPropertyMask(PROP_MASK_TRANSPAT, transPat, NULL,
               plMask, plSkip, pProp);
         SetIntPropertyMask(PROP_MASK_FILL, objFill, NULL,
               plMask, plSkip, pProp);
         SetIntPropertyMask(PROP_MASK_PEN, penPat, NULL,
               plMask, plSkip, pProp);

         SetIntPropertyMask(PROP_MASK_TEXT_JUST, textJust, NULL,
               plMask, plSkip, pProp);
         SetIntPropertyMask(PROP_MASK_VSPACE, textVSpace, NULL,
               plMask, plSkip, pProp);

         SetIntPropertyMask(PROP_MASK_COLOR, colorIndex,
               colorMenuItems[colorIndex], plMask, plSkip, pProp);
         SetIntPropertyMask(PROP_MASK_TEXT_SZ_UNIT, curSzUnit, NULL,
               plMask, plSkip, pProp);
         SetFontPropertyMask(canvasFontDoubleByte, curFont, curStyle, plMask,
               plSkip, pProp);
      }
   } else if (ObjPtr == NULL) {
      double radian=((double)(textRotation))*M_PI/180.0/64.0;
      double sin_val=sin(radian);
      double cos_val=cos(radian);
      struct XfrmMtrxRec ctm;

      ctm.m[CTM_SX] = ctm.m[CTM_SY] = ((double)1000)*cos_val;
      ctm.m[CTM_SIN] = ((double)1000)*sin_val;
      ctm.m[CTM_MSIN] = (-ctm.m[CTM_SIN]);
      ctm.t[CTM_TX] = ctm.t[CTM_TY] = 0;
      SetCTMPropertyMask(&ctm, plMask, plSkip, pProp);

      SetIntPropertyMask(PROP_MASK_COLOR, colorIndex,
            colorMenuItems[colorIndex], plMask, plSkip, pProp);
      SetIntPropertyMask(PROP_MASK_WIDTH, curWidthOfLine[lineWidth],
            curWidthOfLineSpec[lineWidth],
            plMask, plSkip, pProp);
      SetIntPropertyMask(PROP_MASK_AW, curArrowHeadW[lineWidth],
            curArrowHeadWSpec[lineWidth],
            plMask, plSkip, pProp);
      SetIntPropertyMask(PROP_MASK_AH, curArrowHeadH[lineWidth],
            curArrowHeadHSpec[lineWidth],
            plMask, plSkip, pProp);

      SetIntPropertyMask(PROP_MASK_TRANSPAT, transPat, NULL,
            plMask, plSkip, pProp);
      SetIntPropertyMask(PROP_MASK_FILL, objFill, NULL,
            plMask, plSkip, pProp);
      SetIntPropertyMask(PROP_MASK_PEN, penPat, NULL,
            plMask, plSkip, pProp);
      SetIntPropertyMask(PROP_MASK_DASH, curDash, NULL,
            plMask, plSkip, pProp);
      SetIntPropertyMask(PROP_MASK_CURVED, curSpline, NULL,
            plMask, plSkip, pProp);
      SetIntPropertyMask(PROP_MASK_ARROW_STYLE, lineStyle, NULL,
            plMask, plSkip, pProp);
      SetIntPropertyMask(PROP_MASK_RCB_RADIUS, rcbRadius, NULL,
            plMask, plSkip, pProp);

      SetIntPropertyMask(PROP_MASK_TEXT_JUST, textJust, NULL,
            plMask, plSkip, pProp);
      SetIntPropertyMask(PROP_MASK_TEXT_SZ_UNIT, curSzUnit, NULL,
            plMask, plSkip, pProp);
      SetIntPropertyMask(PROP_MASK_VSPACE, textVSpace, NULL,
            plMask, plSkip, pProp);

      SetFontPropertyMask(canvasFontDoubleByte, curFont, curStyle, plMask,
            plSkip, pProp);
   } else {
      struct ObjRec *obj_ptr=NULL;

      switch (ObjPtr->type) {
      case OBJ_POLY: SetPolyPropMask(ObjPtr, plMask, plSkip, pProp); break;
      case OBJ_BOX: SetBoxPropMask(ObjPtr, plMask, plSkip, pProp); break;
      case OBJ_OVAL: SetOvalPropMask(ObjPtr, plMask, plSkip, pProp); break;
      case OBJ_TEXT: SetTextPropMask(ObjPtr, plMask, plSkip, pProp); break;
      case OBJ_POLYGON:
         SetPolygonPropMask(ObjPtr, plMask, plSkip, pProp);
         break;
      case OBJ_ARC: SetArcPropMask(ObjPtr, plMask, plSkip, pProp); break;
      case OBJ_RCBOX: SetRCBoxPropMask(ObjPtr, plMask, plSkip, pProp); break;
      case OBJ_XBM: SetXBmPropMask(ObjPtr, plMask, plSkip, pProp); break;
      case OBJ_XPM: SetXPmPropMask(ObjPtr, plMask, plSkip, pProp); break;

      case OBJ_GROUP:
      case OBJ_ICON:
      case OBJ_SYM:
      case OBJ_PIN:
         for (obj_ptr=ObjPtr->detail.r->first; obj_ptr != NULL;
               obj_ptr=obj_ptr->next) {
            SetPropMask(ObjPtr, plMask, plSkip, pProp);
         }
         break;
      }
   }
}

static
int GetAngleFromCTM(ctm)
   struct XfrmMtrxRec *ctm;
   /* returned angle is degree*64 */
{
   double sin_val=(double)(ctm->m[CTM_SIN]/((double)1000));
   double cos_val=(double)(ctm->m[CTM_SX]/((double)1000));
   double radian=(double)0;
   int angle=0;

   if (fabs(sin_val) < EQ_TOL) {
      if (cos_val > 0) {
         radian = (double)0;
      } else {
         radian = (double)M_PI;
      }
   } else if (fabs(cos_val) < EQ_TOL) {
      if (sin_val> 0) {
         radian = (double)(M_PI/2.0);
      } else {
         radian = (double)(M_PI*3.0/2.0);
      }
   } else {
      radian = (double)atan(sin_val/cos_val);
      if (radian >= 0) {
         if (sin_val < 0) {
            radian += (double)M_PI;
         }
      } else {
         if (cos_val > 0) {
            radian += (double)(M_PI*2.0);
         } else {
            radian += (double)M_PI;
         }
      }
   }
   angle = (int)(radian*((double)180)*((double)64)/M_PI);
   return angle;
}

static
void DoPasteAProperty(lWhich, nPasteSel, pProp)
   long lWhich;
   int nPasteSel;
   struct PropertiesRec *pProp;
{
   char szBuf[MAXSTRING];
   int index=0, new_alloc=FALSE;

   switch (lWhich) {
   case PROP_MASK_COLOR:
      index = QuickFindColorIndex(NULL, pProp->color_str, &new_alloc, FALSE);
      if (index == INVALID) {
         sprintf(gszMsgBox, "%s '%s' %s.",
               "Fail to allocate the", pProp->color_str, "color");
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         return;
      } else {
         ChangeAllSelColor(index, TRUE);
      }
      break;
   case PROP_MASK_WIDTH:
      ChangeAllSelRealLineWidth(CHANGE_WIDTH, pProp->width, (-1), (-1),
            pProp->width_spec, NULL, NULL, TRUE);
      break;
   case PROP_MASK_AW:
      ChangeAllSelRealLineWidth(CHANGE_AW, (-1), pProp->aw, (-1),
            NULL, pProp->aw_spec, NULL, TRUE);
      break;
   case PROP_MASK_AH:
      ChangeAllSelRealLineWidth(CHANGE_AH, (-1), (-1), pProp->ah,
            NULL, NULL, pProp->ah_spec, TRUE);
      break;

   case PROP_MASK_TRANSPAT: ChangeAllSelTransPat(pProp->trans_pat, TRUE); break;
   case PROP_MASK_FILL: ChangeAllSelFill(pProp->fill, TRUE); break;
   case PROP_MASK_PEN: ChangeAllSelPen(pProp->pen, TRUE); break;
   case PROP_MASK_DASH: ChangeAllSelLineDash(pProp->dash, TRUE); break;
   case PROP_MASK_ARROW_STYLE:
      ChangeAllSelLineStyle(pProp->arrow_style, TRUE);
      break;
   case PROP_MASK_CURVED: ChangeAllSelLineType(pProp->curved, TRUE); break;
   case PROP_MASK_RCB_RADIUS: ChangeAllSelRCBRadius(pProp->rcb_radius); break;
   case PROP_MASK_TEXT_JUST: ChangeFontJust(pProp->text_just); break;
   case PROP_MASK_TEXT_SZ_UNIT:
      sprintf(szBuf, "%1d", SzUnitToFontSize(pProp->text_sz_unit));
      SetSelFontSize(szBuf);
      break;
   case PROP_MASK_VSPACE: ChangeVSpace(pProp->v_space); break;

   case PROP_MASK_TEXT_FONT:
      ChangeFont(pProp->text_font, TRUE);
      ChangeFontStyle(pProp->text_style);
      break;

   case PROP_MASK_CTM:
      if (topSel == NULL) {
         if (pProp->transformed) {
            int angle=GetAngleFromCTM(&pProp->ctm);

            FormatAngle(angle, szBuf);
         } else {
            strcpy(szBuf, "0");
         }
         SetTextRotation(szBuf);
      } else {
         SetSelCTM(pProp->transformed, &pProp->ctm);
      }
      break;
   }
}

static char gszPropIniFile[MAXPATHLENGTH];

static char gszCopyPasteSec[MAXSTRING];
static char gszCopyPasteBackupSec[MAXSTRING];
static char gszPropSetSec[MAXSTRING];
static char gszPropProfilePrefix[MAXSTRING];

struct PropInfoRec {
   long bit;
   int checked;
   char *desc;
} gstPropInfo[] = {
   { PROP_MASK_AH,           FALSE, "arrow height" },
   { PROP_MASK_AW,           FALSE, "arrow width" },
   { PROP_MASK_ARROW_STYLE,  FALSE, "arrow style" },
   { PROP_MASK_COLOR,        FALSE, "color" },
   { PROP_MASK_DASH,         FALSE, "dash" },
   { PROP_MASK_WIDTH,        FALSE, "line width" },
   { PROP_MASK_TRANSPAT,     FALSE, "pattern transparency" },
   { PROP_MASK_FILL,         FALSE, "fill" },
   { PROP_MASK_PEN,          FALSE, "pen" },
   { PROP_MASK_RCB_RADIUS,   FALSE, "rcbox radius" },
   { PROP_MASK_CURVED,       FALSE, "spline" },
   { PROP_MASK_TEXT_FONT,    FALSE, "text font" },
   { PROP_MASK_TEXT_STYLE,   FALSE, "text style (no use, part of text font)" },
   { PROP_MASK_TEXT_JUST,    FALSE, "text justification" },
   { PROP_MASK_TEXT_SZ_UNIT, FALSE, "text size" },
   { PROP_MASK_VSPACE,       FALSE, "text vertical spacing" },
   { PROP_MASK_CTM,          FALSE, "transformation matrix" },
   { 0L, FALSE, NULL }
};

static
void WriteIntProp(pszSec, pszKey, nValue)
   char *pszSec, *pszKey;
   int nValue;
{
   sprintf(gszMsgBox, "%1d", nValue);
   tgWriteProfileString(pszSec, pszKey, gszMsgBox, gszPropIniFile);
}

static
void WriteStringProp(pszSec, pszKey, nValue, pszValue)
   char *pszSec, *pszKey;
   int nValue;
   char *pszValue;
{
   sprintf(gszMsgBox, "%1d,%s", nValue, (pszValue==NULL ? "" : pszValue));
   tgWriteProfileString(pszSec, pszKey, gszMsgBox, gszPropIniFile);
}

static
void WriteFontProp(pszSec, pszKey, nFont, nStyle)
   char *pszSec, *pszKey;
   int nFont, nStyle;
{
   char font_str[MAXSTRING];

   *font_str = '\0';
   GetPSFontStr(nFont, nStyle, font_str);
   /* font_str starts with the '/' character */
   sprintf(gszMsgBox, "%1d,%s", nStyle, &font_str[1]);
   tgWriteProfileString(pszSec, pszKey, gszMsgBox, gszPropIniFile);
}

static
void WriteCTMProp(pszSec, pszKey, nTransformed, ctm)
   char *pszSec, *pszKey;
   int nTransformed;
   struct XfrmMtrxRec *ctm;
{
   if (nTransformed) {
      sprintf(gszMsgBox, "%1d,%g,%g,%g,%g,%1d,%1d", nTransformed,
            ctm->m[CTM_SX], ctm->m[CTM_SIN], ctm->m[CTM_MSIN], ctm->m[CTM_SY],
            ctm->t[CTM_TX], ctm->t[CTM_TY]);
   } else {
      sprintf(gszMsgBox, "%1d,%g,%g,%g,%g,%1d,%1d", nTransformed,
            (double)1000, (double)0, (double)0, (double)1000, 0, 0);
   }
   tgWriteProfileString(pszSec, pszKey, gszMsgBox, gszPropIniFile);
}

static
int ParseIntProp(pszBuf, pnValue)
   char *pszBuf;
   int *pnValue;
{
   return (sscanf(pszBuf, "%d", pnValue) == 1);
}

static
int ParseStringProp(pszBuf, pnValue, pszValue)
   char *pszBuf;
   int *pnValue;
   char *pszValue;
{
   char *psz1=strtok(pszBuf, ","), *psz2=NULL;

   if (psz1 == NULL) return FALSE;
   psz2 = strtok(NULL, ",");
   if (psz2 == NULL) return FALSE;

   if (sscanf(psz1, "%d", pnValue) != 1) return FALSE;
   strcpy(pszValue, psz2);

   return TRUE;
}

static
int ParseFontSzUnitProp(pszBuf, pnSzUnit)
   char *pszBuf;
   int *pnSzUnit;
{
   int i, sz_unit=0;

   if (sscanf(pszBuf, "%d", &sz_unit) != 1) return FALSE;
   for (i=0; i < numFontSizes; i++) {
      if (sz_unit == fontSzUnits[i]) {
         *pnSzUnit = sz_unit;
         return TRUE;
      }
   }
   if (topSel == NULL) {
      sprintf(gszMsgBox, "Cannot find font size '%1d'.  %1d %s.",
            SzUnitToFontSize(sz_unit), defaultFontSize, "is used instead");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      *pnSzUnit = FontSizeToSzUnit(defaultFontSize);
   } else {
      *pnSzUnit = sz_unit;
   }
   return TRUE;
}

static
int ParseFontProp(pszBuf, pnDoubleByte, pnFont, pnStyle)
   char *pszBuf;
   int *pnDoubleByte, *pnFont, *pnStyle;
{
   char font_str[MAXSTRING], *psz=NULL;

   *font_str = '\0';
   psz = strchr(pszBuf, ',');
   if (psz == NULL) return FALSE;

   *psz = '\0';
   if (sscanf(pszBuf, "%d", pnStyle) != 1) {
      *psz = ',';
      return FALSE;
   }
   *psz++ = ',';
   strcpy(font_str, psz);
   *pnFont = GetFontIndex(font_str, *pnStyle, TRUE);
   *pnDoubleByte = FALSE;
   if (*pnFont == INVALID) {
      sprintf(gszMsgBox, "Cannot find screen font for '%s'.  %s.",
            font_str, "Times font is used instead");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      *pnDoubleByte = FALSE;
      *pnFont = FONT_TIM;
   }
   return TRUE;
}

static
int ParseCTMProp(pszBuf, pnTransformed, ctm)
   char *pszBuf;
   int *pnTransformed;
   struct XfrmMtrxRec *ctm;
{
   struct PropInfoRec *ppir=NULL;
   char *psz=strchr(pszBuf, ',');

   if (psz == NULL) return FALSE;
   *psz = '\0';
   if (sscanf(pszBuf, "%d", pnTransformed) != 1) {
      *psz = ',';
      return FALSE;
   }
   *psz++ = ',';
   if (sscanf(psz, "%lg , %lg , %lg, %lg , %d , %d",
         &ctm->m[CTM_SX], &ctm->m[CTM_SIN], &ctm->m[CTM_MSIN], &ctm->m[CTM_SY],
         &ctm->t[CTM_TX], &ctm->t[CTM_TY]) != 6) {
      return FALSE;
   }
   if (topSel != NULL) {
      struct SelRec *sel_ptr=NULL;

      for (sel_ptr=topSel; sel_ptr != NULL; sel_ptr=sel_ptr->next) {
         switch (sel_ptr->obj->type) {
         case OBJ_GROUP:
         case OBJ_ICON:
         case OBJ_SYM:
         case OBJ_PIN:
            for (ppir=gstPropInfo; ppir->bit != 0L; ppir++) {
               if (ppir->bit == PROP_MASK_CTM) {
                  sprintf(gszMsgBox, "Cannot paste '%s' %s.  %s, '%s' %s.",
                        ppir->desc, "for compound objects",
                        "Therefore", ppir->desc,
                        "cannot be selected");
                  MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
                  break;
               }
            }
            return FALSE;
         }
      }
   } else if (topSel == NULL && (*pnTransformed)) {
      if ((fabs(ctm->m[CTM_SX]-ctm->m[CTM_SY]) < EQ_TOL) &&
            (fabs(ctm->m[CTM_SIN]+ctm->m[CTM_MSIN]) < EQ_TOL)) {
         return TRUE;
      }
      for (ppir=gstPropInfo; ppir->bit != 0L; ppir++) {
         if (ppir->bit == PROP_MASK_CTM) {
            sprintf(gszMsgBox, "The '%s' %s.  %s.", ppir->desc,
                  "in the cut buffer contains non-rotational components",
                  "Text rotation of 0 degree is used");
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            break;
         }
      }
      *pnTransformed = FALSE;
   }
   return TRUE;
}

static
void WritePropToIni(lWhich, pszSec, pszKey, pProp)
   long lWhich;
   char *pszSec, *pszKey;
   struct PropertiesRec *pProp;
{
   switch (lWhich) {
   case PROP_MASK_AH:
      WriteStringProp(pszSec, pszKey, pProp->ah, pProp->ah_spec);
      break;
   case PROP_MASK_AW:
      WriteStringProp(pszSec, pszKey, pProp->aw, pProp->aw_spec);
      break;
   case PROP_MASK_ARROW_STYLE:
      WriteIntProp(pszSec, pszKey, pProp->arrow_style);
      break;
   case PROP_MASK_COLOR:
      WriteStringProp(pszSec, pszKey, pProp->color, pProp->color_str);
      break;

   case PROP_MASK_DASH: WriteIntProp(pszSec, pszKey, pProp->dash); break;
   case PROP_MASK_WIDTH:
      WriteStringProp(pszSec, pszKey, pProp->width, pProp->width_spec);
      break;
   case PROP_MASK_TRANSPAT:
      WriteIntProp(pszSec, pszKey, pProp->trans_pat);
      break;
   case PROP_MASK_FILL: WriteIntProp(pszSec, pszKey, pProp->fill); break;
   case PROP_MASK_PEN: WriteIntProp(pszSec, pszKey, pProp->pen); break;
   case PROP_MASK_RCB_RADIUS:
      WriteIntProp(pszSec, pszKey, pProp->rcb_radius);
      break;
   case PROP_MASK_CURVED: WriteIntProp(pszSec, pszKey, pProp->curved); break;
   case PROP_MASK_TEXT_FONT:
      WriteFontProp(pszSec, pszKey, pProp->text_font, pProp->text_style);
      break;
   case PROP_MASK_TEXT_JUST:
      WriteIntProp(pszSec, pszKey, pProp->text_just);
      break;
   case PROP_MASK_TEXT_SZ_UNIT:
      WriteIntProp(pszSec, pszKey, pProp->text_sz_unit);
      break;
   case PROP_MASK_VSPACE: WriteIntProp(pszSec, pszKey, pProp->v_space); break;
   case PROP_MASK_CTM:
      WriteCTMProp(pszSec, pszKey, pProp->transformed, &pProp->ctm);
      break;
   }
}

static
int ParseAProp(lWhich, pszBuf, pProp)
   long lWhich;
   char *pszBuf;
   struct PropertiesRec *pProp;
{
   switch (lWhich) {
   case PROP_MASK_AH:
      return ParseStringProp(pszBuf, &pProp->ah, pProp->ah_spec);
   case PROP_MASK_AW:
      return ParseStringProp(pszBuf, &pProp->aw, pProp->aw_spec);
   case PROP_MASK_ARROW_STYLE: return ParseIntProp(pszBuf, &pProp->arrow_style);
   case PROP_MASK_COLOR:
      return ParseStringProp(pszBuf, &pProp->color, pProp->color_str);
   case PROP_MASK_DASH: return ParseIntProp(pszBuf, &pProp->dash);
   case PROP_MASK_WIDTH:
      return ParseStringProp(pszBuf, &pProp->width, pProp->width_spec);
   case PROP_MASK_TRANSPAT: return ParseIntProp(pszBuf, &pProp->trans_pat);
   case PROP_MASK_FILL: return ParseIntProp(pszBuf, &pProp->fill);
   case PROP_MASK_PEN: return ParseIntProp(pszBuf, &pProp->pen);
   case PROP_MASK_RCB_RADIUS: return ParseIntProp(pszBuf, &pProp->rcb_radius);
   case PROP_MASK_CURVED: return ParseIntProp(pszBuf, &pProp->curved);
   case PROP_MASK_TEXT_FONT:
      return ParseFontProp(pszBuf, &pProp->double_byte, &pProp->text_font,
            &pProp->text_style);
   case PROP_MASK_TEXT_JUST: return ParseIntProp(pszBuf, &pProp->text_just);
   case PROP_MASK_TEXT_SZ_UNIT:
      return ParseFontSzUnitProp(pszBuf, &pProp->text_sz_unit);
   case PROP_MASK_VSPACE: return ParseIntProp(pszBuf, &pProp->v_space);
   case PROP_MASK_CTM:
      return ParseCTMProp(pszBuf, &pProp->transformed, &pProp->ctm);
   }
   return FALSE;
}

static
int ReadPropFromIni(pszSec, plMask, pCheckArray, pProp)
   /*
    * This routine sets *plMask according to what's in the ini file.
    *     All associated fields in pProp are filled in.
    * This routine sets pCheckArray->num_cols and pCheckArray->num_rows;
    *     pCheckArray->value is set to NULL.
    */
   char *pszSec;
   long *plMask;
   struct CheckArrayRec *pCheckArray;
   struct PropertiesRec *pProp;
{
   struct PropInfoRec *ppir=NULL;
   char *pszKey=NULL;
   char *pszKeys=tgGetProfileString(pszSec, NULL, gszPropIniFile);

   *plMask = 0L;
   for (ppir=gstPropInfo; ppir->bit != 0L; ppir++) ppir->checked = FALSE;
   pCheckArray->num_cols = 1;

   if (pszKeys == NULL) {
      sprintf(gszMsgBox, "No property to paste!");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return FALSE;
   }
   for (pszKey=pszKeys; *pszKey != '\0'; pszKey++) {
      for (ppir=gstPropInfo; ppir->bit != 0L; ppir++) {
         if (strcmp(pszKey, ppir->desc) == 0) {
            break;
         }
      }
      if (ppir != NULL) {
         char *pszValue=tgGetProfileString(pszSec, pszKey, gszPropIniFile);

         if (topSel == NULL && (ppir->bit == PROP_MASK_WIDTH ||
               ppir->bit == PROP_MASK_AW || ppir->bit == PROP_MASK_AH)) {
         } else {
            if (ParseAProp(ppir->bit, ((pszValue==NULL)?"":pszValue), pProp)) {
               ppir->checked = TRUE;
               pCheckArray->num_rows++;
               *plMask |= ppir->bit;
            }
         }
         if (pszValue != NULL) tgFreeProfileString(pszValue);
      } else {
         sprintf(gszMsgBox, "Invalid key: '%s' in the [%s] section of '%s'.",
               pszKey, pszSec, gszPropIniFile);
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      }
      pszKey += strlen(pszKey);
   }
   tgFreeProfileString(pszKeys);
   return TRUE;
}

#define COPY_BUTTON    101
#define PASTE_BUTTON   102
#define SAVE_BUTTON    103
#define RESTORE_BUTTON 104

void CleanUpProperties()
{
}

void InitProperties()
{
   sprintf(gszPropIniFile, "%s/cutpaste.ini", tgifDir);

   strcpy(gszCopyPasteSec, "Copy/Paste Properties");
   strcpy(gszCopyPasteBackupSec, "Copy/Paste Properties - Backup");
   strcpy(gszPropSetSec, "Property Set Names");
   strcpy(gszPropProfilePrefix, "Property Profile");
}

static
void FormatPropForDisplay(lWhich, pProp, ppir, pszBuf)
   long lWhich;
   struct PropertiesRec *pProp;
   struct PropInfoRec *ppir;
   char *pszBuf;
{
   char font_str[MAXSTRING];
   int nLen=0;

   sprintf(pszBuf, "%s: ", ppir->desc);
   nLen = strlen(pszBuf);

   switch (lWhich) {
   case PROP_MASK_CTM: strcat(pszBuf, "..."); break;

   case PROP_MASK_COLOR: strcat(pszBuf, pProp->color_str); break;
   case PROP_MASK_WIDTH: strcat(pszBuf, pProp->width_spec); break;
   case PROP_MASK_AW: strcat(pszBuf, pProp->aw_spec); break;
   case PROP_MASK_AH: strcat(pszBuf, pProp->ah_spec); break;

   case PROP_MASK_TRANSPAT:
      sprintf(&pszBuf[nLen], "%1d", pProp->trans_pat);
      break;
   case PROP_MASK_FILL: sprintf(&pszBuf[nLen], "%1d", pProp->fill); break;
   case PROP_MASK_PEN: sprintf(&pszBuf[nLen], "%1d", pProp->pen); break;
   case PROP_MASK_DASH: sprintf(&pszBuf[nLen], "%1d", pProp->dash); break;
   case PROP_MASK_ARROW_STYLE:
      sprintf(&pszBuf[nLen], "%1d", pProp->arrow_style);
      break;
   case PROP_MASK_CURVED: sprintf(&pszBuf[nLen], "%1d", pProp->curved); break;
   case PROP_MASK_RCB_RADIUS:
      sprintf(&pszBuf[nLen], "%1d", pProp->rcb_radius);
      break;
   case PROP_MASK_TEXT_JUST:
      sprintf(&pszBuf[nLen], "%1d", pProp->text_just);
      break;
   case PROP_MASK_TEXT_SZ_UNIT:
      if (showFontSizeInPoints) {
         sprintf(&pszBuf[nLen], "%1dpt",
               SzUnitToPointSize(pProp->text_sz_unit));
      } else {
         sprintf(&pszBuf[nLen], "%1d", SzUnitToFontSize(pProp->text_sz_unit));
      }
      break;
   case PROP_MASK_VSPACE: sprintf(&pszBuf[nLen], "%1d", pProp->v_space); break;

   case PROP_MASK_TEXT_FONT:
      *font_str = '\0';
      GetPSFontStr(pProp->text_font, pProp->text_style, font_str);
      /* font_str starts with the '/' character */
      strcat(pszBuf, &font_str[1]);
      break;
   }
}

static
int GetPropSelection(pp_dsp_ptr, ppsz_entries, pn_num_entries, pn_marked_index,
      pp_check_array, cur_buf, p_void)
   DspList **pp_dsp_ptr;
   char ***ppsz_entries, *cur_buf;
   int *pn_num_entries, *pn_marked_index;
   struct CheckArrayRec **pp_check_array;
   void *p_void;
{
   struct PropertiesRec *pProp=(struct PropertiesRec *)p_void;
   struct PropInfoRec *ppir=NULL;
   int num_rows=0;
   char **ppsz=NULL;

   for (ppir=gstPropInfo; ppir->bit != 0L; ppir++) {
      if (ppir->checked) {
         num_rows++;
      }
   }
   *ppsz_entries = (char**)malloc(num_rows*sizeof(char*));
   if (*ppsz_entries == NULL) FailAllocMessage();
   memset(*ppsz_entries, 0, num_rows*sizeof(char*));

   for (ppir=gstPropInfo, ppsz=(*ppsz_entries); ppir->bit != 0L; ppir++) {
      if (ppir->checked) {
         char szBuf[MAXSTRING];

         *szBuf = '\0';
         FormatPropForDisplay(ppir->bit, pProp, ppir, szBuf);
         *ppsz = UtilStrDup(szBuf);
         if ((*ppsz) == NULL) FailAllocMessage();
         ppsz++;
      }
   }
   *pn_num_entries = num_rows;
   return TRUE;
}

static
int DoCopyProperties(pCheckArray, pProp)
   /*
    * Returns FALSE if nothing copied.
    */
   struct CheckArrayRec *pCheckArray;
   struct PropertiesRec *pProp;
{
   struct PropInfoRec *ppir=NULL;
   int index=0, count=0;
   char *pszKeys=NULL;

   tgWriteProfileString(gszCopyPasteSec, NULL, NULL, gszPropIniFile);
   for (ppir=gstPropInfo; ppir->bit != 0L; ppir++) {
      if (ppir->checked) {
         if (pCheckArray->value[0][index]) {
            WritePropToIni(ppir->bit, gszCopyPasteSec, ppir->desc, pProp);
            count++;
         }
         index++;
      }
   }
   tgWriteProfileString(NULL, NULL, NULL, gszPropIniFile);
   if (count == 0) {
      sprintf(gszMsgBox, "No property to copy!");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return FALSE;
   }
   pszKeys = tgGetProfileString(gszCopyPasteSec, NULL, gszPropIniFile);
   if (pszKeys != NULL) {
      int total=(strlen(gszCopyPasteSec)+4), offset=0;
      char *pszKey=NULL;
      char *buf=(char*)malloc(total+1); /* sizeof buf is always total+1 */

      if (buf == NULL) FailAllocMessage();
      memset(buf, 0, total+1);
      *buf = TGIF_HEADER;
      sprintf(&buf[1], "[%s]", gszCopyPasteSec);
      offset = total;

      for (pszKey=pszKeys; *pszKey != '\0'; pszKey++) {
         int key_len=strlen(pszKey);
         char *pszValue=tgGetProfileString(gszCopyPasteSec, pszKey,
               gszPropIniFile);

         if (pszValue != NULL) {
            int value_len=strlen(pszValue);

            buf = (char*)realloc(buf, total+key_len+value_len+3);
            if (buf == NULL) FailAllocMessage();
            sprintf(&buf[offset], "%s=%s", pszKey, pszValue);

            offset += key_len+value_len+2;
            total += key_len+value_len+2;

            tgFreeProfileString(pszValue);
         } else {
            buf = (char*)realloc(buf, total+key_len+3);
            if (buf == NULL) FailAllocMessage();
            sprintf(&buf[offset], "%s=", pszKey);

            offset += key_len+2;
            total += key_len+2;
         }
         pszKey += key_len;
      }
      buf[total] = '\0';
      tgFreeProfileString(pszKeys);

      if (!WriteBufToCutBuffer(buf, total+1)) {
         sprintf(gszMsgBox, "%s  %s", "Copy to cut buffer failed.",
               "Selected object(s) may be too big.");
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      } else {
         sprintf(gszMsgBox, "Copy buffer updated.");
         Msg(gszMsgBox);
      }
      free(buf);
   }
   return TRUE;
}

static
int PropSetExists(pszPropSet)
   char *pszPropSet;
{
   int nReturn=FALSE;
   char *pszKeys=tgGetProfileString(gszPropSetSec, NULL, gszPropIniFile);

   if (pszKeys != NULL) {
      char *pszKey=NULL;

      for (pszKey=pszKeys; *pszKey != '\0'; pszKey++) {
         if (UtilStrICmp(pszPropSet, pszKey) == 0) {
            nReturn = TRUE;
            break;
         }
         pszKey += strlen(pszKey);
      }
      tgFreeProfileString(pszKeys);
   }
   return nReturn;
}

static
int DoSaveProperties(pCheckArray, pProp)
   /*
    * Returns FALSE if nothing saved.
    */
   struct CheckArrayRec *pCheckArray;
   struct PropertiesRec *pProp;
{
   char szPropSet[MAXSTRING], szPropSetSec[MAXSTRING];
   struct PropInfoRec *ppir=NULL;
   int index=0, count=0;

   *szPropSet = *szPropSetSec = '\0';
   sprintf(gszMsgBox, "Please enter a name for the property set.");
   if (Dialog(gszMsgBox, NULL, szPropSet) == INVALID) return FALSE;
   UtilTrimBlanks(szPropSet);
   if (*szPropSet == '\0') return FALSE;
   if (PropSetExists(szPropSet)) {
      sprintf(gszMsgBox, "Property set [%s] already exists.  %s?",
            szPropSet, "Okay to overwrite");
      if (MsgBox(gszMsgBox, TOOL_NAME, YNC_MB) != MB_ID_YES) {
         return FALSE;
      }
   }
   sprintf(szPropSetSec, "%s: %s", gszPropProfilePrefix, szPropSet);
   tgWriteProfileString(gszPropSetSec, szPropSet, "", gszPropIniFile);
   tgWriteProfileString(szPropSetSec, NULL, NULL, gszPropIniFile);
   for (ppir=gstPropInfo; ppir->bit != 0L; ppir++) {
      if (ppir->checked) {
         if (pCheckArray->value[0][index]) {
            WritePropToIni(ppir->bit, szPropSetSec, ppir->desc, pProp);
            count++;
         }
         index++;
      }
   }
   tgWriteProfileString(NULL, NULL, NULL, gszPropIniFile);
   if (count == 0) {
      sprintf(gszMsgBox, "No property to save!");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return FALSE;
   }
   return TRUE;
}

static
int DoPasteProperties(pCheckArray, pProp)
   /*
    * Returns FALSE if nothing pasted.
    */
   struct CheckArrayRec *pCheckArray;
   struct PropertiesRec *pProp;
{
   struct PropInfoRec *ppir=NULL;
   int index=0, count=0;
   int paste_sel=FALSE;

   if ((curChoice == NOTHING || curChoice == VERTEXMODE ||
         curChoice == ROTATEMODE)) {
      paste_sel = (topSel != NULL);
   }
   for (ppir=gstPropInfo; ppir->bit != 0L; ppir++) {
      if (ppir->checked) {
         if (pCheckArray->value[0][index]) {
            count++;
         }
         index++;
      }
   }
   if (count == 0) {
      sprintf(gszMsgBox, "No property to paste!");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return FALSE;
   }
   if (paste_sel) {
      HighLightReverse();
      StartCompositeCmd();
   }
   for (ppir=gstPropInfo, index=0; ppir->bit != 0L; ppir++) {
      if (ppir->checked) {
         if (pCheckArray->value[0][index]) {
            DoPasteAProperty(ppir->bit, paste_sel, pProp);
         }
         index++;
      }
   }
   if (paste_sel) {
      EndCompositeCmd();
      HighLightForward();
   }
   return TRUE;
}

#define COPY_PROP    0
#define PASTE_PROP   1
#define SAVE_PROP    2
#define RESTORE_PROP 3

static
int FreePropSelection(pp_dsp_ptr, ppsz_entries, pn_num_entries, pn_marked_index,
      pp_check_array, cur_buf, btn_id, selected_index, p_void)
   DspList **pp_dsp_ptr;
   char ***ppsz_entries, *cur_buf;
   int *pn_num_entries, *pn_marked_index, btn_id, selected_index;
   struct CheckArrayRec **pp_check_array;
   void *p_void;
{
   struct PropertiesRec *pProp=(struct PropertiesRec *)p_void;
   struct PropInfoRec *ppir=NULL;
   int nReturn=TRUE;

   switch ((int)(pProp->userdata)) {
   case COPY_PROP:
      if (btn_id == COPY_BUTTON) {
         nReturn = !DoCopyProperties(*pp_check_array, pProp);
      }
      break;
   case SAVE_PROP:
      if (btn_id == SAVE_BUTTON) {
         nReturn = !DoSaveProperties(*pp_check_array, pProp);
      }
      break;
   case PASTE_PROP:
      if (btn_id == PASTE_BUTTON) {
         nReturn = !DoPasteProperties(*pp_check_array, pProp);
      }
      break;
   }
   if (*ppsz_entries != NULL) {
      char **ppsz=(*ppsz_entries);

      for (ppir=gstPropInfo; ppir->bit != 0L; ppir++) {
         if (ppir->checked) {
            UtilFree(*ppsz);
            ppsz++;
         }
      }
      free(*ppsz_entries);
      *ppsz_entries = NULL;
   }
   return nReturn;
}

static
int PropSelectionCheckUpdate(pp_dsp_ptr, ppsz_entries, pn_num_entries,
      pn_marked_index, pp_check_array, cur_buf, col, row, p_void)
   DspList **pp_dsp_ptr;
   char ***ppsz_entries, *cur_buf;
   int *pn_num_entries, *pn_marked_index, col, row;
   struct CheckArrayRec **pp_check_array;
   void *p_void;
{
   struct PropertiesRec *pProp=(struct PropertiesRec *)p_void;
   char *psz=NULL, saved_ch='\0', *buf=NULL;
   int checked=FALSE;

   if (ppsz_entries == NULL || *ppsz_entries == NULL ||
         (*ppsz_entries)[row] == NULL || pp_check_array == NULL ||
         (*pp_check_array) == NULL || (*pp_check_array)->value == NULL) {
      return FALSE;
   }
   buf = (*ppsz_entries)[row];
   checked = (*pp_check_array)->value[0][row];

   psz = strchr(buf, ':');
   if (psz != NULL) {
      saved_ch = *psz;
      *psz = '\0';
      switch ((int)(pProp->userdata)) {
      case COPY_PROP:
         sprintf(gszMsgBox, "%s copy the '%s' property",
               (checked ? "Will" : "Will not"), buf);
         break;
      case SAVE_PROP:
         sprintf(gszMsgBox, "%s save the '%s' property",
               (checked ? "Will" : "Will not"), buf);
         break;
      case PASTE_PROP:
         sprintf(gszMsgBox, "%s paste the '%s' property",
               (checked ? "Will" : "Will not"), buf);
         break;
      case RESTORE_PROP:
         sprintf(gszMsgBox, "%s restore the '%s' property",
               (checked ? "Will" : "Will not"), buf);
         break;
      }
      *psz = saved_ch;
   }
   SetStringStatus(gszMsgBox);

   return TRUE;
}

static
void SelectProperties(pszTitle, nWhich, pCheckArray, pProp)
   char *pszTitle;
   int nWhich;
   struct CheckArrayRec *pCheckArray;
   struct PropertiesRec *pProp;
{
   char win_name[128];

   ResetNamesInfo();

   NamesSetTitle(pszTitle);
   switch (nWhich) {
   case COPY_PROP:
      NamesAddButton("COPY", COPY_BUTTON);
      NamesAddButton("CANCEL", BUTTON_CANCEL);
      /* ignore double-click and <CR> */
      NamesSetDefaultBtnId(COPY_BUTTON, INVALID);
      NamesSetEntries(NULL, NULL, 0, pCheckArray, TRUE, INVALID, 0);
      NamesSetStyle(NAMES_SIMPLE_SELECT_NAME, NAMES_LOOP_MANY);
      NamesSetCallback((GetEntriesFunc*)GetPropSelection,
            (AfterLoopFunc*)FreePropSelection,
            (CheckUpdateFunc*)PropSelectionCheckUpdate);
      sprintf(win_name, "%s - Copy Properties", TOOL_NAME);
      break;
   case SAVE_PROP:
      NamesAddButton("SAVE", SAVE_BUTTON);
      NamesAddButton("CANCEL", BUTTON_CANCEL);
      /* ignore double-click and <CR> */
      NamesSetDefaultBtnId(SAVE_BUTTON, INVALID);
      NamesSetEntries(NULL, NULL, 0, pCheckArray, TRUE, INVALID, 0);
      NamesSetStyle(NAMES_SIMPLE_SELECT_NAME, NAMES_LOOP_MANY);
      NamesSetCallback((GetEntriesFunc*)GetPropSelection,
            (AfterLoopFunc*)FreePropSelection,
            (CheckUpdateFunc*)PropSelectionCheckUpdate);
      sprintf(win_name, "%s - Save Properties", TOOL_NAME);
      break;
   case PASTE_PROP:
      NamesAddButton("PASTE", PASTE_BUTTON);
      NamesAddButton("CANCEL", BUTTON_CANCEL);
      /* ignore double-click and <CR> */
      NamesSetDefaultBtnId(PASTE_BUTTON, INVALID);
      NamesSetEntries(NULL, NULL, 0, pCheckArray, TRUE, INVALID, 0);
      NamesSetStyle(NAMES_SIMPLE_SELECT_NAME, NAMES_LOOP_MANY);
      NamesSetCallback((GetEntriesFunc*)GetPropSelection,
            (AfterLoopFunc*)FreePropSelection,
            (CheckUpdateFunc*)PropSelectionCheckUpdate);
      sprintf(win_name, "%s - Paste Properties", TOOL_NAME);
      break;
   case RESTORE_PROP:
      NamesAddButton("RESTORE", RESTORE_BUTTON);
      NamesAddButton("CANCEL", BUTTON_CANCEL);
      /* ignore double-click and <CR> */
      NamesSetDefaultBtnId(RESTORE_BUTTON, INVALID);
      NamesSetEntries(NULL, NULL, 0, pCheckArray, TRUE, INVALID, 0);
      NamesSetStyle(NAMES_SIMPLE_SELECT_NAME, NAMES_LOOP_MANY);
      NamesSetCallback((GetEntriesFunc*)GetPropSelection,
            (AfterLoopFunc*)FreePropSelection,
            (CheckUpdateFunc*)PropSelectionCheckUpdate);
      sprintf(win_name, "%s - Restore Properties", TOOL_NAME);
      break;
   }
   Names(win_name, NULL, NULL, 0, pProp);
}

static
int PrepareToCopyProperties(pProp, lMask, lSkip, pCheckArray)
   struct PropertiesRec *pProp;
   long lMask, lSkip;
   struct CheckArrayRec *pCheckArray;
{
   struct PropInfoRec *ppir=NULL;
   int index=0, num_rows=0;

   for (ppir=gstPropInfo; ppir->bit != 0L; ppir++) {
      ppir->checked = FALSE;
      if ((lSkip & ppir->bit) == 0) {
         num_rows++;
      }
   }
   pCheckArray->num_cols = 1;
   pCheckArray->num_rows = num_rows;

   pCheckArray->value = (int**)malloc(sizeof(int*));
   if (pCheckArray->value == NULL) FailAllocMessage();
   memset(pCheckArray->value, 0, sizeof(int*));

   pCheckArray->value[0] = (int*)malloc(num_rows*sizeof(int));
   if (pCheckArray->value[0] == NULL) FailAllocMessage();
   memset(pCheckArray->value[0], 0, num_rows*sizeof(int));

   for (ppir=gstPropInfo, index=0; ppir->bit != 0L; ppir++) {
      long bit=ppir->bit;

      if ((lMask & bit) != 0L && (lSkip & bit) == 0) {
         ppir->checked = TRUE;
         pCheckArray->value[0][index++] = TRUE;
      }
   }
   return TRUE;
}

static
int PrepareToPasteProperties(pszSec, plMask, pCheckArray, pProp)
   char *pszSec;
   struct PropertiesRec *pProp;
   long *plMask;
   struct CheckArrayRec *pCheckArray;
{
   struct PropInfoRec *ppir=NULL;
   int index=0, num_rows=0;

   for (ppir=gstPropInfo; ppir->bit != 0L; ppir++) {
      ppir->checked = FALSE;
   }
   if (!ReadPropFromIni(pszSec, plMask, pCheckArray, pProp)) {
      return FALSE;
   }
   num_rows = pCheckArray->num_rows;

   pCheckArray->value = (int**)malloc(sizeof(int*));
   if (pCheckArray->value == NULL) FailAllocMessage();
   memset(pCheckArray->value, 0, sizeof(int*));

   pCheckArray->value[0] = (int*)malloc(num_rows*sizeof(int));
   if (pCheckArray->value[0] == NULL) FailAllocMessage();
   memset(pCheckArray->value[0], 0, num_rows*sizeof(int));

   for (ppir=gstPropInfo, index=0; ppir->bit != 0L; ppir++) {
      long bit=ppir->bit;

      if (((*plMask) & bit) != 0L) {
         ppir->checked = TRUE;
         pCheckArray->value[0][index++] = TRUE;
      }
   }
   return TRUE;
}

static
int SetupProperties(pProp, plMask, plSkip, pCheckArray, nWhich)
   struct PropertiesRec *pProp;
   long *plMask, *plSkip;
   struct CheckArrayRec *pCheckArray;
   int nWhich;
{
   int copy_sel=FALSE;

   pProp->userdata = (void*)nWhich;
   if ((curChoice == NOTHING || curChoice == VERTEXMODE ||
         curChoice == ROTATEMODE)) {
      if (topSel == NULL) {
         copy_sel = FALSE;
      } else if (topSel == botSel) {
         copy_sel = TRUE;
      } else {
         switch (nWhich) {
         case COPY_PROP:
            sprintf(gszMsgBox,
                  "Please select only one object for CopyProperties().");
            break;
         case SAVE_PROP:
            sprintf(gszMsgBox,
                  "Please select only one object for SaveProperties().");
            break;
         }
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         return FALSE;
      }
   }
   if (copy_sel) {
      SetPropMask(topSel->obj, plMask, plSkip, pProp);
   } else {
      SetPropMask(NULL, plMask, plSkip, pProp);
   }
   *plSkip = ~(*plMask); /* shouldn't do this if we support multiple objects */

   return PrepareToCopyProperties(pProp, *plMask, *plSkip, pCheckArray);
}

void CopyProperties(nPrompt)
   int nPrompt;
{
   struct PropertiesRec properties;
   long lMask=0L, lSkip=0L;
   struct CheckArrayRec check_array;

   memset(&properties, 0, sizeof(struct PropertiesRec));
   memset(&check_array, 0, sizeof(struct CheckArrayRec));
   if (!SetupProperties(&properties, &lMask, &lSkip, &check_array, COPY_PROP)) {
      return;
   }
   tgWriteProfileString(gszCopyPasteSec, NULL, NULL, gszPropIniFile);
   tgWriteProfileString(NULL, NULL, NULL, gszPropIniFile);
   if (nPrompt) {
      SelectProperties("Please uncheck properties you don't want to copy ...",
            COPY_PROP, &check_array, &properties);
   } else {
      DoCopyProperties(&check_array, &properties);
   }
   CleanUpCheckArray(&check_array);
}

void SaveProperties()
{
   struct PropertiesRec properties;
   long lMask=0L, lSkip=0L;
   struct CheckArrayRec check_array;

   memset(&properties, 0, sizeof(struct PropertiesRec));
   memset(&check_array, 0, sizeof(struct CheckArrayRec));
   if (!SetupProperties(&properties, &lMask, &lSkip, &check_array, SAVE_PROP)) {
      return;
   }
   SelectProperties("Please uncheck properties you don't want to save ...",
         SAVE_PROP, &check_array, &properties);
   CleanUpCheckArray(&check_array);
}

void PasteProperties(nPrompt)
   int nPrompt;
{
   struct PropertiesRec properties;
   long lMask=0L, lSkip=0L;
   struct CheckArrayRec check_array;
   int len=0;
   char *cut_buffer=NULL;

   cut_buffer = (char*)FetchCutBuffer(&len);
   if (len == 0) {
      MsgBox("Cut buffer is empty", TOOL_NAME, INFO_MB);
      return;
   }
   sprintf(gszMsgBox, "[%s]", gszCopyPasteSec);
   if (((unsigned char)(*cut_buffer)) != TGIF_HEADER ||
         strcmp(&cut_buffer[1], gszMsgBox) != 0) {
      XFree(cut_buffer);
      return;
   }
   if (!WriteBufToIni(&cut_buffer[1], gszPropIniFile)) {
      sprintf(gszMsgBox, "%s.",
            "Malformed cut buffer read while pasting properties");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);

      XFree(cut_buffer);
      return;
   }
   XFree(cut_buffer);

   memset(&check_array, 0, sizeof(struct CheckArrayRec));
   memset(&properties, 0, sizeof(struct PropertiesRec));
   properties.userdata = (void*)PASTE_PROP;
   if (!PrepareToPasteProperties(gszCopyPasteSec, &lMask, &check_array,
         &properties)) {
      return;
   }
   lSkip = (~lMask);

   if (nPrompt) {
      SelectProperties("Please uncheck properties you don't want to paste ...",
            PASTE_PROP, &check_array, &properties);
   } else {
      DoPasteProperties(&check_array, &properties);
   }
   CleanUpCheckArray(&check_array);
}

static
int GetPropSetNames(pp_dsp_ptr, ppsz_entries, pn_num_entries, pn_marked_index,
      pp_check_array, cur_buf, p_void)
   DspList **pp_dsp_ptr;
   char ***ppsz_entries, *cur_buf;
   int *pn_num_entries, *pn_marked_index;
   struct CheckArrayRec **pp_check_array;
   void *p_void;
{
   int num_rows=0;
   char **ppsz=NULL, *pszName=NULL;
   char *pszNames=tgGetProfileString(gszPropSetSec, NULL, gszPropIniFile);

   if (pszNames == NULL) {
      sprintf(gszMsgBox, "Cannot find any property sets.  Nothing to restore.");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return FALSE;
   }
   for (pszName=pszNames; *pszName != '\0'; pszName++, num_rows++) {
      pszName += strlen(pszName);
   }
   if (num_rows == 0) {
      tgWriteProfileString(gszPropSetSec, NULL, NULL, gszPropIniFile);
      tgWriteProfileString(NULL, NULL, NULL, gszPropIniFile);
      sprintf(gszMsgBox, "Cannot find any property sets.  Nothing to restore.");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return FALSE;
   }
   *ppsz_entries = (char**)malloc(num_rows*sizeof(char*));
   if (*ppsz_entries == NULL) FailAllocMessage();
   memset(*ppsz_entries, 0, num_rows*sizeof(char*));

   for (pszName=pszNames, ppsz=(*ppsz_entries); *pszName != '\0'; pszName++) {
      *ppsz = UtilStrDup(pszName);
      if ((*ppsz) == NULL) FailAllocMessage();
      ppsz++;

      pszName += strlen(pszName);
   }
   tgFreeProfileString(pszNames);
   *pn_num_entries = num_rows;
   return TRUE;
}

static
int FreePropSetNames(pp_dsp_ptr, ppsz_entries, pn_num_entries, pn_marked_index,
      pp_check_array, cur_buf, btn_id, selected_index, p_void)
   DspList **pp_dsp_ptr;
   char ***ppsz_entries, *cur_buf;
   int *pn_num_entries, *pn_marked_index, btn_id, selected_index;
   struct CheckArrayRec **pp_check_array;
   void *p_void;
{
   int nReturn=TRUE;

   if (btn_id == BUTTON_OK) {
      nReturn = FALSE;
   }
   if (*ppsz_entries != NULL) {
      char **ppsz=(*ppsz_entries);
      int i=0;

      for (i=0; i < *pn_num_entries; i++) {
         UtilFree(*ppsz);
         ppsz++;
      }
      free(*ppsz_entries);
      *ppsz_entries = NULL;
   }
   return nReturn;
}

static
int SelectPropSetForRestore(pszTitle, pszPropSet, nPropSetSize)
   char *pszTitle, *pszPropSet;
   int nPropSetSize;
{
   char win_name[128];

   ResetNamesInfo();

   NamesSetTitle(pszTitle);
   NamesAddButton("OK", BUTTON_OK);
   NamesAddButton("CANCEL", BUTTON_CANCEL);
   NamesSetDefaultBtnId(BUTTON_OK, BUTTON_OK);
   NamesSetEntries(NULL, NULL, 0, NULL, TRUE, INVALID, 0);
   NamesSetStyle(NAMES_SIMPLE_SELECT_NAME, NAMES_LOOP_MANY);
   NamesSetCallback((GetEntriesFunc*)GetPropSetNames,
         (AfterLoopFunc*)FreePropSetNames, NULL);
   sprintf(win_name, "%s - Restore Properties", TOOL_NAME);

   Names(win_name, NULL, pszPropSet, nPropSetSize, pszPropSet);
   return (*pszPropSet != '\0');
}

void RestoreProperties()
{
   struct PropertiesRec properties;
   long lMask=0L, lSkip=0L;
   struct CheckArrayRec check_array;
   char szPropSet[MAXSTRING], szPropSetSec[MAXSTRING];

   memset(&check_array, 0, sizeof(struct CheckArrayRec));
   memset(&properties, 0, sizeof(struct PropertiesRec));
   properties.userdata = (void*)RESTORE_PROP;
   *szPropSet = '\0';
   if (!SelectPropSetForRestore(
         "Please select a property set to restore ...",
         szPropSet, sizeof(szPropSet))) { 
      return;
   }
   sprintf(szPropSetSec, "%s: %s", gszPropProfilePrefix, szPropSet);
   if (!PrepareToPasteProperties(szPropSetSec, &lMask, &check_array,
         &properties)) {
      return;
   }
   lSkip = (~lMask);

   SelectProperties("Please uncheck properties you don't want to restore ...",
         RESTORE_PROP, &check_array, &properties);
   CleanUpCheckArray(&check_array);
}

void BackupCopiedProperties()
{
   char *pszKeys=tgGetProfileString(gszCopyPasteSec, NULL, gszPropIniFile);

   tgWriteProfileString(gszCopyPasteBackupSec, NULL, NULL, gszPropIniFile);
   if (pszKeys != NULL) {
      char *pszKey=NULL;

      for (pszKey=pszKeys; *pszKey != '\0'; pszKey++) {
         char *pszValue=tgGetProfileString(gszCopyPasteSec, pszKey,
               gszPropIniFile);

         if (pszValue != NULL) {
            tgWriteProfileString(gszCopyPasteBackupSec, pszKey, pszValue,
                  gszPropIniFile);
            tgFreeProfileString(pszValue);
         }
         pszKey += strlen(pszKey);
      }
      tgFreeProfileString(pszKeys);
   }
   tgWriteProfileString(NULL, NULL, NULL, gszPropIniFile);
}

void RestoreCopiedProperties()
{
   char *pszKeys=tgGetProfileString(gszCopyPasteBackupSec, NULL,
         gszPropIniFile);

   tgWriteProfileString(gszCopyPasteSec, NULL, NULL, gszPropIniFile);
   if (pszKeys != NULL) {
      char *pszKey=NULL;

      for (pszKey=pszKeys; *pszKey != '\0'; pszKey++) {
         char *pszValue=tgGetProfileString(gszCopyPasteBackupSec, pszKey,
               gszPropIniFile);

         if (pszValue != NULL) {
            tgWriteProfileString(gszCopyPasteSec, pszKey, pszValue,
                  gszPropIniFile);
            tgFreeProfileString(pszValue);
         }
         pszKey += strlen(pszKey);
      }
      tgFreeProfileString(pszKeys);
   }
   tgWriteProfileString(NULL, NULL, NULL, gszPropIniFile);
}

