/*
 * Author:      William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990-1999, William Chia-Wei Cheng.
 *
 * Permission limited to the use, copy, display, distribute without
 * charging for a fee, and produce derivative works of "tgif" and
 * its documentation for not-for-profit purpose is hereby granted by
 * the Author, provided that the above copyright notice appears in
 * all copies made of "tgif" and that both the copyright notice
 * and this permission notice appear in supporting documentation,
 * and that the name of the Author not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission.  The Author makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.  All other rights (including, but not limited to, the
 * right to sell "tgif", the right to sell or distribute derivative
 * works of "tgif", the right to distribute "tgif" for a fee, and
 * the right to include "tgif" or derivative works of "tgif" in a
 * for-sale product) are reserved by the Author.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
 * OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * @(#)$Header: /mm/src/tgif/v4/RCS/grid.c,v 4.6 1999/02/05 22:53:08 william Exp $
 */

#define _INCLUDE_FROM_GRID_C_

#include "tgifdefs.h"
#include "cmdids.h"
#include "expfdefs.h"

#include "choice.e"
#include "color.e"
#include "cursor.e"
#include "dialog.e"
#include "drawing.e"
#include "dup.e"
#include "file.e"
#include "grid.e"
#include "mainmenu.e"
#include "menu.e"
#include "menuinfo.e"
#include "msg.e"
#include "navigate.e"
#include "obj.e"
#include "page.e"
#include "pattern.e"
#include "raster.e"
#include "ruler.e"
#include "scroll.e"
#include "select.e"
#include "setup.e"
#include "stretch.e"
#include "text.e"
#include "util.e"
#include "xprtfltr.e"

#define ENGLISH_GRID_STEP 8
#define METRIC_GRID_STEP 5

int	gridSystem = ENGLISH_GRID;
int	snapOn = TRUE;
int	xyEnglishGrid = DEFAULT_ENGLISH_GRID;
int	xyMetricGrid = DEFAULT_METRIC_GRID;
int	pageStyle = PORTRAIT;
int	whereToPrint = LATEX_FIG;
int	moveMode = UNCONST_MOVE;
int	gridShown = TRUE;
int	mapShown = TRUE;
int	usePaperSizeStoredInFile = FALSE;
int	oneMotionSelectMove = FALSE;
int	queryZoomInPoint=FALSE;

int shiftForDiagMouseMove=TRUE;
#ifdef RESTRICTED_MOUSE_MOVE
int useRecentForDiagMouseMove=TRUE;
#else /* ~RESTRICTED_MOUSE_MOVE */
int useRecentForDiagMouseMove=FALSE;
#endif /* RESTRICTED_MOUSE_MOVE */

static int zoomInDontRedraw=FALSE;

void MyHDotLine(Win, Y, XStart, XEnd)
   Window Win;
   int Y, XStart, XEnd;
{
   int x, step=0;

   switch (gridSystem) {
   case ENGLISH_GRID: step = ENGLISH_GRID_STEP; break;
   case METRIC_GRID: step = METRIC_GRID_STEP; break;
   }

   for (x=XStart; x < XEnd; x += step) {
      PUT_A_POINT(mainDisplay, Win, defaultGC, x, Y);
   }
}

void MyVDotLine(Win, X, YStart, YEnd)
   Window Win;
   int X, YStart, YEnd;
{
   int y, step=0;

   switch (gridSystem) {
   case ENGLISH_GRID: step = ENGLISH_GRID_STEP; break;
   case METRIC_GRID: step = METRIC_GRID_STEP; break;
   }

   for (y=YStart; y < YEnd; y += step) {
      PUT_A_POINT(mainDisplay, Win, defaultGC, X, y);
   }
}

#define FAKE_CM 80

static
void DoDrawGridLines(Win, LtX, LtY, W, H, WithRestriction)
   Window Win;
   int  LtX, LtY, W, H; /* screen offsets */
   int WithRestriction;
{
   int i, inc=0, abs_grid=0, step=ENGLISH_GRID_STEP;
   int x_start, y_start, x_end, y_end, x_grid_start, y_grid_start;

   if (!gridShown || (inSlideShow && !visibleGridInSlideShow)) return;

   switch (gridSystem) {
   case ENGLISH_GRID:
      inc = HALF_INCH;
      step = ENGLISH_GRID_STEP;
      break;
   case METRIC_GRID:
      if (zoomedIn && zoomScale > 1) {
         inc = (FAKE_CM>>1);
      } else {
         inc = ONE_CM;
      }
      step = METRIC_GRID_STEP;
      break;
   }
   abs_grid = ABS_SIZE(inc);

   if (drawOrigX % abs_grid == 0) {
      x_start = 0;
   } else {
      x_start = OFFSET_X(((int)(drawOrigX / abs_grid) + 1) * abs_grid);
   }
   if (drawOrigY % abs_grid == 0) {
      y_start = 0;
   } else {
      y_start = OFFSET_Y(((int)(drawOrigY / abs_grid) + 1) * abs_grid);
   }
   x_end = min(OFFSET_X(paperWidth), OFFSET_X(drawOrigX+drawWinW));
   y_end = min(OFFSET_Y(paperHeight), OFFSET_Y(drawOrigY+drawWinH));

   if (ZOOMED_SIZE(drawOrigX) % step == 0) {
      x_grid_start = 0;
   } else {
      x_grid_start = ((int)(ZOOMED_SIZE(drawOrigX) / step)) * step -
            ZOOMED_SIZE(drawOrigX);
   }
   if (ZOOMED_SIZE(drawOrigY) % step == 0) {
      y_grid_start = 0;
   } else {
      y_grid_start = ((int)(ZOOMED_SIZE(drawOrigY) / step)) * step -
            ZOOMED_SIZE(drawOrigY);
   }
   if (WithRestriction) {
      int rbx=LtX+W+1, rby=LtY+H+1;

      while (x_start+inc+1 < LtX) x_start += inc;
      while (y_start+inc+1 < LtY) y_start += inc;
      while (x_end-inc > rbx) x_end -= inc;
      while (y_end-inc > rby) y_end -= inc;
   }
   for (i=x_start; i < x_end; i += inc) {
      MyVDotLine(Win, i, y_grid_start, y_end);
   }
   for (i=y_start; i < y_end; i += inc) {
      MyHDotLine(Win, i, x_grid_start, x_end);
   }
}

void RedrawGridLines(Win)
   Window Win;
{
   DoDrawGridLines(Win, 0, 0, 0, 0, FALSE);
}

void DrawGridLines(Win, LtX, LtY, W, H)
   Window Win;
   int  LtX, LtY, W, H; /* screen offsets */
{
   DoDrawGridLines(Win, LtX, LtY, W, H, TRUE);
}

void RedrawPageLines(Win)
   Window Win;
{
   register int i, x, y;
   int end;

   if (pageLayoutMode == PAGE_STACK || !pageLineShownInTileMode) return;

   end = OFFSET_Y(min(drawOrigY+drawWinH,paperHeight));
   for (i=0; i < paperWidth; i += onePageWidth) {
      x = OFFSET_X(i);
      if (i >= drawOrigX && i < drawOrigX+drawWinW) {
         XDrawLine(mainDisplay, Win, defaultGC, x, 0, x, end);
      }
      if (i > drawOrigX+drawWinW) {
         break;
      }
   }
   end = OFFSET_X(min(drawOrigX+drawWinW,paperWidth));
   for (i=0; i < paperHeight; i += onePageHeight) {
      y = OFFSET_Y(i);
      if (i >= drawOrigY && i < drawOrigY+drawWinH) {
         XDrawLine(mainDisplay, Win, defaultGC, 0, y, end, y);
      }
      if (i > drawOrigY+drawWinH) {
         break;
      }
   }
}

void DrawPageLines(Win, LtX, LtY, W, H)
   Window Win;
   int  LtX, LtY, W, H; /* screen offsets */
{
   int i, x, y, start, end;

   if (pageLayoutMode == PAGE_STACK || !pageLineShownInTileMode) return;

   start = max(0,LtY);
   end = min(LtY+H,OFFSET_Y(min(drawOrigY+drawWinH,paperHeight)));
   for (i=0; i < paperWidth; i += onePageWidth) {
      x = OFFSET_X(i);
      if (x >= LtX && x < LtX+W) {
         XDrawLine(mainDisplay, Win, defaultGC, x, start, x, end);
      }
      if (i > drawOrigX+drawWinW) {
         break;
      }
   }
   start = max(0,LtX);
   end = min(LtX+W,OFFSET_X(min(drawOrigX+drawWinW,paperWidth)));
   for (i=0; i < paperHeight; i += onePageHeight) {
      y = OFFSET_Y(i);
      if (y >= LtY && y < LtY+H) {
         XDrawLine(mainDisplay, Win, defaultGC, start, y, end, y);
      }
      if (i > drawOrigY+drawWinH) {
         break;
      }
   }
}

void ToggleGridSystem()
{
   int saved_orig_x, saved_orig_y;

   switch (gridSystem) {
   case ENGLISH_GRID: Msg("Using Metric system."); break;
   case METRIC_GRID: Msg("Using English system."); break;
   }
   gridSystem = !gridSystem;
   if (drawOrigX != 0 || drawOrigY != 0) {
      saved_orig_x = drawOrigX;
      saved_orig_y = drawOrigY;
      drawOrigX = drawOrigY = 0;
      RedrawScrollBars();
      UpdDrawWinBBox();
      AdjSplineVs();
      AdjustCurText(-(saved_orig_x), -(saved_orig_y));
   }
   ClearAndRedrawDrawWindow();
   RedrawRulers();
   UpdatePinnedMenu(MENU_LAYOUT);
}

void IncGrid()
{
   if (!snapOn) {
      Msg("Snap is not on, grid size not changed.");
   } else if (gridSystem == ENGLISH_GRID && xyEnglishGrid < HALF_INCH) {
      xyEnglishGrid <<= 1;
      RedrawRulers();
   } else if (gridSystem == METRIC_GRID && xyMetricGrid < ONE_CM) {
      switch (xyMetricGrid) {
      case ONE_MM: xyMetricGrid = TWO_MM; break;
      case TWO_MM: xyMetricGrid = FIVE_MM; break;
      case FIVE_MM: xyMetricGrid = ONE_CM; break;
      }
      RedrawRulers();
   } else {
      Msg("Already at maximun grid, grid size not changed.");
   }
}

void DecGrid()
{
   if (!snapOn) {
      Msg("Snap is not on, grid size not changed.");
   } else if (gridSystem == ENGLISH_GRID && xyEnglishGrid > 4) {
      xyEnglishGrid >>= 1;
      RedrawRulers();
   } else if (gridSystem == METRIC_GRID && xyMetricGrid > ONE_MM) {
      switch (xyMetricGrid) {
      case TWO_MM: xyMetricGrid = ONE_MM; break;
      case FIVE_MM: xyMetricGrid = TWO_MM; break;
      case ONE_CM: xyMetricGrid = FIVE_MM; break;
      }
      RedrawRulers();
   } else {
      Msg("Already at minimum grid, grid size not changed.");
   }
}

void ToggleGridShown()
{
   gridShown = !gridShown;
   ClearAndRedrawDrawWindow();
   UpdatePinnedMenu(MENU_LAYOUT);
}

void ToggleSnapOn()
{
   snapOn = !snapOn;
   if (snapOn) {
      Msg("Snapping to grid point activated.");
   } else {
      Msg("Snapping to grid point disabled.");
   }
   RedrawRulers();
   UpdatePinnedMenu(MENU_LAYOUT);
}

void ToggleColorPostScript()
{
   colorDump = !colorDump;
   if (colorDump) {
      Msg("Will print in color (PS and pixmap).");
   } else {
      Msg("Will print in black-and-white (PS and bitmap).");
   }
   ShowWhereToPrint();
   SetFileModified(TRUE);
   UpdatePinnedMenu(MENU_LAYOUT);
   UpdatePinnedMenu(MENU_FILE);
}

void ToggleMoveMode()
{
   moveMode = !moveMode;
   switch (moveMode) {
   case CONST_MOVE: Msg("Constrained move."); break;
   case UNCONST_MOVE: Msg("Unconstrained move."); break;
   }
   ShowMoveMode();
   UpdatePinnedMenu(MENU_LAYOUT);
   UpdatePinnedMenu(MENU_MOVEMODE);
}

void ToggleMapShown()
{
   mapShown = !mapShown;
   if (mapShown) {
      Msg("Will display bitmap/pixmap.");
   } else {
      Msg("Will not display bitmap/pixmap.");
   }
   ClearAndRedrawDrawWindow();
   UpdatePinnedMenu(MENU_LAYOUT);
}

void ToggleUseGray()
{
   useGray = !useGray;
   if (useGray) {
      Msg("Gray scale enabled in black&white printing.");
   } else {
      Msg("Gray scale disabled in black&white printing.");
   }
   UpdatePinnedMenu(MENU_LAYOUT);
}

void SetMeasureUnit()
{
   char buf[MAXSTRING+1], spec[MAXSTRING+1];

   GetUnitSpec(gszMsgBox);
   sprintf(buf, "( current spec: '%s' )", gszMsgBox);
   sprintf(gszMsgBox, "Please enter '%s' (%s):",
         "<num> {pixel|cm|in}/<unit>", "e.g., 36 in/yd or 0.1 in/mil");
   Dialog(gszMsgBox, buf, spec);
   UtilTrimBlanks(spec);
   if (*spec == '\0') return;
   if (SetUnit(spec)) {
      RedrawRulers();
      ShowUnitMsg();
      SetFileModified(TRUE);
   }
}

void ToggleShowMenubar()
{
   noMenubar = !noMenubar;
   if (noMenubar) {
      XUnmapWindow(mainDisplay, menubarWindow);
   } else {
      XMapWindow(mainDisplay, menubarWindow);
   }
   Reconfigure(TRUE);
   UpdatePinnedMenu(MENU_LAYOUT);
}

void ToggleShowStatus()
{
   noStatusWindow = !noStatusWindow;
   if (noStatusWindow) {
      XUnmapWindow(mainDisplay, userRedrawWindow);
      XUnmapWindow(mainDisplay, statusWindow);
   } else {
      XMapWindow(mainDisplay, userRedrawWindow);
      XMapWindow(mainDisplay, statusWindow);
   }
   Reconfigure(TRUE);
   UpdatePinnedMenu(MENU_LAYOUT);
}

void ToggleShowMode()
{
   noModeWindow = !noModeWindow;
   if (noModeWindow) {
      XUnmapWindow(mainDisplay, modeWindow);
   } else {
      XMapWindow(mainDisplay, modeWindow);
   }
   Reconfigure(TRUE);
   UpdatePinnedMenu(MENU_LAYOUT);
}

static
void WhereToPrintMsg()
{
   if (whereToPrint >= MAXDEFWHERETOPRINT) {
      sprintf(gszMsgBox, "Will export %s file.",
            GetExportName(whereToPrint, EXPORT_THIS));
      Msg(gszMsgBox);
   } else {
      switch (whereToPrint) {
      case PRINTER: Msg("Print device set to printer."); break;
      case LATEX_FIG:
         Msg("Will export Encapsulated PostScript (LaTeX-Figure) file.");
         break;
      case PS_FILE: Msg("Will export raw PostScript file."); break;
      case XBM_FILE:
         if (colorDump) {
            Msg("Will export X11 Pixmap file.");
         } else {
            Msg("Will export X11 Bitmap file.");
         }
         break;
      case TEXT_FILE: Msg("Will export ASCII text file."); break;
      case EPSI_FILE:
         Msg("Will export EPSI file (EPS with preview bitmap).");
         break;
      case GIF_FILE:
         Msg("Will export GIF/ISMAP (needs xpm->gif filters) file.");
         break;
      case HTML_FILE:
         Msg("Will export HTML (with Client-side imagemap) file.");
         break;
      case PDF_FILE:
         Msg("Will export PDF (needs ps->pdf filter) file.");
         break;
      case TIFFEPSI_FILE:
         Msg("Will export DOS/Windows EPS/EPSI file.");
         break;
      }
   }
}

void SetWhereToPrint(nIndex)
   int nIndex;
{
   whereToPrint = nIndex;
   WhereToPrintMsg();
   ShowWhereToPrint();
   UpdatePinnedMenu(MENU_LAYOUT);
}

void ToggleWhereToPrint()
{
   int nIndex=0;
   char szLeft[MAXSTRING], szRight[MAXSTRING];

   whereToPrint = (whereToPrint+1) % maxWhereToPrint;
   WhereToPrintMsg();
   ShowWhereToPrint();
   /* UpdatePinnedMenu(MENU_LAYOUT); -- no need since we have ExportFormat popup */

   nIndex = GetExportIndex(whereToPrint, EXPORT_NEXT);
   if (nIndex == PRINTER) {
      sprintf(szLeft, "%s print mode",
            GetExportName(whereToPrint, EXPORT_NEXT));
   } else {
      sprintf(szLeft, "%s export mode",
            GetExportName(whereToPrint, EXPORT_NEXT));
   }
   nIndex = GetExportIndex(whereToPrint, EXPORT_PREV);
   if (nIndex == PRINTER) {
      sprintf(szRight, "%s print mode",
            GetExportName(whereToPrint, EXPORT_PREV));
   } else {
      sprintf(szRight, "%s export mode",
            GetExportName(whereToPrint, EXPORT_PREV));
   }
   SetMouseStatus(szLeft, "Layout Menu", szRight);
}

void ToggleOneMotionSelectMove()
{
   oneMotionSelectMove = !oneMotionSelectMove;
   if (oneMotionSelectMove) {
      Msg("Click-select-move in one motion mode selected.");
   } else {
      Msg("Click-select-click-move mode selected.");
   }
   UpdatePinnedMenu(MENU_LAYOUT);
}

void ToggleColorLayers()
{
   colorLayers = !colorLayers;
   if (colorLayers) {
      XMapWindow(mainDisplay, colorWindow);
      XMapWindow(mainDisplay, colorDummyWindow);
      Msg("Color layers enabled.");
   } else {
      XUnmapWindow(mainDisplay, colorWindow);
      XUnmapWindow(mainDisplay, colorDummyWindow);
      Msg("Color layers disabled.");
   }
   Reconfigure(TRUE);
   UpdatePinnedMenu(MENU_LAYOUT);
   ClearAndRedrawDrawWindow();
}

void ToggleStretchableText()
{
   stretchableText = !stretchableText;
   if (stretchableText) {
      Msg("Text objects are stretchable.");
   } else {
      Msg("Text objects are not stretchable.");
   }
   ShowStretchableTextMode();
   UpdatePinnedMenu(MENU_STRETCHTEXT);
}

void ToggleTransPatMode()
{
   transPat = !transPat;
   if (transPat) {
      Msg("Fill and pen patterns are transparent.");
   } else {
      Msg("Fill and pen patterns are opaque.");
   }
   ShowTransPatMode();
   UpdatePinnedMenu(MENU_TRANSPAT);
}

static
void PostZoom(xc, yc)
   int xc, yc;
{
   int new_orig_x=0, new_orig_y=0;

   UpdDrawWinWH();

   new_orig_x = ((xc<<1)-drawWinW)>>1;
   new_orig_y = ((yc<<1)-drawWinH)>>1;

   drawOrigX = max(0,new_orig_x);
   drawOrigY = max(0,new_orig_y);

   if (drawOrigX>=0 && drawWinW<paperWidth && drawOrigX+drawWinW>=paperWidth) {
      switch (gridSystem) {
      case ENGLISH_GRID:
         if ((paperWidth-drawWinW) % ABS_SIZE(HALF_INCH) == 0) {
            drawOrigX = paperWidth-drawWinW;
         } else {
            drawOrigX = max(0, ((int)((paperWidth-drawWinW)/
                  ABS_SIZE(HALF_INCH)) + 1) * ABS_SIZE(HALF_INCH));
         }
         break;
      case METRIC_GRID:
         if (zoomedIn && zoomScale > 1) {
            if ((paperWidth-drawWinW) % ABS_SIZE(FAKE_CM) == 0) {
               drawOrigX = paperWidth-drawWinW;
            } else {
               drawOrigX = max(0, ((int)((paperWidth-drawWinW)/
                     ABS_SIZE(FAKE_CM)) + 1) * ABS_SIZE(FAKE_CM));
            }
         } else {
            if ((paperWidth-drawWinW) % ABS_SIZE(ONE_CM) == 0) {
               drawOrigX = paperWidth-drawWinW;
            } else {
               drawOrigX = max(0, ((int)((paperWidth-drawWinW)/
                     ABS_SIZE(ONE_CM)) + 1) * ABS_SIZE(ONE_CM));
            }
         }
         break;
      }
   } else if (drawOrigX < 0 || drawWinW >= paperWidth) {
      drawOrigX = 0;
   }
   if (drawOrigY>=0 && drawWinH<paperHeight &&
         drawOrigY+drawWinH>=paperHeight) {
      switch (gridSystem) {
      case ENGLISH_GRID:
         if ((paperHeight-drawWinH) % ABS_SIZE(HALF_INCH) == 0) {
            drawOrigY = paperHeight-drawWinH;
         } else {
            drawOrigY = max(0, ((int)((paperHeight-drawWinH)/
                  ABS_SIZE(HALF_INCH)) + 1) * ABS_SIZE(HALF_INCH));
         }
         break;
      case METRIC_GRID:
         if (zoomedIn && zoomScale > 1) {
            if ((paperHeight-drawWinH) % ABS_SIZE(FAKE_CM) == 0) {
               drawOrigY = paperHeight-drawWinH;
            } else {
               drawOrigY = max(0, ((int)((paperHeight-drawWinH)/
                     ABS_SIZE(FAKE_CM)) + 1) * ABS_SIZE(FAKE_CM));
            }
         } else {
            if ((paperHeight-drawWinH) % ABS_SIZE(ONE_CM) == 0) {
               drawOrigY = paperHeight-drawWinH;
            } else {
               drawOrigY = max(0, ((int)((paperHeight-drawWinH)/
                     ABS_SIZE(ONE_CM)) + 1) * ABS_SIZE(ONE_CM));
            }
         }
         break;
      }
   } else if (drawOrigY < 0 || drawWinH >= paperHeight) {
      drawOrigY = 0;
   }
   drawOrigX = ABS_SIZE(ZOOMED_SIZE(drawOrigX));
   drawOrigY = ABS_SIZE(ZOOMED_SIZE(drawOrigY));

   UpdDrawWinBBox();
   if (!zoomInDontRedraw) {
      SetWatchCursor(drawWindow);
      SetWatchCursor(mainWindow);
      RedrawRulers();
      RedrawScrollBars();
      AdjSplineVs();
      AdjCaches();
      ShowZoom();
      /* use to be ClearAndRedrawDrawWindowDontDrawCurText() */
      ClearAndRedrawDrawWindowNoCurT();
      SetDefaultCursor(mainWindow);
      ShowCursor();
   }
   justDupped = FALSE;
}

void DefaultZoom()
{
   int xc, yc;

   if (zoomScale == 0) {
      Msg("Already at the default zoom.");
      return;
   }
   if (curChoice==DRAWTEXT && textCursorShown) {
      PrepareZoomCurText(&xc, &yc);
   } else {
      xc = (topSel==NULL) ? ((drawOrigX<<1)+drawWinW)>>1 : (selLtX+selRbX)>>1;
      yc = (topSel==NULL) ? ((drawOrigY<<1)+drawWinH)>>1 : (selLtY+selRbY)>>1;
   }
   if (curChoice == VERTEXMODE) {
      TieLooseEnds();
      SetCurChoice(NOTHING);
   }
   zoomScale = 0;
   zoomedIn = FALSE;

   if (curChoice==DRAWTEXT && textCursorShown) PreZoomCurText();
   PostZoom(xc, yc);
   if (curChoice==DRAWTEXT && textCursorShown) PostZoomCurText(xc, yc);
}

static int zoomInAbsX=0, zoomInAbsY=0;
static int useZoomInAtCursor=FALSE;

static
void FixUpZoomCenter(p_abs_cx, p_abs_cy, zooming_in)
   int *p_abs_cx, *p_abs_cy, zooming_in;
{
   int saved_zoomed_in=zoomedIn, saved_zoom_scale=zoomScale;
   int anchor_x_off=OFFSET_X(*p_abs_cx), anchor_y_off=OFFSET_Y(*p_abs_cy);
   int new_anchor_abs_x=0, new_anchor_abs_y=0;
   int abs_dx=0, abs_dy=0, new_abs_cx=0, new_abs_cy=0;

   if (zooming_in) {
      if (zoomedIn) {
         zoomScale++;
      } else if (zoomScale == 0) {
         zoomedIn = TRUE;
         zoomScale++;
      } else {
         zoomScale--;
      }
   } else {
      if (!zoomedIn) {
         zoomScale++;
      } else if (zoomScale == 1) {
         zoomedIn = FALSE;
         zoomScale--;
      } else {
         zoomScale--;
      }
   }
   UpdDrawWinWH();

   new_anchor_abs_x = ABS_X(anchor_x_off);
   new_anchor_abs_y = ABS_Y(anchor_y_off);

   new_abs_cx = (((drawOrigX<<1)+drawWinW)>>1);
   new_abs_cy = (((drawOrigY<<1)+drawWinH)>>1);

   abs_dx = new_abs_cx - new_anchor_abs_x;
   abs_dy = new_abs_cy - new_anchor_abs_y;

   if (zooming_in) {
      *p_abs_cx += (abs_dx);
      *p_abs_cy += (abs_dy);
   } else {
      *p_abs_cx += (abs_dx);
      *p_abs_cy += (abs_dy);
   }
   zoomedIn = saved_zoomed_in;
   zoomScale = saved_zoom_scale;
   UpdDrawWinWH();
}

void ZoomIn()
{
   int xc, yc;

   if (zoomedIn && zoomScale == MAX_ZOOMED_IN) {
      MsgBox("Already at highest magnification, can no longer zoom in.",
            TOOL_NAME, INFO_MB);
      return;
   }

   if (curChoice==DRAWTEXT && textCursorShown) {
      if (useZoomInAtCursor) {
         sprintf(gszMsgBox, "    (%s %s)",
               "the cursor position is ignored even",
               "though <Cntrl><Shift>Btn1 is used");
         TwoLineMsg("Zooming in around text being edited...", gszMsgBox);
      }
      PrepareZoomCurText(&xc, &yc);
   } else if (useZoomInAtCursor) {
      xc = zoomInAbsX;
      yc = zoomInAbsY;
      FixUpZoomCenter(&xc, &yc, TRUE);
   } else if (queryZoomInPoint==TRUE || (queryZoomInPoint==INVALID &&
         topSel==NULL)) {
      if (queryZoomInPoint==INVALID && topSel==NULL) {
         MakeQuiescent();
      }
      SaveStatusStrings();
      SetMouseStatus("Select zoom center", "Cancel", "Cancel");
      if (DrawWindowLoop(&xc, &yc, magCursor, FALSE) != Button1) {
         RestoreStatusStrings();
         Msg("ZoomIn canceled by the user.");
         return;
      }
      RestoreStatusStrings();
      xc = ABS_X(xc);
      yc = ABS_Y(yc);
      FixUpZoomCenter(&xc, &yc, TRUE);
   } else if (queryZoomInPoint==BAD) {
      Window root_win, child_win;
      int root_x, root_y;
      unsigned int status;

      XQueryPointer(mainDisplay, drawWindow, &root_win, &child_win,
            &root_x, &root_y, &xc, &yc, &status);
      xc = ABS_X(xc);
      yc = ABS_Y(yc);
      FixUpZoomCenter(&xc, &yc, TRUE);
   } else {
      xc = (topSel==NULL) ? ((drawOrigX<<1)+drawWinW)>>1 : (selLtX+selRbX)>>1;
      yc = (topSel==NULL) ? ((drawOrigY<<1)+drawWinH)>>1 : (selLtY+selRbY)>>1;
   }
   if (curChoice == VERTEXMODE) {
      TieLooseEnds();
      SetCurChoice(NOTHING);
   }
   if (zoomedIn) {
      zoomScale++;
   } else if (zoomScale == 0) {
      zoomedIn = TRUE;
      zoomScale++;
   } else {
      zoomScale--;
   }
   if (curChoice==DRAWTEXT && textCursorShown) PreZoomCurText();
   PostZoom(xc, yc);
   if (curChoice==DRAWTEXT && textCursorShown) PostZoomCurText(xc, yc);
}

void ZoomInAtCursor(abs_x, abs_y)
   int abs_x, abs_y;
{
   zoomInAbsX = abs_x;
   zoomInAbsY = abs_y;
   useZoomInAtCursor = TRUE;
   ZoomIn();
   useZoomInAtCursor = FALSE;
}

void ZoomWayOut()
{
   int saved_zoom_scale=zoomScale, saved_zoomed_in=zoomedIn;

   while ((drawWinW>>1) >= paperWidth && (drawWinH>>1) >= paperHeight) {
      if (zoomedIn) {
         zoomScale++;
      } else if (zoomScale == 0) {
         zoomedIn = TRUE;
         zoomScale++;
      } else {
         zoomScale--;
      }
      drawWinW >>= 1;
      drawWinH >>= 1;
   }
   while (drawWinW < paperWidth || drawWinH < paperHeight) {
      if (!zoomedIn) {
         zoomScale++;
      } else if (zoomScale == 1) {
         zoomedIn = FALSE;
         zoomScale--;
      } else {
         zoomScale--;
      }
      drawWinW <<= 1;
      drawWinH <<= 1;
   }
   if (saved_zoom_scale==zoomScale && saved_zoomed_in==zoomedIn &&
         drawOrigX==0 && drawOrigY==0) {
      return;
   }
   TieLooseEnds();
   SetCurChoice(NOTHING);

   PostZoom(0, 0);
}

void ZoomOut()
{
   int xc, yc;

   if (paperWidth <= drawWinW && paperHeight <= drawWinH) {
      strcpy(gszMsgBox, "Already at paper boundaries, can no longer zoom out.");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return;
   }
   if (curChoice==DRAWTEXT && textCursorShown) {
      PrepareZoomCurText(&xc, &yc);
   } else if (queryZoomInPoint==BAD) {
      Window root_win, child_win;
      int root_x, root_y;
      unsigned int status;

      XQueryPointer(mainDisplay, drawWindow, &root_win, &child_win,
            &root_x, &root_y, &xc, &yc, &status);
      xc = ABS_X(xc);
      yc = ABS_Y(yc);
      FixUpZoomCenter(&xc, &yc, FALSE);
   } else {
      xc = (topSel==NULL) ? ((drawOrigX<<1)+drawWinW)>>1 : (selLtX+selRbX)>>1;
      yc = (topSel==NULL) ? ((drawOrigY<<1)+drawWinH)>>1 : (selLtY+selRbY)>>1;
   }
   if (curChoice == VERTEXMODE) {
      TieLooseEnds();
      SetCurChoice(NOTHING);
   }
   if (!zoomedIn) {
      zoomScale++;
   } else if (zoomScale == 1) {
      zoomedIn = FALSE;
      zoomScale--;
   } else {
      zoomScale--;
   }
   if (curChoice==DRAWTEXT && textCursorShown) PreZoomCurText();
   PostZoom(xc, yc);
   if (curChoice==DRAWTEXT && textCursorShown) PostZoomCurText(xc, yc);
}

void CenterAtCursor(abs_x, abs_y)
   int abs_x, abs_y;
{
   zoomInDontRedraw = TRUE;
   ZoomInAtCursor(abs_x, abs_y);
   zoomInDontRedraw = FALSE;
   ZoomOut();
}

void PreciseZoom(zoomed_in, zoom_scale, force_zoom)
   int zoomed_in, zoom_scale, force_zoom;
{
   int xc, yc;

   if (!force_zoom && zoomedIn == zoomed_in && zoomScale == zoom_scale) return;

   if (curChoice==DRAWTEXT && textCursorShown) {
      PrepareZoomCurText(&xc, &yc);
   } else {
      xc = (topSel==NULL) ? ((drawOrigX<<1)+drawWinW)>>1 : (selLtX+selRbX)>>1;
      yc = (topSel==NULL) ? ((drawOrigY<<1)+drawWinH)>>1 : (selLtY+selRbY)>>1;
   }
   if (curChoice == VERTEXMODE) {
      TieLooseEnds();
      SetCurChoice(NOTHING);
   }
   zoomedIn = zoomed_in;
   zoomScale = zoom_scale;

   if (curChoice==DRAWTEXT && textCursorShown) PreZoomCurText();
   PostZoom(xc, yc);
   if (curChoice==DRAWTEXT && textCursorShown) PostZoomCurText(xc, yc);
}

void SetPSPageWidthHeight()
{

   if (psYOffStr != NULL && psYOff != NULL) {
      float fval=(float)0;

      fval = (float)(((float)onePageHeight)/((float)PIX_PER_INCH));
      FormatFloat(&fval, psYOffStr[0]);
      psYOff[0] = ((float)onePageHeight)/((float)PIX_PER_INCH);
      psPageWidthInInch[0]=((float)onePageWidth)/((float)PIX_PER_INCH);
      psPageWidthInInch[1]=((float)onePageHeight)/((float)PIX_PER_INCH);
      psPageHeightInInch[0]=((float)onePageHeight)/((float)PIX_PER_INCH);
      psPageHeightInInch[1]=((float)onePageWidth)/((float)PIX_PER_INCH);
   }
}

void ResetOnePageSize()
{  /* reset 8.5in x 11in */
#ifndef A4PAPER
   onePageWidth = (85*PIX_PER_INCH)/10;
   onePageHeight = 11*PIX_PER_INCH;
#else /* A4PAPER */
   onePageWidth = (825*PIX_PER_INCH)/100;
   onePageHeight = (117*PIX_PER_INCH)/10;
#endif /* A4PAPER */
   SetPSPageWidthHeight();
}

int UpdPageStyle(NewPageStyle)
   int NewPageStyle;
{
   int changed=FALSE, old_w=paperWidth, old_h=paperHeight;

   pageStyle = NewPageStyle;
   onePageWidth =
         round(psPageWidthInInch[NewPageStyle]*PIX_PER_INCH*100.0/printMag);
   onePageHeight =
         round(psPageHeightInInch[NewPageStyle]*PIX_PER_INCH*100.0/printMag);
   paperWidth = onePageWidth * paperCol;
   paperHeight = onePageHeight * paperRow;
   if (drawOrigX+drawWinW > paperWidth || drawOrigX+drawWinW > old_w) {
      if (drawOrigX+drawWinW > paperWidth) {
         int tmp_x=paperWidth-drawWinW;

         switch (gridSystem) {
         case ENGLISH_GRID:
            if (zoomedIn) {
               if ((ZOOMED_SIZE(tmp_x) % HALF_INCH) == 0) {
                  drawOrigX = max(0,tmp_x);
               } else {
                  drawOrigX = max(0,((int)(ZOOMED_SIZE(tmp_x)/HALF_INCH)) *
                        HALF_INCH);
               }
            } else {
               if ((tmp_x % (ABS_SIZE(HALF_INCH))) == 0) {
                  drawOrigX = max(0,tmp_x);
               } else {
                  drawOrigX = max(0,((int)(tmp_x/ABS_SIZE(HALF_INCH))) *
                        ABS_SIZE(HALF_INCH));
               }
            }
            break;
         case METRIC_GRID:
            if (zoomedIn) {
               if (zoomScale > 1) {
                  if ((ZOOMED_SIZE(tmp_x) % ONE_CM) == 0) {
                     drawOrigX = max(0,tmp_x);
                  } else {
                     drawOrigX = max(0,((int)(ZOOMED_SIZE(tmp_x)/ONE_CM)) *
                           ONE_CM);
                  }
               } else {
                  if ((ZOOMED_SIZE(tmp_x) % FAKE_CM) == 0) {
                     drawOrigX = max(0,tmp_x);
                  } else {
                     drawOrigX = max(0,((int)(ZOOMED_SIZE(tmp_x)/FAKE_CM)) *
                           FAKE_CM);
                  }
               }
            } else {
               if ((tmp_x % (ABS_SIZE(ONE_CM))) == 0) {
                  drawOrigX = max(0,tmp_x);
               } else {
                  drawOrigX = max(0,((int)(tmp_x/ABS_SIZE(ONE_CM))) *
                        ABS_SIZE(ONE_CM));
               }
            }
            break;
         }
      }
      drawOrigX = ABS_SIZE(ZOOMED_SIZE(drawOrigX));
      changed = TRUE;
   }
   if (drawOrigY+drawWinH > paperHeight || drawOrigY+drawWinH > old_h) {
      if (drawOrigY+drawWinH > paperHeight) {
         int tmp_y=paperHeight-drawWinH;

         switch (gridSystem) {
         case ENGLISH_GRID:
            if (zoomedIn) {
               if ((ZOOMED_SIZE(tmp_y) % HALF_INCH) == 0) {
                  drawOrigY = max(0,tmp_y);
               } else {
                  drawOrigY = max(0,((int)(ZOOMED_SIZE(tmp_y)/HALF_INCH)) *
                        HALF_INCH);
               }
            } else {
               if ((tmp_y % (ABS_SIZE(HALF_INCH))) == 0) {
                  drawOrigY = max(0,tmp_y);
               } else {
                  drawOrigY = max(0,((int)(tmp_y/ABS_SIZE(HALF_INCH))) *
                        ABS_SIZE(HALF_INCH));
               }
            }
            break;
         case METRIC_GRID:
            if (zoomedIn) {
               if (zoomScale > 1) {
                  if ((ZOOMED_SIZE(tmp_y) % FAKE_CM) == 0) {
                     drawOrigY = max(0,tmp_y);
                  } else {
                     drawOrigY = max(0,((int)(ZOOMED_SIZE(tmp_y)/FAKE_CM)) *
                           FAKE_CM);
                  }
               } else {
                  if ((ZOOMED_SIZE(tmp_y) % ONE_CM) == 0) {
                     drawOrigY = max(0,tmp_y);
                  } else {
                     drawOrigY = max(0,((int)(ZOOMED_SIZE(tmp_y)/ONE_CM)) *
                           ONE_CM);
                  }
               }
            } else {
               if ((tmp_y % (ABS_SIZE(ONE_CM))) == 0) {
                  drawOrigY = max(0,tmp_y);
               } else {
                  drawOrigY = max(0,((int)(tmp_y/ABS_SIZE(ONE_CM))) *
                        ABS_SIZE(ONE_CM));
               }
            }
            break;
         }
      }
      drawOrigY = ABS_SIZE(ZOOMED_SIZE(drawOrigY));
      changed = TRUE;
   }
   return changed;
}

void ChangePageStyle(NewPageStyle, PageStyleStr)
   int NewPageStyle;
   char *PageStyleStr;
{
   char msg[MAXSTRING];

   if (pageStyle != NewPageStyle) {
      if (UpdPageStyle(NewPageStyle)) {
         UpdDrawWinBBox();
         AdjSplineVs();
         ClearAndRedrawDrawWindow();
      }
      ShowFile();
      UpdatePinnedMenu(MENU_LAYOUT);
      RedrawScrollBars();
      RedrawRulers();
      SetFileModified(TRUE);
      sprintf(msg, "Page style changed to %s.", PageStyleStr);
      Msg(msg);
   }
}

int RefreshPageStyleMenu(menu)
   TgMenu *menu;
{
   int ok=TRUE;

   /* Portrait */
   ok &= TgSetMenuItemRadioById(menu, CMDID_PORTRAITMODE,
         (pageStyle==PORTRAIT));
   /* Landscape */
   ok &= TgSetMenuItemRadioById(menu, CMDID_LANDSCAPEMODE,
         (pageStyle==LANDSCAPE));

   return ok;
}

TgMenu *CreatePageStyleMenu(parent_menu, x, y, menu_info)
   TgMenu *parent_menu;
   int x, y;
   TgMenuInfo *menu_info;
{
   TgMenu *menu=TgCreateMenuFromMenuInfo(parent_menu, x, y, menu_info);

   if (menu != NULL) {
      if (!RefreshPageStyleMenu(menu)) {
         return TgDestroyMenu(menu, TRUE);
      }
   }
   return menu;
}

int PageStyleMenu(X, Y, TrackMenubar)
   int X, Y, TrackMenubar;
{
   int rc=INVALID;
   TgMenu *menu=(pageStyleMenuInfo.create_proc)(NULL, X, Y, &pageStyleMenuInfo);

   activeMenu = INVALID;
   if (menu != NULL) {
      rc = TgMenuLoop(menu);
      TgDestroyMenu(menu, TRUE);
   }
   return rc;
}

void RefreshExportFormatMenu(menu)
   TgMenu *menu;
{
   int i, num_items=menu->num_items;
   TgMenuItem *menuitems=menu->menuitems;

   for (i=0; i < num_items; i++) {
      TgMenuItem *menu_item=(&menuitems[i]);
      TgMenuItem stMenuItem;

      memset(&stMenuItem, 0, sizeof(TgMenuItem));
      stMenuItem.state = TGBS_NORMAL;
      stMenuItem.checked = (i == whereToPrint);
      TgSetMenuItemInfo(menu_item, TGMU_MASK_STATE|TGMU_MASK_CHECK,
            &stMenuItem);
   }
}

static char printerDesc[]  = "Set printing mode to print to a printer";
static char latexFigDesc[] = "Set printing mode to generate EPS files";
static char psFileDesc[]   = "Set printing mode to generate PS files";
static char xbmDesc[]      = "Set printing mode to generate X11 Bitmap files";
static char xpmDesc[]      = "Set printing mode to generate X11 Pixmap files";
static char textFileDesc[] = "Set printing mode to generate ASCII files";
static char epsiFileDesc[] = "Set printing mode to generate EPSI files";
static char gifDesc[]      = "Set printing mode to generate GIF/ISMAP files";
static char htmlDesc[]     = "Set printing mode to generate HTML files";
static char pdfDesc[]      = "Set printing mode to generate PDF files";
static char tiffEpsiDesc[] =
      "Set printing mode to generate DOS/Windows EPSI files";

TgMenu *CreateExportFormatMenu(parent_menu, x, y, menu_info)
   TgMenu *parent_menu;
   int x, y;
   TgMenuInfo *menu_info;
{
   int i=0;
   TgMenu *menu=NULL;
   TgMenuInfo stMenuInfo;
   TgMenuItemInfo *item_info=NULL;

   memcpy(&stMenuInfo, menu_info, sizeof(TgMenuInfo));
   stMenuInfo.items = (TgMenuItemInfo*)malloc(
         (maxWhereToPrint+1)*sizeof(TgMenuItemInfo));
   if (stMenuInfo.items == NULL) FailAllocMessage();
   memset(stMenuInfo.items, 0, (maxWhereToPrint+1)*sizeof(TgMenuItemInfo));
   for (item_info=stMenuInfo.items, i=0; i < maxWhereToPrint;
         item_info++, i++) {
      item_info->menu_str = (char*)(Pixmap*)(&whereToPrintPixmap[i]);
      item_info->shortcut_str = NULL;
      switch (i) {
      case PRINTER: strcpy(gszMsgBox, printerDesc); break;
      case LATEX_FIG: strcpy(gszMsgBox, latexFigDesc); break;
      case PS_FILE: strcpy(gszMsgBox, psFileDesc); break;
      case XBM_FILE: strcpy(gszMsgBox, colorDump ? xpmDesc : xbmDesc); break;
      case TEXT_FILE: strcpy(gszMsgBox, textFileDesc); break;
      case EPSI_FILE: strcpy(gszMsgBox, epsiFileDesc); break;
      case GIF_FILE: strcpy(gszMsgBox, gifDesc); break;
      case HTML_FILE: strcpy(gszMsgBox, htmlDesc); break;
      case PDF_FILE: strcpy(gszMsgBox, pdfDesc); break;
      case TIFFEPSI_FILE: strcpy(gszMsgBox, tiffEpsiDesc); break;
      default:
         sprintf(gszMsgBox, "Set printing mode to generate %s files",
               GetExportName(whereToPrint, EXPORT_THIS));
         break;
      }
      item_info->status_str = UtilStrDup(gszMsgBox);
      if (item_info->status_str == NULL) FailAllocMessage();
      item_info->submenu_info = NULL;
      item_info->cmdid = CMDID_SETWHERETOPRINT;
   }
   stMenuInfo.items[maxWhereToPrint].cmdid = INVALID;

   menu = TgCreateMenuFromMenuInfo(parent_menu, x, y, &stMenuInfo);
   for (item_info=stMenuInfo.items, i=0; i < maxWhereToPrint;
         item_info++, i++) {
      UtilFree(item_info->status_str);
   }
   memset(stMenuInfo.items, 0, (maxWhereToPrint+1)*sizeof(TgMenuItemInfo));
   free(stMenuInfo.items);
   stMenuInfo.items = NULL;
   if (menu != NULL) {
      TgMenuItem *menu_item=NULL;
      TgMenuItem stMenuItem;

      menu->track_menubar = TRUE;
      TgAdjustMenuGeometry(menu, choiceImageW, choiceImageH, maxWhereToPrint);
      menu_item = (&menu->menuitems[whereToPrint]);

      memset(&stMenuItem, 0, sizeof(TgMenuItem));
      stMenuItem.checked = TRUE;
      if (!TgSetMenuItemInfo(menu_item, TGMU_MASK_CHECK, &stMenuItem)) {
         return TgDestroyMenu(menu, TRUE);
      }
      menu->refresh_proc = ((RefreshMenuFunc*)RefreshExportFormatMenu);
   }
   return menu;
}

int ExportFormatMenu(X, Y, TrackMenubar)
   int X, Y, TrackMenubar;
{
   int rc=INVALID;
   TgMenu *menu=(exportFormatMenuInfo.create_proc)(NULL, X, Y,
         &exportFormatMenuInfo);

   activeMenu = INVALID;
   if (menu != NULL) {
      rc = TgMenuLoop(menu);
      TgDestroyMenu(menu, TRUE);
   }
   return rc;
}

int RefreshShowHideMenu(menu)
   TgMenu *menu;
{
   int ok=TRUE;

   /* ShowBit/Pixmap */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLEMAPSHOWN, mapShown);
   /* ShowMeasurement */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLESHOWMEASUREMENT,
         showMeasurement);
   /* ShowMenubar */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLESHOWMENUBAR, !noMenubar);
   /* ShowStatus */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLESHOWSTATUS,
         !noStatusWindow);

   return ok;
}

TgMenu *CreateShowHideMenu(parent_menu, x, y, menu_info)
   TgMenu *parent_menu;
   int x, y;
   TgMenuInfo *menu_info;
{
   TgMenu *menu=TgCreateMenuFromMenuInfo(parent_menu, x, y, menu_info);

   if (menu != NULL) {
      if (!RefreshShowHideMenu(menu)) {
         return TgDestroyMenu(menu, TRUE);
      }
   }
   return menu;
}

int RefreshLayoutMenu(menu)
   TgMenu *menu;
{
   int ok=TRUE;

   /* VisibleGrid */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLEGRID, gridShown);
   /* SnapToGrid */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLESNAP, snapOn);
   /* Landscape */
   ok &= TgSetMenuItemRadioById(menu, CMDID_LANDSCAPEMODE,
         (pageStyle==LANDSCAPE));
   /* Portrait */
   ok &= TgSetMenuItemRadioById(menu, CMDID_PORTRAITMODE,
         (pageStyle==PORTRAIT));
   /* Print/ExportInColor */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLEBWCOLORPS, colorDump);
   /* MetricGrid */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLEGRIDSYSTEM,
         (gridSystem==METRIC_GRID));
   /* OneMotionSelMove */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLEONEMOTIONSELMOVE,
         oneMotionSelectMove);
   /* UseGrayScale */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLEUSEGRAYSCALE, useGray);
   /* UseColorLayers */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLECOLORLAYERS, colorLayers);
   /* StackedPage */
   ok &= TgEnableMenuItemById(menu, MENU_STACKEDPAGE,
         (pageLayoutMode==PAGE_STACK));
   /* TiledPage */
   ok &= TgEnableMenuItemById(menu, MENU_TILEDPAGE,
         (pageLayoutMode!=PAGE_STACK));

   return ok;
}

TgMenu *CreateLayoutMenu(parent_menu, x, y, menu_info)
   TgMenu *parent_menu;
   int x, y;
   TgMenuInfo *menu_info;
{
   TgMenu *menu=TgCreateMenuFromMenuInfo(parent_menu, x, y, menu_info);

   if (menu != NULL) {
      if (!RefreshLayoutMenu(menu)) {
         return TgDestroyMenu(menu, TRUE);
      }
   }
   return menu;
}

int LayoutMenu(X, Y, TrackMenubar)
   int X, Y, TrackMenubar;
{
   int rc=INVALID;
   TgMenu *menu=(layoutMenuInfo.create_proc)(NULL, X, Y, &layoutMenuInfo);

   activeMenu = MENU_LAYOUT;
   if (menu != NULL) {
      rc = TgMenuLoop(menu);
      TgDestroyMenu(menu, TRUE);
   }
   return rc;
}

int GridSize()
{
   int grid=1;

   if (snapOn) {
      switch (gridSystem) {
      case ENGLISH_GRID: grid = xyEnglishGrid; break;
      case METRIC_GRID: grid = xyMetricGrid; break;
      }
      if (zoomedIn) {
         grid <<= zoomScale;
      }
   } else {
      if (zoomedIn) {
         grid = 1 << zoomScale;
      }
   }
   return grid;
}

void GridXY(X, Y, GridX, GridY)
   int X, Y, *GridX, *GridY;
{
   int dx, dy, grid=GridSize();

   if (snapOn) {
      if (zoomedIn) {
         dx = (X+(drawOrigX<<zoomScale)) % grid;
         dy = (Y+(drawOrigY<<zoomScale)) % grid;
      } else {
         dx = (X+(drawOrigX>>zoomScale)) % grid;
         dy = (Y+(drawOrigY>>zoomScale)) % grid;
      }
      *GridX = ((dx<<1) < grid) ? X-dx : X-dx+grid;
      *GridY = ((dy<<1) < grid) ? Y-dy : Y-dy+grid;
   } else {
      if (zoomedIn) {
         dx = (X+(drawOrigX<<zoomScale)) % grid;
         dy = (Y+(drawOrigY<<zoomScale)) % grid;
         *GridX = ((dx<<1) < grid) ? X-dx : X-dx+grid;
         *GridY = ((dy<<1) < grid) ? Y-dy : Y-dy+grid;
      } else {
         *GridX = X;
         *GridY = Y;
      }
   }
}

int ControlKeyPressedEvent(ev)
   XEvent *ev;
{
   if (ev->type == KeyPress || ev->type == KeyRelease) {
      char s[80];
      KeySym key_sym;

      XLookupString(&(ev->xkey), s, sizeof(s), &key_sym, NULL);
      if (key_sym == XK_Control_L || key_sym == XK_Control_R) {
         return (ev->type == KeyPress);
      }
   }
   return FALSE;
}

int DiagEventCheck(ev)
   XEvent *ev;
{
   static int prev_diag_state=FALSE;

   if (ev->type == MotionNotify) {
      prev_diag_state = (ev->xmotion.state & ShiftMask);
   } else if (ev->type == KeyPress || ev->type == KeyRelease) {
      char s[80];
      KeySym key_sym;

      XLookupString(&(ev->xkey), s, sizeof(s), &key_sym, NULL);
      if (key_sym == XK_Shift_L || key_sym == XK_Shift_R) {
         prev_diag_state = (ev->type == KeyPress);
         return TRUE;
      }
   }
   return prev_diag_state;
}

void DiagGridXY(OrigX, OrigY, NewX, NewY)
   int OrigX, OrigY, *NewX, *NewY;
{
   int X=(*NewX), Y=(*NewY);
   int dx, dy;
   int p0, p1, p2, p3;

   dx = OrigX - X;
   dy = OrigY - Y;

   p0 = dy - (dx >> 1); /* y = x/2 */
   p1 = (dy >> 1) - dx; /* y = 2x */
   p2 = (dy >> 1) + dx; /* y = -2x */
   p3 = dy + (dx >> 1); /* y = -x/2 */

   if ( (p3 ^ p0) < 0) {
      Y = OrigY;
   } else if ( (p0 ^ p1) < 0) {
      X = OrigX - ((dx + dy)>>1);
      Y = OrigY - ((dx + dy)>>1);
   } else if ( (p1 ^ p2) >= 0) {
      X = OrigX;
   } else {
      X = OrigX - ((dx - dy)>>1);
      Y = OrigY + ((dx - dy)>>1);
      if ((GridSize() & 0x1) == 0) {
         X--;
      }
   }
   (*NewX) = X;
   (*NewY) = Y;
}

void CleanUpGrids()
{
}

void MoveModeSubMenu(index)
   int index;
{
   moveMode = index;
   switch (moveMode) {
   case CONST_MOVE: Msg("Constrained move."); break;
   case UNCONST_MOVE: Msg("Unconstrained move."); break;
   }
   ShowMoveMode();
   UpdatePinnedMenu(MENU_LAYOUT);
   UpdatePinnedMenu(MENU_MOVEMODE);
}

void RefreshMoveModeMenu(menu)
   TgMenu *menu;
{
   int i, num_items=menu->num_items;
   TgMenuItem *menuitems=menu->menuitems;

   for (i=0; i < num_items; i++) {
      TgMenuItem *menu_item=(&menuitems[i]);
      TgMenuItem stMenuItem;

      memset(&stMenuItem, 0, sizeof(TgMenuItem));
      stMenuItem.state = TGBS_NORMAL;
      stMenuItem.checked = (i == moveMode);
      TgSetMenuItemInfo(menu_item, TGMU_MASK_STATE|TGMU_MASK_CHECK,
            &stMenuItem);
   }
}

TgMenu *CreateMoveModeMenu(parent_menu, x, y, menu_info)
   TgMenu *parent_menu;
   int x, y;
   TgMenuInfo *menu_info;
{
   TgMenu *menu=TgCreateMenuFromMenuInfo(parent_menu, x, y, menu_info);

   if (menu != NULL) {
      TgMenuItem *menu_item=NULL;
      TgMenuItem stMenuItem;

      menu->track_menubar = TRUE;
      TgAdjustMenuGeometry(menu, choiceImageW, choiceImageH, MAXMOVEMODES);
      menu_item = (&menu->menuitems[moveMode]);

      memset(&stMenuItem, 0, sizeof(TgMenuItem));
      stMenuItem.checked = TRUE;
      if (!TgSetMenuItemInfo(menu_item, TGMU_MASK_CHECK, &stMenuItem)) {
         return TgDestroyMenu(menu, TRUE);
      }
      menu->refresh_proc = ((RefreshMenuFunc*)RefreshMoveModeMenu);
   }
   return menu;
}

int MoveModeMenu(X, Y, TrackMenubar)
   int X, Y, TrackMenubar;
{
   int rc=INVALID;
   TgMenu *menu=(moveModeMenuInfo.create_proc)(NULL, X, Y, &moveModeMenuInfo);

   activeMenu = MENU_MOVEMODE;
   if (menu != NULL) {
      rc = TgMenuLoop(menu);
      TgDestroyMenu(menu, TRUE);
   }
   return rc;
}

void StretchableTextModeSubMenu(index)
   int index;
{
   stretchableText = index;
   switch (stretchableText) {
   case NO_STRETCHABLE_TEXT: Msg("Text objects are not stretchable."); break;
   case STRETCHABLE_TEXT: Msg("Text objects are stretchable."); break;
   }
   ShowStretchableTextMode();
   UpdatePinnedMenu(MENU_STRETCHTEXT);
}

void RefreshStretchableTextModeMenu(menu)
   TgMenu *menu;
{
   int i, num_items=menu->num_items;
   TgMenuItem *menuitems=menu->menuitems;

   for (i=0; i < num_items; i++) {
      TgMenuItem *menu_item=(&menuitems[i]);
      TgMenuItem stMenuItem;

      memset(&stMenuItem, 0, sizeof(TgMenuItem));
      stMenuItem.state = TGBS_NORMAL;
      stMenuItem.checked = (i == stretchableText);
      TgSetMenuItemInfo(menu_item, TGMU_MASK_STATE|TGMU_MASK_CHECK,
            &stMenuItem);
   }
}

TgMenu *CreateStretchableTextModeMenu(parent_menu, x, y, menu_info)
   TgMenu *parent_menu;
   int x, y;
   TgMenuInfo *menu_info;
{
   TgMenu *menu=TgCreateMenuFromMenuInfo(parent_menu, x, y, menu_info);

   if (menu != NULL) {
      TgMenuItem *menu_item=NULL;
      TgMenuItem stMenuItem;

      menu->track_menubar = TRUE;
      TgAdjustMenuGeometry(menu, choiceImageW, choiceImageH,
            MAXSTRETCHABLEMODES);
      menu_item = (&menu->menuitems[stretchableText]);

      memset(&stMenuItem, 0, sizeof(TgMenuItem));
      stMenuItem.checked = TRUE;
      if (!TgSetMenuItemInfo(menu_item, TGMU_MASK_CHECK, &stMenuItem)) {
         return TgDestroyMenu(menu, TRUE);
      }
      menu->refresh_proc = ((RefreshMenuFunc*)RefreshStretchableTextModeMenu);
   }
   return menu;
}

int StretchableTextModeMenu(X, Y, TrackMenubar)
   int X, Y, TrackMenubar;
{
   int rc=INVALID;
   TgMenu *menu=(stretchableTextModeMenuInfo.create_proc)(NULL, X, Y,
         &stretchableTextModeMenuInfo);

   activeMenu = MENU_STRETCHTEXT;
   if (menu != NULL) {
      rc = TgMenuLoop(menu);
      TgDestroyMenu(menu, TRUE);
   }
   return rc;
}

void RefreshTransPatModeMenu(menu)
   TgMenu *menu;
{
   int i, num_items=menu->num_items;
   TgMenuItem *menuitems=menu->menuitems;

   for (i=0; i < num_items; i++) {
      TgMenuItem *menu_item=(&menuitems[i]);
      TgMenuItem stMenuItem;

      memset(&stMenuItem, 0, sizeof(TgMenuItem));
      stMenuItem.state = TGBS_NORMAL;
      stMenuItem.checked = (i == transPat);
      TgSetMenuItemInfo(menu_item, TGMU_MASK_STATE|TGMU_MASK_CHECK,
            &stMenuItem);
   }
}

TgMenu *CreateTransPatModeMenu(parent_menu, x, y, menu_info)
   TgMenu *parent_menu;
   int x, y;
   TgMenuInfo *menu_info;
{
   TgMenu *menu=TgCreateMenuFromMenuInfo(parent_menu, x, y, menu_info);

   if (menu != NULL) {
      TgMenuItem *menu_item=NULL;
      TgMenuItem stMenuItem;

      menu->track_menubar = TRUE;
      TgAdjustMenuGeometry(menu, choiceImageW, choiceImageH, MAXTRANSPATMODES);
      menu_item = (&menu->menuitems[transPat]);

      memset(&stMenuItem, 0, sizeof(TgMenuItem));
      stMenuItem.checked = TRUE;
      if (!TgSetMenuItemInfo(menu_item, TGMU_MASK_CHECK, &stMenuItem)) {
         return TgDestroyMenu(menu, TRUE);
      }
      menu->refresh_proc = ((RefreshMenuFunc*)RefreshTransPatModeMenu);
   }
   return menu;
}

int TransPatModeMenu(X, Y, TrackMenubar)
   int X, Y, TrackMenubar;
{
   int rc=INVALID;
   TgMenu *menu=(transPatModeMenuInfo.create_proc)(NULL, X, Y,
         &transPatModeMenuInfo);

   activeMenu = MENU_TRANSPAT;
   if (menu != NULL) {
      rc = TgMenuLoop(menu);
      TgDestroyMenu(menu, TRUE);
   }
   return rc;
}

