/*
 * Author:      William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990-1999, William Chia-Wei Cheng.
 *
 * Permission limited to the use, copy, display, distribute without
 * charging for a fee, and produce derivative works of "tgif" and
 * its documentation for not-for-profit purpose is hereby granted by
 * the Author, provided that the above copyright notice appears in
 * all copies made of "tgif" and that both the copyright notice
 * and this permission notice appear in supporting documentation,
 * and that the name of the Author not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission.  The Author makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.  All other rights (including, but not limited to, the
 * right to sell "tgif", the right to sell or distribute derivative
 * works of "tgif", the right to distribute "tgif" for a fee, and
 * the right to include "tgif" or derivative works of "tgif" in a
 * for-sale product) are reserved by the Author.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
 * OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * @(#)$Header: /mm/src/tgif/v4/RCS/menuinfo.c,v 4.7 1999/02/05 22:53:09 william Exp $
 */

#define _INCLUDE_FROM_MENUINFO_C_

#include "tgifdefs.h"
#include "cmdids.h"

#include "align.e"
#include "color.e"
#include "edit.e"
#include "file.e"
#include "font.e"
#include "grid.e"
#include "help.e"
#include "menu.e"
#include "menuinfo.e"
#include "navigate.e"
#include "page.e"
#include "pattern.e"
#include "raster.e"
#include "scroll.e"
#include "shape.e"
#include "stream.e"
#include "text.e"

#ifdef _TANGRAM_II
#include "tangram2.e"
#endif /* _TANGRAM_II */

/* ------------------ color menus ------------------ */

TgMenuInfo colorMenuInfo={ TGMUTYPE_COLOR, NULL, CreateColorMenu };

/* ------------------ bitmap menus ------------------ */

static TgMenuItemInfo modeMenuItemInfo[] = {
   { (char*)(&choicePixmap[NOTHING]), NULL,
     "Enter select/move/resize object mode", NULL, CMDID_SELECTMODE },
   { (char*)(&choicePixmap[DRAWTEXT]), NULL,
     "Enter draw text mode", NULL, CMDID_DRAWTEXT },
   { (char*)(&choicePixmap[DRAWBOX]), NULL,
     "Enter draw rectangle mode", NULL, CMDID_DRAWBOX },
   { (char*)(&choicePixmap[DRAWCORNEROVAL]), NULL,
     "Enter draw corner oval mode", NULL, CMDID_DRAWCORNEROVAL },
   { (char*)(&choicePixmap[DRAWCENTEROVAL]), NULL,
     "Enter draw center oval mode", NULL, CMDID_DRAWCENTEROVAL },
   { (char*)(&choicePixmap[DRAWEDGECIRCLE]), NULL,
     "Enter draw edge circle mode", NULL, CMDID_DRAWEDGEOVAL },
   { (char*)(&choicePixmap[DRAWPOLY]), NULL,
     "Enter draw poly/open spline mode", NULL, CMDID_DRAWPOLY },
   { (char*)(&choicePixmap[DRAWPOLYGON]), NULL,
     "Enter draw polygon/close spline mode", NULL, CMDID_DRAWPOLYGON },
   { (char*)(&choicePixmap[DRAWARC]), NULL,
     "Enter draw arc mode (specify center first)", NULL, CMDID_DRAWARC },
   { (char*)(&choicePixmap[DRAWEDGEARC]), NULL,
     "Enter draw arc mode (specify endpoints first)", NULL, CMDID_DRAWEDGEARC },
   { (char*)(&choicePixmap[DRAWRCBOX]), NULL,
     "Enter draw rcbox mode", NULL, CMDID_DRAWRCBOX },
   { (char*)(&choicePixmap[FREEHAND]), NULL,
     "Enter draw freehand poly/open spline mode", NULL, CMDID_FREEHANDMODE },
   { (char*)(&choicePixmap[VERTEXMODE]), NULL,
     "Enter select/move vertices mode", NULL, CMDID_SELECTVERTEXMODE },
   { (char*)(&choicePixmap[ROTATEMODE]), NULL,
     "Enter rotate/shear object mode", NULL, CMDID_ROTATESHEARMODE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo modeMenuInfo={ TGMUTYPE_BITMAP, modeMenuItemInfo, CreateModeMenu };

TgMenuInfo fillMenuInfo={ TGMUTYPE_BITMAP, NULL, CreateFillMenu };
TgMenuInfo penMenuInfo={ TGMUTYPE_BITMAP, NULL, CreatePenMenu };
TgMenuInfo lineWidthMenuInfo={ TGMUTYPE_BITMAP, NULL, CreateLineWidthMenu };

static TgMenuItemInfo lineStyleMenuItemInfo[] = {
   { (char*)(&lineStylePixmap[LS_PLAIN]), NULL,
     "No arrows for a poly/open spline", NULL, CMDID_CHANGEALLSELLINESTYLE },
   { (char*)(&lineStylePixmap[LS_RIGHT]), NULL,
     "An arrow at the end of a poly/open spline",
     NULL, CMDID_CHANGEALLSELLINESTYLE },
   { (char*)(&lineStylePixmap[LS_LEFT]), NULL,
     "An arrow at the beginning of a poly/open spline",
     NULL, CMDID_CHANGEALLSELLINESTYLE },
   { (char*)(&lineStylePixmap[LS_DOUBLE]), NULL,
     "Arrows at both ends of a poly/open spline",
     NULL, CMDID_CHANGEALLSELLINESTYLE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo lineStyleMenuInfo={ TGMUTYPE_BITMAP, lineStyleMenuItemInfo,
                               CreateLineStyleMenu };

static TgMenuItemInfo lineTypeMenuItemInfo[] = {
   { (char*)(&lineTypePixmap[LT_STRAIGHT]), NULL,
     "Straight line segments", NULL, CMDID_CHANGEALLSELLINETYPE },
   { (char*)(&lineTypePixmap[LT_SPLINE]), NULL,
     "Open/closed splines", NULL, CMDID_CHANGEALLSELLINETYPE },
   { (char*)(&lineTypePixmap[LT_INTSPLINE]), NULL,
     "Interpolated splines", NULL, CMDID_CHANGEALLSELLINETYPE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo lineTypeMenuInfo={ TGMUTYPE_BITMAP, lineTypeMenuItemInfo,
                              CreateLineTypeMenu };

TgMenuInfo lineDashMenuInfo={ TGMUTYPE_BITMAP, NULL, CreateLineDashMenu };

static TgMenuItemInfo horiAlignMenuItemInfo[] = {
   { (char*)(&alignHoriPixmap[ALIGN_N]), NULL,
     "Set horizontal alignment to none",
     NULL, CMDID_SETHORIZONTALALIGN },
   { (char*)(&alignHoriPixmap[ALIGN_L]), NULL,
     "Set horizontal alignment to left aligned",
     NULL, CMDID_SETHORIZONTALALIGN },
   { (char*)(&alignHoriPixmap[ALIGN_C]), NULL,
     "Set horizontal alignment to center aligned",
     NULL, CMDID_SETHORIZONTALALIGN },
   { (char*)(&alignHoriPixmap[ALIGN_R]), NULL,
     "Set horizontal alignment to right aligned",
     NULL, CMDID_SETHORIZONTALALIGN },
   { (char*)(&alignHoriPixmap[ALIGN_S]), NULL,
     "Set horizontal distribution to equal-spacing",
     NULL, CMDID_SETHORIZONTALALIGN },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo horiAlignMenuInfo={ TGMUTYPE_BITMAP, horiAlignMenuItemInfo,
                               CreateHoriAlignMenu };

static TgMenuItemInfo vertAlignMenuItemInfo[] = {
   { (char*)(&alignVertPixmap[ALIGN_N]), NULL,
     "Set vertical alignment to none",
     NULL, CMDID_SETVERTICALALIGN },
   { (char*)(&alignVertPixmap[ALIGN_T]), NULL,
     "Set vertical alignment to top aligned",
     NULL, CMDID_SETVERTICALALIGN },
   { (char*)(&alignVertPixmap[ALIGN_M]), NULL,
     "Set vertical alignment to middle aligned",
     NULL, CMDID_SETVERTICALALIGN },
   { (char*)(&alignVertPixmap[ALIGN_B]), NULL,
     "Set vertical alignment to bottom aligned",
     NULL, CMDID_SETVERTICALALIGN },
   { (char*)(&alignVertPixmap[ALIGN_S]), NULL,
     "Set vertical distribution to equal-spacing",
     NULL, CMDID_SETVERTICALALIGN },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo vertAlignMenuInfo={ TGMUTYPE_BITMAP, vertAlignMenuItemInfo,
                               CreateVertAlignMenu };

static TgMenuItemInfo moveModeMenuItemInfo[] = {
   { (char*)(&moveModePixmap[CONST_MOVE]), NULL,
     "Set constrained move mode", NULL, CMDID_SETMOVEMODE },
   { (char*)(&moveModePixmap[UNCONST_MOVE]), NULL,
     "Set unconstrained move mode", NULL, CMDID_SETMOVEMODE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo moveModeMenuInfo={ TGMUTYPE_BITMAP, moveModeMenuItemInfo,
                              CreateMoveModeMenu };

static TgMenuItemInfo stretchableTextModeMenuItemInfo[] = {
   { (char*)(&stretchableModePixmap[NO_STRETCHABLE_TEXT]), NULL,
     "Make text objects stretchable", NULL, CMDID_SETSTRETCHTEXTMODE },
   { (char*)(&stretchableModePixmap[STRETCHABLE_TEXT]), NULL,
     "Make text objects not stretchable", NULL, CMDID_SETSTRETCHTEXTMODE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo stretchableTextModeMenuInfo={ TGMUTYPE_BITMAP,
                                         stretchableTextModeMenuItemInfo,
                                         CreateStretchableTextModeMenu };

TgMenuInfo shapeMenuInfo={ TGMUTYPE_BITMAP, NULL, CreateShapeMenu };

static TgMenuItemInfo pageLayoutMenuItemInfo[] = {
   { (char*)(&pageLayoutPixmap[PAGE_STACK]), NULL,
     "Stacked page mode", NULL, CMDID_SETPAGELAYOUTMODE },
   { (char*)(&pageLayoutPixmap[PAGE_TILE]), NULL,
     "Tiled page mode", NULL, CMDID_SETPAGELAYOUTMODE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo pageLayoutMenuInfo={ TGMUTYPE_BITMAP, pageLayoutMenuItemInfo,
                                CreatePageLayoutMenu };

static TgMenuItemInfo transPatModeMenuItemInfo[] = {
   { (char*)(&transPatPixmap[NO_TRANSPAT_MODE]), NULL,
     "Make fill and pen patterns opaque", NULL, CMDID_SETTRANSPATMODE },
   { (char*)(&transPatPixmap[TRANSPAT_MODE]), NULL,
     "Make fill and pen patterns transparent", NULL, CMDID_SETTRANSPATMODE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo transPatModeMenuInfo={ TGMUTYPE_BITMAP, transPatModeMenuItemInfo,
                                  CreateTransPatModeMenu };

static TgMenuItemInfo pageStyleMenuItemInfo[] = {
   { "Portrait", NULL, "Portrait page style", NULL, CMDID_PORTRAITMODE },
   { "Landscape", NULL, "Lanscape page style", NULL,
     CMDID_LANDSCAPEMODE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo pageStyleMenuInfo={ TGMUTYPE_TEXT, pageStyleMenuItemInfo,
                               CreatePageStyleMenu };

/* ------------------ text menus ------------------ */

TgMenuInfo fontMenuInfo={ TGMUTYPE_TEXT, NULL, CreateFontMenu };

static TgMenuItemInfo fontStyleMenuItemInfo[] = {
   { "Roman", "Alt+Cntrl+o", "Roman text style", NULL,
     CMDID_CHANGEALLSELFONTSTYLE },
   { "Bold", "Alt+Cntrl+b", "Bold text style", NULL,
     CMDID_CHANGEALLSELFONTSTYLE },
   { "Italic", "Alt+Cntrl+t", "Italic text style", NULL,
     CMDID_CHANGEALLSELFONTSTYLE },
   { "BoldItalic", "Alt+Cntrl+p", "Bold-italic text style", NULL,
     CMDID_CHANGEALLSELFONTSTYLE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Left", "Alt+Cntrl+l", "Left justified text", NULL,
     CMDID_CHANGEALLSELFONTJUST },
   { "Center", "Alt+Cntrl+c", "Center justified text", NULL,
     CMDID_CHANGEALLSELFONTJUST },
   { "Right", "Alt+Cntrl+r", "Right justified text", NULL,
     CMDID_CHANGEALLSELFONTJUST },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo fontStyleMenuInfo={ TGMUTYPE_TEXT, fontStyleMenuItemInfo,
                               CreateFontStyleMenu };

TgMenuInfo fontSizeMenuInfo={ TGMUTYPE_TEXT, NULL, CreateFontSizeMenu };

static TgMenuItemInfo stackedPageMenuItemInfo[] = {
   { "NextPage", NULL, "Go to next page", NULL, CMDID_NEXTPAGE },
   { "PrevPage", NULL, "Go to previous page", NULL, CMDID_PREVPAGE },
   { "NamePages...", NULL, "Assign names to pages", NULL, CMDID_NAMEPAGES },
   { "GotoPage...", NULL, "Go to a specified page", NULL, CMDID_GOTOPAGE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "AddPageBefore...", NULL, "Add a page before the current page", NULL,
     CMDID_ADDPAGEBEFORE },
   { "AddPageAfter...", NULL, "Add a page after the current page", NULL,
     CMDID_ADDPAGEAFTER },
   { "DeleteCurPage", NULL, "Delete the current page", NULL,
     CMDID_DELETECURPAGE },
   { "DeletePages...", NULL, "Delete user-specified pages", NULL,
     CMDID_DELETEPAGES },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "PrintOnePage", NULL, "Print the curret page", NULL, CMDID_PRINTONEPAGE },
   { "PrintOneFilePerPage", NULL, "Print into a separate file for each page",
     NULL, CMDID_PRINTONEFILEPERPAGE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SpecifyPaperSize...", NULL, "Set the physical size of the paper", NULL,
     CMDID_SPECIFYPAPERSIZE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo stackedPageMenuInfo={ TGMUTYPE_TEXT, stackedPageMenuItemInfo,
                                 TgCreateMenuFromMenuInfo };

static TgMenuItemInfo tiledPageMenuItemInfo[] = {
   { "TogglePageLineShown", NULL, "Toggle the showing of page boundaries", NULL,
     CMDID_TOGGLEPAGELINESHOWN },
   { "SpecifyDrawingSize...", NULL, "Change the drawing size", NULL,
     CMDID_SPECIFYDRAWINGSIZE },
   { "PrintOnePage", NULL, "Select and print one of the pages", NULL,
     CMDID_PRINTONEPAGE },
   { "SpecifyPaperSize...", NULL, "Set the physical size of the paper", NULL,
     CMDID_SPECIFYPAPERSIZE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo tiledPageMenuInfo={ TGMUTYPE_TEXT, tiledPageMenuItemInfo,
                               TgCreateMenuFromMenuInfo };

static TgMenuItemInfo pageMenuItemInfo[] = {
   { "StackedPage", TGMUITEM_PINNABLESUBMENU,
     "StackedPage Menu", &stackedPageMenuInfo, MENU_STACKEDPAGE },
   { "TiledPage", TGMUITEM_PINNABLESUBMENU,
     "TiledPage Menu", &tiledPageMenuInfo, MENU_TILEDPAGE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo pageMenuInfo={ TGMUTYPE_TEXT, pageMenuItemInfo, CreatePageMenu };

static TgMenuItemInfo reduceColorMenuItemInfo[] = {
   { "ReduceColors...", NULL, "Reduce the number of colors of a selected image",
     NULL, CMDID_REDUCECOLORS },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SetDefaultColorLevels...", NULL,
     "Set the number of R G B color levels for the ``standard colors''", NULL,
     CMDID_SETDEFAULTCOLORLEVELS },
   { "ReduceToPixmapColors...", NULL,
     "Reduce the colors of a selected image to a the colors in an XPM file",
     NULL, CMDID_REDUCETOPIXMAPCOLORS },
   { "ReduceToDefaultColors", NULL,
     "Reduce the colors of a selected image to standard colors", NULL,
     CMDID_REDUCETODEFAULTCOLORs },
   { "DefaultErrorDiffuse", NULL,
     "Error diffuse to reduce colors of a selected image to standard colors",
     NULL, CMDID_DEFAULTERRORDIFFUSE },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo reduceColorMenuInfo={ TGMUTYPE_TEXT, reduceColorMenuItemInfo,
                                        TgCreateMenuFromMenuInfo };

static TgMenuItemInfo effectsMenuItemInfo[] = {
   { "EdgeDetect", NULL, "Perform edge-detection on a selected image", NULL,
     CMDID_EDGEDETECT },
   { "Emboss", NULL, "Convert a selected image to gray and then emboss", NULL,
     CMDID_EMBOSS },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Spread...", NULL, "Spread the pixels of a selected image around", NULL,
     CMDID_SPREAD },
   { "Sharpen", NULL, "Sharpen a selected image", NULL, CMDID_SHARPEN },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Blur(3x3)", NULL, "Blur a selected image using a 3 by 3 filter", NULL,
     CMDID_BLUR3 },
   { "Blur(5x5)", NULL, "Blur a selected image using a 5 by 5 filter", NULL,
     CMDID_BLUR5 },
   { "Blur(7x7)", NULL, "Blur a selected image using a 7 by 7 filter", NULL,
     CMDID_BLUR7 },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo effectsMenuInfo={ TGMUTYPE_TEXT, effectsMenuItemInfo,
                                    TgCreateMenuFromMenuInfo };

static TgMenuItemInfo generateMenuItemInfo[] = {
   { "RunBggen...", NULL,
     "Run 'bggen' to create an image using the size of a selected image", NULL,
     CMDID_RUNBGGEN },
   { "CircularBggen...", NULL,
     "Create a gray circular image using the size of a selected image", NULL,
     CMDID_CIRCULARBGGEN },
   { "SimpleRectBggen...", NULL,
     "Create a single-color rectanglar image using the size of selected image",
     NULL, CMDID_SIMPLERECTBGGEN },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo generateMenuInfo={ TGMUTYPE_TEXT, generateMenuItemInfo,
                                     TgCreateMenuFromMenuInfo };

static TgMenuItemInfo imageProcMenuItemInfo[] = {
   { "MakeGray", NULL, "Convert a selected image into gray-scale", NULL,
     CMDID_MAKEGRAY },
   { "Brighten/Darken...", NULL, "Brighten or darken a selected image", NULL,
     CMDID_BRIGHTENDARKEN },
   { "ContrastEnhance...", NULL, "Adjust contrast of a selected image", NULL,
     CMDID_CONTRASTENHANCE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ChangeSaturation...", NULL, "Change saturation of a selected image", NULL,
     CMDID_CHANGESATURATION },
   { "ChangeHue...", NULL, "Change hue of a selected image", NULL,
     CMDID_CHANGEHUE },
   { "GammaCorrect...", NULL, "Apply gamma correction a selected image", NULL,
     CMDID_GAMMA },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "InvertColor", NULL,
     "Subtract every pixel of a selected image from White (in RGB-model)",
     NULL, CMDID_INVERTCOLOR },
   { "InterpolateColor...", NULL,
     "Map pixel intensities of a selected image between two specified colors",
     NULL, CMDID_INTERPOLATECOLOR },
   { "ColorBalance...", NULL, "Balance the RGB components of a selected image",
     NULL, CMDID_COLORBALANCE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ReduceNumberOfColors", TGMUITEM_SUBMENU,
     "Reduce number of colors submenu >>>", &reduceColorMenuInfo, INVALID },
   { "Effects", TGMUITEM_SUBMENU,
     "Effects submenu >>>", &effectsMenuInfo, INVALID },
   { "Generate", TGMUITEM_SUBMENU,
     "Generate submenu >>>", &generateMenuInfo, INVALID },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "RegenerateImage", NULL,
     "Regenerated a selected image", NULL, CMDID_REGENERATEIMAGE },
   { "CropImage", NULL, "Crop a selected image", NULL, CMDID_CROPIMAGE },
   { "GetColor", NULL,
     "Pick a color from a selected image to be the current color", NULL,
     CMDID_GETCOLOR },
   { "ReplaceColor", NULL,
     "Replace a color in a selected image with the current color", NULL,
     CMDID_REPLACECOLOR },
   { "FloodFill", NULL, "Flood-filling a selected image with the current color",
     NULL, CMDID_FLOODFILL },
   { "CreateContour", NULL,
     "Generate a contour from a point in a selected image", NULL,
     CMDID_CREATECONTOUR },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Subtract", NULL,
     "Subtract one image from another", NULL, CMDID_SUBTRACT },
   { "AlphaCombine", NULL,
     "Combine 2 images using another image as the alpha channel", NULL,
     CMDID_ALPHACOMBINE },
   { "XorColor", NULL, "XOR the colors of one image into another", NULL,
     CMDID_XORCOLORS },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "VectorWarp", NULL,
     "Warp selected image in a specified direction for a specified amount",
     NULL, CMDID_VECTORWARP },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo imageProcMenuInfo={ TGMUTYPE_TEXT, imageProcMenuItemInfo,
                               TgCreateMenuFromMenuInfo };

static TgMenuItemInfo driverMenuItemInfo[] = {
   { "Solve", "Alt+s", "Escape to driver", NULL, CMDID_SOLVE },
   { "Simulate", "Alt+y", "Escape to driver", NULL, CMDID_SIMULATE },
   { "Probe", "Alt+b", "Escape to driver", NULL, CMDID_PROBE },
   { "Animate", "Cntrl+z", "Escape to driver", NULL, CMDID_ANIMATE },
   { "Escape", "Alt+x", "Escape to driver", NULL, CMDID_ESCAPE },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo driverMenuInfo={ TGMUTYPE_TEXT, driverMenuItemInfo,
                                   TgCreateMenuFromMenuInfo };

static TgMenuItemInfo importMenuItemInfo[] = {
   { "Import...", "Alt+p", "Embed/Import another drawing", NULL, CMDID_IMPORT },
   { "ImportXBitmap...", "Alt+Cntrl+.", "Embed/Import an X11 bitmap file", NULL,
     CMDID_IMPORTXBITMAP },
   { "ImportXPixmap...", "Alt+Cntrl+,", "Embed/Import an X11 pixmap file", NULL,
     CMDID_IMPORTXPIXMAP },
   { "ImportEPSFile...", "Alt+(", "Link to an PS/EPS file", NULL,
     CMDID_IMPORTEPSFILE },
   { "ImportGIFFile...", NULL, "Embed/Import a GIF file", NULL,
     CMDID_IMPORTGIFFILE },
   { "ImportOtherFile...", NULL, "Embed/Import another type of file", NULL,
     CMDID_IMPORTOTHERFILE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "EmbedEPSFile...", NULL, "Embed/Import an PS/EPS file", NULL,
     CMDID_EMBEDEPSFILE },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo importMenuInfo={ TGMUTYPE_TEXT, importMenuItemInfo,
                                   TgCreateMenuFromMenuInfo };

static TgMenuItemInfo browseMenuItemInfo[] = {
   { "BrowseXBitmap", NULL, "Recursively read in all X11 bitmap files", NULL,
     CMDID_BROWSEXBITMAP },
   { "BrowseXPixmap", NULL, "Recursively read in all X11 pixmap files", NULL,
     CMDID_BROWSEXPIXMAP },
   { "BrowseOther...", NULL, "Recursively read in all another type of files",
     NULL, CMDID_BROWSEOTHER },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo browseMenuInfo={ TGMUTYPE_TEXT, browseMenuItemInfo,
                                   TgCreateMenuFromMenuInfo };

TgMenuInfo exportFormatMenuInfo={ TGMUTYPE_BITMAP, NULL,
                                  CreateExportFormatMenu };

static TgMenuItemInfo gifAnimateMenuItemInfo[] = {
   { "ExportXPixmapDeckToGIF", NULL,
     "Export a deck of X11 pixmap images to an animated GIF file", NULL,
     CMDID_EXPORTXPMDECKTOGIF },
   { "ImportGIFToXPixmapDeck...", NULL,
     "Import an animated GIF file into a deck of X11 pixmap images", NULL,
     CMDID_IMPORTGIFTOXPMDECK },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo gifAnimateMenuInfo={ TGMUTYPE_TEXT, gifAnimateMenuItemInfo,
                                       TgCreateMenuFromMenuInfo };

#ifdef _TGIF_DBG
static TgMenuItemInfo mimeMenuItemInfo[] = {
   { "FakeUserAgent...", NULL,
     "Set 'UserAgent' to be used when making an HTTP request", NULL,
     CMDID_FAKEUSERAGENT },
   { "FakeReferer...", NULL,
     "Set 'Referer' to be used when making an HTTP request", NULL,
     CMDID_FAKEREFERER },
   { "ToggleKeepAlive", NULL,
     "Toggle between using Keep-Alive and non-Keep-Alive HTTP connections",
     NULL, CMDID_TOGGLEKEEPALIVE },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo mimeMenuInfo={ TGMUTYPE_TEXT, mimeMenuItemInfo,
                                 TgCreateMenuFromMenuInfo };
#endif /* _TGIF_DBG */

static TgMenuItemInfo fileMenuItemInfo[] = {
   { "New", "Cntrl+n", "Start with a blank/new drawing", NULL, CMDID_NEW },
   { "Open...", "Cntrl+o", "Open an existing drawing", NULL, CMDID_OPEN },
   { "Import", TGMUITEM_SUBMENU,
     "Import submenu >>>", &importMenuInfo, INVALID },
   { "Browse", TGMUITEM_SUBMENU,
     "Browse submenu >>>", &browseMenuInfo, INVALID },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Save", "Cntrl+s", "Save current drawing", NULL, CMDID_SAVE },
   { "SaveNew...", "Alt+Cntrl+s", "Save current drawing in a different file",
     NULL, CMDID_SAVENEW },
   { "SaveSelectedAs...", "Alt+~", "Save selected objects in a different file",
     NULL, CMDID_SAVESELECTEDAS },
   { "SaveSymInLibrary...", NULL,
     "Save symbol file in a library path within the current domain", NULL,
     CMDID_SAVESYMINLIBRARY },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ChangeDomain...", "Cntrl+c", "Change to a different domain", NULL,
     CMDID_CHANGEDOMAIN },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Print", "Cntrl+p", "Print/export current drawing", NULL, CMDID_PRINT },
   { "Print/ExportFormat", TGMUITEM_SUBMENU,
     "Select print/export format submenu >>>", &exportFormatMenuInfo, INVALID },
   { "Print/ExportInColor", "Alt+Cntrl+k",
     "Toggle between color and black & white printing modes", NULL,
     CMDID_TOGGLEBWCOLORPS },
   { "PrintWithCmd...", "Cntrl+-",
     "Print/export current drawing with a specific command", NULL,
     CMDID_PRINTWITHCMD },
   { "PrintSelectedObjs", NULL, "Print/export only selected objects", NULL,
     CMDID_PRINTSELECTEDOBJS },
   { "PrintOnePage", NULL, "Print a page in a multipage drawing", NULL,
     CMDID_PRINTONEPAGE },
   { "PrintOneFilePerPage", NULL, "Print into a separate file for each page",
     NULL, CMDID_PRINTONEFILEPERPAGE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "GIFAnimation", TGMUITEM_SUBMENU,
     "GIFAnimation submenu >>>", &gifAnimateMenuInfo, INVALID },
#ifdef _TGIF_DBG
   { "Mime", TGMUITEM_SUBMENU, "Mime submenu >>>",
     &mimeMenuInfo, INVALID },
#endif /* _TGIF_DBG */
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "PrintMsgBuffer...", NULL,
     "Display/save the contents of the message window", NULL,
     CMDID_PRINTMSGBUFFER },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SetExportPixelTrim...", NULL,
     "Specify the number of pixels to trim when exporting", NULL,
     CMDID_SETEXPORTPIXELTRIM },
   { "SetTemplate...", NULL,
     "Select a template file be used in STACKED page mode", NULL,
     CMDID_SETTEMPLATE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Return", TGMUITEM_SUBMENU,
     "Return submenu >>>", &driverMenuInfo, INVALID },
   { "Quit", "Cntrl+q", "Exit <<PROGRAM_NAME>>", NULL, CMDID_QUIT },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo fileMenuInfo={ TGMUTYPE_TEXT, fileMenuItemInfo, CreateFileMenu };

static TgMenuItemInfo transformMenuItemInfo[] = {
   { "PreciseScale...", "Alt+)", "Scale selected objects", NULL,
     CMDID_PRECISESCALE },
   { "PreciseRotate...", NULL,
     "Rotate selected objects by a specified number of degrees", NULL,
     CMDID_PRECISEROTATE },
   { "RestoreImageWH", NULL,
     "Restore a seleted bitmap/pixmap/EPS object to its original dimension",
     NULL, CMDID_RESTOREIMAGEWH },
   { "NoTransform", NULL,
     "Get rid of transformations on selected simple objects", NULL,
     CMDID_NOTRANSFORM },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo transformMenuInfo={ TGMUTYPE_TEXT, transformMenuItemInfo,
                                      TgCreateMenuFromMenuInfo };

static TgMenuItemInfo bitPixmapMenuItemInfo[] = {
   { "InvertXBitmap", "Alt+Cntrl+f",
     "Invert pixels for selected bitmap objects", NULL, CMDID_INVERTXBITMAP },
   { "CutBit/Pixmap...", "Alt+;", "Trim/scale a selected bitmap/pixmap object",
     NULL, CMDID_CUTBITPIXMAP },
   { "BreakUpBit/Pixmap...", NULL,
     "Break selected bitmap/pixmap object into smaller objects", NULL,
     CMDID_BREAKUPBITPIXMAP },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "UpdateEPS", NULL, "Refresh a selected EPS object", NULL,
     CMDID_UPDATEEPS },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo bitPixmapMenuInfo={ TGMUTYPE_TEXT, bitPixmapMenuItemInfo,
                                      TgCreateMenuFromMenuInfo };

static TgMenuItemInfo polyMenuItemInfo[] = {
   { "DeletePoint", "Alt+Cntrl+d", "Delete vertices from a poly/polygon/spline",
     NULL, CMDID_DELETEPOINT },
   { "AddPoint", "Alt+Cntrl+a", "Add vertices to a polygon/polygon/spline",
     NULL, CMDID_ADDPOINT },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ConvertIntSpline", NULL,
     "Convert selected interpolated splines to regular splines", NULL,
     CMDID_CONVERTINTSPLINE },
   { "Smooth<->Hinge", NULL,
     "Toggle smooth/hinge vertices for a selected poly/polygon/spline", NULL,
     CMDID_TOGGLESMOOTHHINGE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "JoinPoly", NULL, "Join polylines or open-splines at endpoints", NULL,
     CMDID_JOINPOLY },
   { "CutPoly", NULL, "Cut a poly/polygon at a vertex into two pieces", NULL,
     CMDID_CUTPOLY },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo polyMenuInfo={ TGMUTYPE_TEXT, polyMenuItemInfo,
                                 TgCreateMenuFromMenuInfo };

static TgMenuItemInfo textMenuItemInfo[] = {
   { "BreakUpText", "Alt+#",
     "Break up selected text objects into one-character text objects", NULL,
     CMDID_BREAKUPTEXT },
   { "SetTextBackgroundColor", "Alt+#",
     "Set the background color of a text object to be the current color", NULL,
     CMDID_SETTEXTBACKGROUNDCOLOR },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "InsertRightSuperscript", NULL,
     "Insert superscript to the right of the text cursor", NULL,
     CMDID_INSERTRIGHTSUPERSCRIPT },
   { "InsertRightSubscript", NULL,
     "Insert subscript to the right of the text cursor", NULL,
     CMDID_INSERTRIGHTSUBSCRIPT },
   { "InsertLeftSuperscript", NULL,
     "Insert superscript to the left of the text cursor", NULL,
     CMDID_INSERTLEFTSUPERSCRIPT },
   { "InsertLeftSubscript", NULL,
     "Insert subscript to the left of the text cursor", NULL,
     CMDID_INSERTLEFTSUBSCRIPT },
   { "InsertCenterSuperscript", NULL,
     "Insert superscript above the text cursor", NULL,
     CMDID_INSERTCENTERSUPERSCRIPT },
   { "InsertCenterSubscript", NULL,
     "Insert subscript below the text cursor", NULL,
     CMDID_INSERTCENTERSUBSCRIPT },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "InsertThinSpace...", NULL,
     "Insert a think space at the text cursor", NULL,
     CMDID_INSERTTHINSPACE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "InsertVerticalOffset...", NULL,
     "Insert a vertical offset at the text cursor", NULL,
     CMDID_INSERTVERTICALOFFSET },
   { "RemoveVerticalOffset...", NULL,
     "Remove any vertical offset at the text cursor", NULL,
     CMDID_REMOVEVERTICALOFFSET },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SetScriptFraction", NULL,
     "Set the relative size of super/subscript text", NULL,
     CMDID_SETSCRIPTFRACTION },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo textMenuInfo={ TGMUTYPE_TEXT, textMenuItemInfo,
                                 CreateEditTextMenu };

static TgMenuItemInfo createObjectMenuItemInfo[] = {
   { "InputPolyPts", NULL, "Read points from terminal and create a poly", NULL,
     CMDID_INPUTPOLYPTS },
   { "InputPolygonPts", NULL, "Read points from terminal and create a polygon",
     NULL, CMDID_INPUTPOLYGONPTS },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SpecifyAnArc...", "Alt+9", "Create and specify an arc", NULL,
     CMDID_SPECIFYANARC },
   { "MakeRegPolygon...", "Alt+\"",
     "Make a selected polygon/closed spline into a regular one", NULL,
     CMDID_MAKEREGPOLYGON },
   { "GetBoundingBox", NULL,
     "Get rectangular bounding boxes of selected objects", NULL,
     CMDID_GETBOUNDINGBOX },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo createObjectMenuInfo={ TGMUTYPE_TEXT,
                                         createObjectMenuItemInfo,
                                         TgCreateMenuFromMenuInfo };

static TgMenuItemInfo editMenuItemInfo[] = {
   { "Redraw", "Cntrl+r", "Redraw canvas window", NULL, CMDID_REDRAW },
   { "Duplicate", "Cntrl+d", "Duplicate selected objects", NULL,
     CMDID_DUPLICATE },
   { "Delete", "Cntrl+x", "Delete selected objects", NULL, CMDID_DELETE },
   { "SelectAll", "Cntrl+a", "Select all objects on the current page", NULL,
     CMDID_SELECTALL },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Undo", "Alt+u", "Undo the most recent command", NULL, CMDID_UNDO },
   { "Redo", "Alt+*", "Redo the most undone command", NULL, CMDID_REDO },
   { "FlushUndoBuffer", NULL, "Flush undo buffer and unused colors", NULL,
     CMDID_FLUSHUNDOBUFFER },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Copy", "Cntrl+y", "Copy selected objects into the cut buffer", NULL,
     CMDID_COPY },
   { "CopyPlainTextAsObject", NULL,
     "Copy highlighted plain text as an object into the cut buffer", NULL,
     CMDID_COPYPLAINTEXTASOBJECT },
   { "Cut", NULL, "Cut selected objects into the cut buffer", NULL, CMDID_CUT },
   { "Paste", "Alt+Cntrl+y", "Paste from the cut buffer", NULL, CMDID_PASTE },
   { "PasteFromFile...", NULL, "Select a file and paste the content as text",
     NULL, CMDID_PASTEFROMFILE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Shape", TGMUITEM_PINNABLESUBMENU,
     "Shape submenu >>>", &shapeMenuInfo, MENU_SHAPE },
   { "SetShapeShadow...", NULL, "Specify the shadow for shapes", NULL,
     CMDID_SETSHAPESHADOW },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ImageProc", TGMUITEM_PINNABLESUBMENU,
     "ImageProc submenu >>>", &imageProcMenuInfo, MENU_IMAGEPROC },
   { "Properties", TGMUITEM_PINNABLESUBMENU,
     "Properties submenu >>>", &propertiesMenuInfo, MENU_PROPERTIES },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Transformation", TGMUITEM_SUBMENU,
     "Transformation submenu >>>", &transformMenuInfo, INVALID },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Bitmap/Pixmap/EPS", TGMUITEM_SUBMENU,
     "Bitmap/Pixmap/EPS submenu >>>", &bitPixmapMenuInfo, INVALID },
   { "Poly/Polygon", TGMUITEM_SUBMENU,
     "Poly/Polygon submenu >>>", &polyMenuInfo, INVALID },
   { "Text", TGMUITEM_SUBMENU, "Text submenu >>>", &textMenuInfo, INVALID },
   { "CreateObject", TGMUITEM_SUBMENU,
     "Create object submenu >>>", &createObjectMenuInfo, INVALID },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "PushCurChoice", "Cntrl+e", "Save/restore current drawing mode", NULL,
     CMDID_PUSHCURCHOICE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "FindCaseSensitive...", "Alt+Cntrl+0",
     "Find string in visible text objects (case sensitive)", NULL,
     CMDID_FINDCASESENSITIVE },
   { "FindNoCase...", "Alt+Cntrl+9",
     "Find string in visible text objects (case insensitive)", NULL,
     CMDID_FINDNOCASE },
   { "FindAgain", "Alt+Cntrl+=", "Repeat the last Find command", NULL,
     CMDID_FINDAGAIN },
/*
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ConnectPins", NULL, "Connect pins", NULL, CMDID_CONNECTPINS },
 */
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo editMenuInfo={ TGMUTYPE_TEXT, editMenuItemInfo, CreateEditMenu };

static TgMenuItemInfo alignMenuItemInfo[] = {
   { "AlignObjsTop", "Alt+{", "Align top sides of selected objects", NULL,
     CMDID_ALIGNOBJSTOP },
   { "AlignObjsMiddle", "Alt++", "Align vertical centers of selected objects",
     NULL, CMDID_ALIGNOBJSMIDDLE },
   { "AlignObjsBottom", "Alt+}", "Align bottom sides of selected objects", NULL,
     CMDID_ALIGNOBJSBOTTOM },
   { "AlignObjsLeft", "Alt+[", "Align left sides of selected objects", NULL,
     CMDID_ALIGNOBJSLEFT },
   { "AlignObjsCenter", "Alt+=", "Align horizontal centers of selected objects",
     NULL, CMDID_ALIGNOBJSCENTER },
   { "AlignObjsRight", "Alt+]", "Align right sides of selected objects", NULL,
     CMDID_ALIGNOBJSRIGHT },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo alignMenuInfo={ TGMUTYPE_TEXT, alignMenuItemInfo,
                                  TgCreateMenuFromMenuInfo };

static TgMenuItemInfo sizeMenuItemInfo[] = {
   { "SizeToWidest", NULL,
     "Resize all selected objects to be as wide as the widest one", NULL,
     CMDID_SIZETOWIDEST },
   { "SizeToNarrowest", NULL,
     "Resize all selected objects to be as narrow as the narrowest one", NULL,
     CMDID_SIZETONARROWEST },
   { "SizeToTallest", NULL,
     "Resize all selected objects to be as tall as the tallest one", NULL,
     CMDID_SIZETOTALLEST },
   { "SizeToShortest", NULL,
     "Resize all selected objects to be as short as the shortest one", NULL,
     CMDID_SIZETOSHORTEST },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SizeToGivenWidthHeight...", NULL,
     "Resize all selected objects to a specified width and height", NULL,
     CMDID_SIZETOGIVENWIDTHHEIGHT },
   { "SizeToGivenWidth...", NULL,
     "Resize all selected objects to a specified width", NULL,
     CMDID_SIZETOGIVENWIDTH },
   { "SizeToGivenHeight...", NULL,
     "Resize all selected objects to a specified height", NULL,
     CMDID_SIZETOGIVENHEIGHT },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo sizeMenuInfo={ TGMUTYPE_TEXT, sizeMenuItemInfo,
                                 TgCreateMenuFromMenuInfo };

static TgMenuInfo alignObjsDirectMenuInfo={ TGMUTYPE_BITMAP, NULL,
                                            CreateAlignObjsDirectMenu };
static TgMenuInfo alignToGridDirectMenuInfo={ TGMUTYPE_BITMAP, NULL,
                                              CreateAlignToGridDirectMenu };
static TgMenuInfo alignToPageDirectMenuInfo={ TGMUTYPE_BITMAP, NULL,
                                              CreateAlignToPageDirectMenu };

static TgMenuInfo distrObjsDirectMenuInfo={ TGMUTYPE_BITMAP, NULL,
                                            CreateDistributeDirectMenu };

static TgMenuItemInfo arrangeMenuItemInfo[] = {
   { "Front", "Cntrl+f", "Bring selected objects to the front", NULL,
     CMDID_FRONT },
   { "Back", "Cntrl+b", "Send selected objects to the back", NULL, CMDID_BACK },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Group", "Cntrl+g", "Group selected objects into a compound object", NULL,
     CMDID_GROUP },
   { "UnGroup", "Cntrl+u",
     "Break up selected grouped objects into its components", NULL,
     CMDID_UNGROUP },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Lock", "Alt+<", "Lock the positions of selected objects", NULL,
     CMDID_LOCK },
   { "UnLock", "Alt+>", "UnLock the positions of selected objects", NULL,
     CMDID_UNLOCK },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "HoriAlign", TGMUITEM_PINNABLESUBMENU,
     "HoriAlign submenu >>>", &horiAlignMenuInfo, MENU_HORIALIGN },
   { "VertAlign", TGMUITEM_PINNABLESUBMENU,
     "VertAlign submenu >>>", &vertAlignMenuInfo,
     MENU_VERTALIGN },
   { "AlignObjs", "Cntrl+l",
     "Align selected objects with each other using the current alignments",
     NULL, CMDID_ALIGNOBJS },
   { "AlignToGrid", "Cntrl+t",
     "Align selected objects to grid points using the current alignments", NULL,
     CMDID_ALIGNTOGRID },
   { "AlignToPage", "Alt+&",
     "Align selected objects to the page using the current alignments", NULL,
     CMDID_ALIGNTOPAGE },
   { "AlignObjsDirect", TGMUITEM_SUBMENU,
     "Align objects direct submenu >>>", &alignObjsDirectMenuInfo, INVALID },
   { "AlignToGridDirect", TGMUITEM_SUBMENU,
     "Align to grid direct submenu >>>", &alignToGridDirectMenuInfo, INVALID },
   { "AlignToPageDirect", TGMUITEM_SUBMENU,
     "Align to page direct submenu >>>", &alignToPageDirectMenuInfo, INVALID },
   { "More Object Alignments", TGMUITEM_SUBMENU,
     "Align objects submenu >>>", &alignMenuInfo, INVALID },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SizeObjects", TGMUITEM_SUBMENU,
     "Size objects submenu >>>", &sizeMenuInfo, INVALID },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "DistributeObjs", "Alt+l",
     "Distribute selected objects using the current alignments", NULL,
     CMDID_DISTRIBUTEOBJS },
   { "DistributeObjsDirect", TGMUITEM_SUBMENU,
     "Distribute objects direct submenu >>>", &distrObjsDirectMenuInfo,
     INVALID },
   { "LayoutOnArc", NULL, "Layout objects on an arc", NULL, CMDID_LAYOUTONARC },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "FlipHorizontal", "Alt+h", "Flip selected objects horizontally", NULL,
     CMDID_FLIPHORIZONTAL },
   { "FlipVertical", "Alt+v", "Flip selected objects vertically", NULL,
     CMDID_FLIPVERTICAL },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "RotateClockWise", "Alt+w", "Rotate selected objects clockwise", NULL,
     CMDID_ROTATECLOCKWISE },
   { "RotateCounter", "Alt+c", "Rotate selected objects counter-clockwise",
     NULL, CMDID_ROTATECOUNTER },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SetTextRotation...", NULL, "Set text rotation in degrees", NULL,
     CMDID_SETTEXTROTATION },
   { "SetRotationIncrement...", NULL, "Set rotation increment in degrees", NULL,
     CMDID_SETROTATIONINCREMENT },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "AbutHorizontal", "Alt+_", "Squish selected objects together horizontally",
     NULL, CMDID_ABUTHORIZONTAL },
   { "AbutVertical", "Alt+|", "Squish selected objects together vertically",
     NULL, CMDID_ABUTVERTICAL },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "CenterAnEndPoint", NULL,
     "Move an endpoint of a selected poly to the center of another object",
     NULL, CMDID_CENTERANENDPOINT },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo arrangeMenuInfo={ TGMUTYPE_TEXT, arrangeMenuItemInfo,
                             TgCreateMenuFromMenuInfo };

static TgMenuItemInfo scrollModeMenuItemInfo[] = {
   { "Off", NULL, "Do not update canvas window while scrolling", NULL,
     CMDID_CHANGESCROLLMODE },
   { "Jump", NULL, "Jump update canvas window while scrolling", NULL,
     CMDID_CHANGESCROLLMODE },
   { "Smooth", NULL, "Update canvas window smoothly while scrolling", NULL,
     CMDID_CHANGESCROLLMODE },
};
TgMenuInfo scrollModeMenuInfo={ TGMUTYPE_TEXT, scrollModeMenuItemInfo,
                                CreateScrollModeMenu };

static TgMenuItemInfo showHideMenuItemInfo[] = {
   { "ShowBit/Pixmap", NULL,
     "Toggle between hiding/showing detailed bitmap/pixmap objects", NULL,
     CMDID_TOGGLEMAPSHOWN },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ShowMeasurement", NULL,
     "Show cursor measurements", NULL, CMDID_TOGGLESHOWMEASUREMENT },
   { "SetMeasureUnit...", NULL, "Set the unit of measurements", NULL,
     CMDID_SETMEASUREUNIT },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ShowMenubar", NULL, "Show menubar window", NULL,
     CMDID_TOGGLESHOWMENUBAR },
   { "ShowStatus", NULL, "Show status window", NULL, CMDID_TOGGLESHOWSTATUS },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo showHideMenuInfo={ TGMUTYPE_TEXT, showHideMenuItemInfo,
                                     CreateShowHideMenu };

static TgMenuItemInfo slideShowMenuItemInfo[] = {
   { "SlideShow", "Alt+Cntrl+8", "Toggle SlideShow mode", NULL,
     CMDID_TOGGLESLIDESHOW },
   { "SetSlideShowBorderColor...", NULL,
     "Set the color of the slideshow border", NULL,
     CMDID_SETSLIDESHOWBRDRCOLOR },
   { "SetSlideShowWindowOffsets...", NULL,
     "Set the offsets of the slideshow window", NULL,
     CMDID_SETSLIDESHOWWINOFFSETS },
   { "VisibleGridInSlideShow", NULL,
     "Toggle visible grid in slideshow mode", NULL,
     CMDID_VISIBLEGRIDINSLIDESHOW },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo slideShowMenuInfo={ TGMUTYPE_TEXT, slideShowMenuItemInfo,
                                      CreateSlideShowMenu };

static TgMenuItemInfo layoutMenuItemInfo[] = {
   { "+Grid", "Alt+i", "Increment grid size", NULL, CMDID_INCGRID },
   { "-Grid", "Alt+d", "Decrement grid size", NULL, CMDID_DECGRID },
   { "VisibleGrid", "Alt+g", "Hide/show grid lines", NULL, CMDID_TOGGLEGRID },
   { "SnapToGrid", "Alt+Cntrl+g", "Toggle the snapping to grid points effect",
     NULL, CMDID_TOGGLESNAP },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SaveOrigin", NULL, "Save current location as the 'origin' of the drawing",
     NULL, CMDID_SAVEORIGIN },
   { "ScrollToOrig", "Alt+^", "Scroll to the 'origin' of the drawing", NULL,
     CMDID_SCROLLTOORIG },
   { "ScrollMode", TGMUITEM_SUBMENU,
     "ScrollMode submenu >>>", &scrollModeMenuInfo, INVALID },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ZoomIn", "Alt+z", "Zoom in", NULL, CMDID_ZOOMIN },
   { "ZoomOut", "Alt+o", "Zoom out", NULL, CMDID_ZOOMOUT },
   { "DefaultZoom", "Alt+:", "No zoom", NULL, CMDID_DEFAULTZOOM },
   { "ZoomWayOut", "Alt+`", "Zoom way out to see the whole drawing", NULL,
     CMDID_ZOOMWAYOUT },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Portrait", NULL, "Portrait page style", NULL, CMDID_PORTRAITMODE },
   { "Landscape", NULL, "Lanscape page style", NULL,
     CMDID_LANDSCAPEMODE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "StackedPage", TGMUITEM_PINNABLESUBMENU,
     "StackedPage submenu >>>", &stackedPageMenuInfo, MENU_STACKEDPAGE },
   { "TiledPage", TGMUITEM_PINNABLESUBMENU,
     "TiledPage submenu >>>", &tiledPageMenuInfo, MENU_TILEDPAGE },
   { "PageLayout", TGMUITEM_PINNABLESUBMENU,
     "PageLayout submenu >>>", &pageLayoutMenuInfo, MENU_PAGELAYOUT },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Print/ExportFormat", TGMUITEM_SUBMENU,
     "Select print/export format submenu >>>", &exportFormatMenuInfo, INVALID },
   { "Print/ExportInColor", "Alt+Cntrl+k",
     "Toggle between color and black & white printing modes", NULL,
     CMDID_TOGGLEBWCOLORPS },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SetReduction...", "Alt+%",
     "Specify reduction/magnification of the whole drawing", NULL,
     CMDID_SETREDUCTION },
   { "MetricGrid", "Alt+Cntrl+-",
     "Toggle between English and Metric grid systems", NULL,
     CMDID_TOGGLEGRIDSYSTEM },
   { "OneMotionSelMove", NULL,
     "Toggle between click-select-move and click-select-click-move modes", NULL,
     CMDID_TOGGLEONEMOTIONSELMOVE },
   { "UseGrayScale", NULL,
     "Toggle between using gray scales to tile patterns to speed up printing",
     NULL, CMDID_TOGGLEUSEGRAYSCALE },
   { "UseColorLayers", NULL, "Toggle color layers", NULL,
     CMDID_TOGGLECOLORLAYERS },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Show/Hide", TGMUITEM_SUBMENU,
     "Show/Hide submenu >>>", &showHideMenuInfo, INVALID },
   { "SlideShow", TGMUITEM_SUBMENU,
     "SlideShow submenu >>>", &slideShowMenuInfo, INVALID },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo layoutMenuInfo={ TGMUTYPE_TEXT, layoutMenuItemInfo,
                            CreateLayoutMenu };

static TgMenuItemInfo propertiesMenuItemInfo[] = {
   { "CopyProperties...", NULL,
     "Copy properties (e.g., color, fill, etc.) into the cut buffer", NULL,
     CMDID_COPYPROPERTIES },
   { "PasteProperties...", NULL,
     "Paste properties (e.g., color, fill, etc.) from the cut buffer", NULL,
     CMDID_PASTEPROPERTIES },
   { "SaveProperties...", NULL,
     "Save and name a property (e.g., color, fill, etc.) set", NULL,
     CMDID_SAVEPROPERTIES },
   { "RestoreProperties...", NULL,
     "Restore a named property (e.g., color, fill, etc.) set", NULL,
     CMDID_RESTOREPROPERTIES },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "UpdateObjects", "Alt+0",
     "Update selected objects to have current drawing properties", NULL,
     CMDID_UPDATE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SetSelLineWidth...", NULL,
     "Change the line width (and arrow width/height) of selected objects", NULL,
     CMDID_SETSELLINEWIDTH },
   { "SetSelFontSize...", NULL,
     "Change the font size of selected objects", NULL, CMDID_SETSELFONTSIZE },
   { "SetEditTextSize...", NULL,
     "Set the text size to be used in editing existing text objects", NULL,
     CMDID_SETEDITTEXTSIZE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "AddColor...", NULL, "Add colors to the current palette", NULL,
     CMDID_ADDCOLOR },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Font", TGMUITEM_PINNABLESUBMENU,
     "Font submenu >>>", &fontMenuInfo, MENU_FONT },
   { "TextStyle", TGMUITEM_PINNABLESUBMENU,
     "TextStyle submenu >>>", &fontStyleMenuInfo, MENU_STYLE },
   { "TextSize", TGMUITEM_PINNABLESUBMENU,
     "TextSize submenu >>>", &fontSizeMenuInfo, MENU_SIZE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "LineDash", TGMUITEM_PINNABLESUBMENU,
     "LineDash submenu >>>", &lineDashMenuInfo, MENU_LINEDASH },
   { "LineStyle", TGMUITEM_PINNABLESUBMENU,
     "LineStyle submenu >>>", &lineStyleMenuInfo, MENU_LINESTYLE },
   { "LineType", TGMUITEM_PINNABLESUBMENU,
     "LineType submenu >>>", &lineTypeMenuInfo, MENU_LINETYPE },
   { "LineWidth", TGMUITEM_PINNABLESUBMENU,
     "LineWidth submenu >>>", &lineWidthMenuInfo, MENU_LINEWIDTH },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Fill", TGMUITEM_PINNABLESUBMENU,
     "Fill submenu >>>", &fillMenuInfo, MENU_FILL },
   { "Pen", TGMUITEM_PINNABLESUBMENU,
     "Pen submenu >>>", &penMenuInfo, MENU_PEN },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "MoveMode", TGMUITEM_PINNABLESUBMENU,
     "MoveMode submenu >>>", &moveModeMenuInfo, MENU_MOVEMODE },
   { "StretchText", TGMUITEM_PINNABLESUBMENU,
     "StretchText submenu >>>", &stretchableTextModeMenuInfo,
     MENU_STRETCHTEXT },
   { "TransPattern", TGMUITEM_PINNABLESUBMENU,
     "TransPattern submenu >>>", &transPatModeMenuInfo, MENU_TRANSPAT },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo propertiesMenuInfo={ TGMUTYPE_TEXT, propertiesMenuItemInfo,
                                TgCreateMenuFromMenuInfo };

static TgMenuItemInfo navigateMenuItemInfo[] = {
   { "GoBack", NULL, "Go back one file", NULL, CMDID_GOBACK },
   { "GoForward", NULL, "Go forward one file", NULL, CMDID_GOFORWARD },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "RefreshCurrent", NULL, "Reload the current file", NULL,
     CMDID_REFRESHCURRENT },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "HotList...", NULL, "Navigate using the hot-list", NULL, CMDID_HOTLIST },
   { "AddCurrentToHotList", NULL, "Add the current file to the hot-list", NULL,
     CMDID_ADDCURRENTTOHOTLIST },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "SessionHistory...", NULL, "Go to a file visited during this session",
     NULL, CMDID_SESSIONHISTORY },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "HyperSpace", NULL, "Toggle HyperSpace mode", NULL,
     CMDID_TOGGLEHYPERSPACE },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo navigateMenuInfo={ TGMUTYPE_TEXT, navigateMenuItemInfo,
                              CreateNavigateMenu };

static TgMenuItemInfo attrMenuItemInfo[] = {
   { "AttachAttrs", "Alt+a",
     "Attach selected text objects as attributes for the non-text object", NULL,
     CMDID_ATTACHATTRS },
   { "DetachAttrs", "Alt+t",
     "Detach all attributes of selected objects", NULL,
     CMDID_DETACHATTRS },
   { "AttachFileAttrs", NULL,
     "Attach selected text objects as file attributes", NULL,
     CMDID_ATTACHFILEATTRS },
   { "DetachFileAttrs", NULL,
     "Detach all file attributes", NULL,
     CMDID_DETACHFILEATTRS },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "EditAttrs...", NULL,
     "Edit attributes of a selected object", NULL,
     CMDID_EDITATTRS },
   { "EditFileAttrs...", NULL,
     "Edit file attributes", NULL,
     CMDID_EDITFILEATTRS },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ShowAttr", "Alt+-",
     "Make all attributes of selected objects visible", NULL,
     CMDID_SHOWATTR },
   { "ShowAttrName", "Alt+n",
     "Make all attribute names of selected objects visible", NULL,
     CMDID_SHOWATTRNAME },
   { "HideAttr", "Alt+Cntrl+h",
     "Hide all attributes for selected objects", NULL,
     CMDID_HIDEATTR },
   { "HideAttrName", "Alt+j",
     "Hide all attribute names for selected objects", NULL,
     CMDID_HIDEATTRNAME },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Move/JustfyAttr", "Alt+m",
     "Move/justify an attribute for a selected object", NULL,
     CMDID_MOVEJUSTIFYATTR },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ImportAttrs...", NULL,
     "Import attributes of a selected object from a text file", NULL,
     CMDID_IMPORTATTRS },
   { "ExportAttrs...", NULL,
     "Export attributes of a selected object to a text file", NULL,
     CMDID_EXPORTATTRS },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo attrMenuInfo={ TGMUTYPE_TEXT, attrMenuItemInfo,
                                 TgCreateMenuFromMenuInfo };

static TgMenuItemInfo animateMenuItemInfo[] = {
   { "AnimateSend", "Alt+e",
     "Animate a little token on a selected poly (not very useful)", NULL,
     CMDID_ANIMATESEND },
   { "AnimateFlash", "Alt+f",
     "Flash a selected poly (not very useful)", NULL,
     CMDID_ANIMATEFLASH },
   { NULL, NULL, NULL, NULL, INVALID }
};
static TgMenuInfo animateMenuInfo={ TGMUTYPE_TEXT, animateMenuItemInfo,
                                    TgCreateMenuFromMenuInfo };

static TgMenuItemInfo specialMenuItemInfo[] = {
   { "MakeSymbolic", "Alt+Cntrl+m",
     "Turn a selected grouped/icon object into a symbol object", NULL,
     CMDID_MAKESYMBOLIC },
   { "UnMakeSymbolic", "Alt+Cntrl+n",
     "Turn a selected symbol object into a grouped object", NULL,
     CMDID_UNMAKESYMBOLIC },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Instantiate...", "Cntrl+i",
     "Instantiate a building-block object from the current domain", NULL,
     CMDID_INSTANTIATE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "MakeIconic...", "Alt+Cntrl+i",
     "Turn a selected grouped/symbol object into an icon object", NULL,
     CMDID_MAKEICONIC },
   { "UnMakeIconic", "Alt+Cntrl+j",
     "Turn an icon object into a grouped object", NULL,
     CMDID_UNMAKEICONIC },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Push", "Cntrl+v",
     "Push into (edit) the symbol file which defines the selected icon object",
     NULL, CMDID_PUSH },
   { "Pop", "Cntrl+k",
     "Pop back to a high lever (reverse of Push)", NULL,
     CMDID_POP },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Attribute", TGMUITEM_SUBMENU,
     "Attribute submenu >>>", &attrMenuInfo, INVALID },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "UpdateSymbols", "Alt+Cntrl+u",
     "Refresh selected icon objects from their definition files", NULL,
     CMDID_UPDATESYMBOLS },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Misc", TGMUITEM_SUBMENU,
     "Misc submenu >>>", &animateMenuInfo, INVALID },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "MergeWithTable...", NULL,
     "Merge selected object with a table file", NULL,
     CMDID_MERGEWITHTABLE },
   { "ExportToTable...", NULL,
     "Export attributes of selected objects to a table file", NULL,
     CMDID_EXPORTTOTABLE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ExecCmdsFromFile...", NULL, "Execute internal commands from a file", NULL,
     CMDID_EXECCMDSFROMFILE },
#ifdef _TGIF_WB
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "WhiteBoard", NULL,
     "Start/Enable the White Board communication", NULL, CMDID_WHITEBOARD },
#endif /* _TGIF_WB */
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo specialMenuInfo={ TGMUTYPE_TEXT, specialMenuItemInfo,
                             TgCreateMenuFromMenuInfo };

static TgMenuItemInfo helpMenuItemInfo[] = {
   { "LatestReleaseInfo", NULL,
     "Find out what is the version number of the most recently released tgif",
     NULL, CMDID_LATESTRELEASEINFO },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Copyright", NULL, "Copyright information", NULL, CMDID_COPYRIGHT },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "About", NULL,
     "Version information, where to send bug reports, etc.", NULL,
     CMDID_ABOUT },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo helpMenuInfo={ TGMUTYPE_TEXT, helpMenuItemInfo,
                          TgCreateMenuFromMenuInfo };

/* ------------------ main menu ------------------ */

static TgMenuItemInfo mainMenuItemInfo[] = {
   { "File", TGMUITEM_PINNABLESUBMENU, "File Menu", &fileMenuInfo, MENU_FILE },
   { "Edit", TGMUITEM_PINNABLESUBMENU, "Edit Menu", &editMenuInfo, MENU_EDIT },
   { "Layout", TGMUITEM_PINNABLESUBMENU,
     "Layout Menu", &layoutMenuInfo, MENU_LAYOUT },
   { "Arrange", TGMUITEM_PINNABLESUBMENU,
     "Arrange Menu", &arrangeMenuInfo, MENU_ARRANGE },
   { "Properties", TGMUITEM_PINNABLESUBMENU,
     "Properties Menu", &propertiesMenuInfo, MENU_PROPERTIES },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "StackedPage", TGMUITEM_PINNABLESUBMENU,
     "StackedPage Menu", &stackedPageMenuInfo, MENU_STACKEDPAGE },
   { "TiledPage", TGMUITEM_PINNABLESUBMENU,
     "TiledPage Menu", &tiledPageMenuInfo, MENU_TILEDPAGE },
   { "PageLayout", TGMUITEM_PINNABLESUBMENU,
     "PageLayout Menu", &pageLayoutMenuInfo, MENU_PAGELAYOUT },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "HoriAlign", TGMUITEM_PINNABLESUBMENU,
     "HoriAlign Menu", &horiAlignMenuInfo, MENU_HORIALIGN },
   { "VertAlign", TGMUITEM_PINNABLESUBMENU,
     "VertAlign Menu", &vertAlignMenuInfo, MENU_VERTALIGN },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Font", TGMUITEM_PINNABLESUBMENU, "Font Menu", &fontMenuInfo, MENU_FONT },
   { "TextStyle", TGMUITEM_PINNABLESUBMENU,
     "TextStyle Menu", &fontStyleMenuInfo, MENU_STYLE },
   { "TextSize", TGMUITEM_PINNABLESUBMENU,
     "TextSize Menu", &fontSizeMenuInfo, MENU_SIZE },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "LineDash", TGMUITEM_PINNABLESUBMENU,
     "LineDash Menu", &lineDashMenuInfo, MENU_LINEDASH },
   { "LineStyle", TGMUITEM_PINNABLESUBMENU,
     "LineStyle Menu", &lineStyleMenuInfo, MENU_LINESTYLE },
   { "LineType", TGMUITEM_PINNABLESUBMENU,
     "LineType Menu", &lineTypeMenuInfo, MENU_LINETYPE },
   { "LineWidth", TGMUITEM_PINNABLESUBMENU,
     "LineWidth Menu", &lineWidthMenuInfo, MENU_LINEWIDTH },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Fill", TGMUITEM_PINNABLESUBMENU, "Fill Menu", &fillMenuInfo, MENU_FILL },
   { "Pen", TGMUITEM_PINNABLESUBMENU, "Pen Menu", &penMenuInfo, MENU_PEN },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "MoveMode", TGMUITEM_PINNABLESUBMENU,
     "MoveMode Menu", &moveModeMenuInfo, MENU_MOVEMODE },
   { "Shape", TGMUITEM_PINNABLESUBMENU,
     "Shape Menu", &shapeMenuInfo, MENU_SHAPE },
   { "StretchText", TGMUITEM_PINNABLESUBMENU,
     "StretchText Menu", &stretchableTextModeMenuInfo, MENU_STRETCHTEXT },
   { "TransPattern", TGMUITEM_PINNABLESUBMENU,
     "TransPattern Menu", &transPatModeMenuInfo,
         MENU_TRANSPAT },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "ImageProc", TGMUITEM_PINNABLESUBMENU,
     "ImageProc Menu", &imageProcMenuInfo, MENU_IMAGEPROC },
   { TGMUITEM_SEPARATOR, NULL, NULL, NULL, INVALID },
   { "Color", TGMUITEM_PINNABLESUBMENU,
     "Color Menu", &colorMenuInfo, MENU_COLOR },
   { "Navigate", TGMUITEM_PINNABLESUBMENU,
     "Navigate Menu", &navigateMenuInfo, MENU_NAVIGATE },
   { "Special", TGMUITEM_PINNABLESUBMENU,
     "Special Menu", &specialMenuInfo, MENU_SPECIAL },
#ifdef _TANGRAM_II
   { "Tangram-II", TGMUITEM_PINNABLESUBMENU,
     "Tangram-II Menu", &tangram2MenuInfo, MENU_TANGRAM2 },
#endif /* _TANGRAM_II */
   { "Help", TGMUITEM_PINNABLESUBMENU, "Help Menu", &helpMenuInfo, MENU_HELP },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo mainMenuInfo={ TGMUTYPE_TEXT, mainMenuItemInfo, CreateMainMenu };

/* ------------------ menubar ------------------ */

/*
 * Don't put any separators here!
 */

static TgMenuItemInfo minimalMenubarMenuItemInfo[] = {
   { "File", TGMUITEM_PINNABLESUBMENU, "File Menu", &fileMenuInfo, MENU_FILE },
   { "Edit", TGMUITEM_PINNABLESUBMENU, "Edit Menu", &editMenuInfo, MENU_EDIT },
   { "Layout", TGMUITEM_PINNABLESUBMENU,
     "Layout Menu", &layoutMenuInfo, MENU_LAYOUT },
   { "Arrange", TGMUITEM_PINNABLESUBMENU,
     "Arrange Menu", &arrangeMenuInfo, MENU_ARRANGE },
   { "Properties", TGMUITEM_PINNABLESUBMENU,
     "Properties Menu", &propertiesMenuInfo, MENU_PROPERTIES },
   { "Color", TGMUITEM_PINNABLESUBMENU,
     "Color Menu", &colorMenuInfo, MENU_COLOR },
   { "Navigate", TGMUITEM_PINNABLESUBMENU,
     "Navigate Menu", &navigateMenuInfo, MENU_NAVIGATE },
   { "Special", TGMUITEM_PINNABLESUBMENU,
     "Special Menu", &specialMenuInfo, MENU_SPECIAL },
#ifdef _TANGRAM_II
   { "Tangram-II", TGMUITEM_PINNABLESUBMENU,
     "Tangram-II Menu", &tangram2MenuInfo, MENU_TANGRAM2 },
#endif /* _TANGRAM_II */
   { "Help", TGMUITEM_PINNABLESUBMENU, "Help Menu", &helpMenuInfo, MENU_HELP },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo minimalMenubarMenuInfo={ TGMUTYPE_TEXT, minimalMenubarMenuItemInfo,
                                    NULL };

static TgMenuItemInfo maximalMenubarMenuItemInfo[] = {
   { "File", TGMUITEM_PINNABLESUBMENU, "File Menu", &fileMenuInfo, MENU_FILE },
   { "Edit", TGMUITEM_PINNABLESUBMENU, "Edit Menu", &editMenuInfo, MENU_EDIT },
   { "Layout", TGMUITEM_PINNABLESUBMENU,
     "Layout Menu", &layoutMenuInfo, MENU_LAYOUT },
   { "Arrange", TGMUITEM_PINNABLESUBMENU,
     "Arrange Menu", &arrangeMenuInfo, MENU_ARRANGE },
   { "Properties", TGMUITEM_PINNABLESUBMENU,
     "Properties Menu", &propertiesMenuInfo, MENU_PROPERTIES },
   { "MoveMode", TGMUITEM_PINNABLESUBMENU,
     "MoveMode Menu", &moveModeMenuInfo, MENU_MOVEMODE },
   { "Page", TGMUITEM_PINNABLESUBMENU,
     "Page Menu", &pageMenuInfo, MENU_PAGE },
   { "PageLayout", TGMUITEM_PINNABLESUBMENU,
     "PageLayout Menu", &pageLayoutMenuInfo, MENU_PAGELAYOUT },
   { "HoriAlign", TGMUITEM_PINNABLESUBMENU,
     "HoriAlign Menu", &horiAlignMenuInfo, MENU_HORIALIGN },
   { "VertAlign", TGMUITEM_PINNABLESUBMENU,
     "VertAlign Menu", &vertAlignMenuInfo, MENU_VERTALIGN },
   { "Font", TGMUITEM_PINNABLESUBMENU, "Font Menu", &fontMenuInfo, MENU_FONT },
   { "TextStyle", TGMUITEM_PINNABLESUBMENU,
     "TextStyle Menu", &fontStyleMenuInfo, MENU_STYLE },
   { "TextSize", TGMUITEM_PINNABLESUBMENU,
     "TextSize Menu", &fontSizeMenuInfo, MENU_SIZE },
   { "Shape", TGMUITEM_PINNABLESUBMENU,
     "Shape Menu", &shapeMenuInfo, MENU_SHAPE },
   { "StretchText", TGMUITEM_PINNABLESUBMENU,
     "StretchText Menu", &stretchableTextModeMenuInfo, MENU_STRETCHTEXT },
   { "LineDash", TGMUITEM_PINNABLESUBMENU,
     "LineDash Menu", &lineDashMenuInfo, MENU_LINEDASH },
   { "LineStyle", TGMUITEM_PINNABLESUBMENU,
     "LineStyle Menu", &lineStyleMenuInfo, MENU_LINESTYLE },
   { "LineType", TGMUITEM_PINNABLESUBMENU,
     "LineType Menu", &lineTypeMenuInfo, MENU_LINETYPE },
   { "LineWidth", TGMUITEM_PINNABLESUBMENU,
     "LineWidth Menu", &lineWidthMenuInfo, MENU_LINEWIDTH },
   { "Fill", TGMUITEM_PINNABLESUBMENU, "Fill Menu", &fillMenuInfo, MENU_FILL },
   { "Pen", TGMUITEM_PINNABLESUBMENU, "Pen Menu", &penMenuInfo, MENU_PEN },
   { "TransPattern", TGMUITEM_PINNABLESUBMENU,
     "TransPattern Menu", &transPatModeMenuInfo, MENU_TRANSPAT },
   { "Color", TGMUITEM_PINNABLESUBMENU,
     "Color Menu", &colorMenuInfo, MENU_COLOR },
   { "ImageProc", TGMUITEM_PINNABLESUBMENU,
     "ImageProc Menu", &imageProcMenuInfo, MENU_IMAGEPROC },
   { "Navigate", TGMUITEM_PINNABLESUBMENU,
     "Navigate Menu", &navigateMenuInfo, MENU_NAVIGATE },
   { "Special", TGMUITEM_PINNABLESUBMENU,
     "Special Menu", &specialMenuInfo, MENU_SPECIAL },
#ifdef _TANGRAM_II
   { "Tangram-II", TGMUITEM_PINNABLESUBMENU,
     "Tangram-II Menu", &tangram2MenuInfo, MENU_TANGRAM2 },
#endif /* _TANGRAM_II */
   { "Help", TGMUITEM_PINNABLESUBMENU, "Help Menu", &helpMenuInfo, MENU_HELP },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo maximalMenubarMenuInfo={ TGMUTYPE_TEXT, maximalMenubarMenuItemInfo,
                                    NULL };

