/*
 * Author:      William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990-1999, William Chia-Wei Cheng.
 *
 * Permission limited to the use, copy, display, distribute without
 * charging for a fee, and produce derivative works of "tgif" and
 * its documentation for not-for-profit purpose is hereby granted by
 * the Author, provided that the above copyright notice appears in
 * all copies made of "tgif" and that both the copyright notice
 * and this permission notice appear in supporting documentation,
 * and that the name of the Author not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission.  The Author makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.  All other rights (including, but not limited to, the
 * right to sell "tgif", the right to sell or distribute derivative
 * works of "tgif", the right to distribute "tgif" for a fee, and
 * the right to include "tgif" or derivative works of "tgif" in a
 * for-sale product) are reserved by the Author.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
 * OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * @(#)$Header: /mm/src/tgif/v4/RCS/wb.c,v 4.4 1999/01/07 15:39:57 william Exp $
 */

#define _INCLUDE_FROM_WB_C_

#include "tgifdefs.h"

#include "auxtext.e"
#include "drawing.e"
#include "cmd.e"
#include "dialog.e"
#include "file.e"
#include "grid.e"
#include "http.e"
#include "mark.e"
#include "msg.e"
#include "obj.e"
#include "remote.e"
#include "select.e"
#include "stk.e"
#include "tcp.e"
#include "util.e"
#include "wb.e"

struct WhiteBoardRec	gstWBInfo;

int	cmdLineNoWhiteBoard=FALSE;
int	cmdLineWhiteBoardListenOnly=FALSE;

/* =================== Remote Related Functions =================== */

static
void CleanUpWBSockets()
{
   if (gstWBInfo.listening) {
      if (gstWBInfo.listen_socket != (-1)) {
         close(gstWBInfo.listen_socket);
         gstWBInfo.listen_socket = (-1);
      }
      if (gstWBInfo.listen_fp != NULL) {
         fclose(gstWBInfo.listen_fp);
         gstWBInfo.listen_fp = NULL;
      }
      if (*gstWBInfo.listen_fname != '\0') {
         unlink(gstWBInfo.listen_fname);
         *gstWBInfo.listen_fname = '\0';
      }
      gstWBInfo.listening = FALSE;
   }
   if (gstWBInfo.send_socket != (-1)) {
      close(gstWBInfo.send_socket);
      gstWBInfo.send_socket = (-1);
   }
   if (gstWBInfo.send_fp != NULL) {
      fclose(gstWBInfo.send_fp);
      gstWBInfo.send_fp = NULL;
   }
   if (*gstWBInfo.send_fname != '\0') {
      unlink(gstWBInfo.send_fname);
      *gstWBInfo.send_fname = '\0';
   }
}

static
void CleanUpWBCmds()
{
   struct CmdRec *cmd_ptr;
   int num_records=0;

   for (cmd_ptr=gstWBInfo.last_cmd; cmd_ptr != NULL; cmd_ptr=cmd_ptr->prev) {
      num_records++;
   }
   if (num_records > 0) {
      struct CmdRec *prev_cmd;
      double inc=(100.0/((double)num_records)), percent_start=0.0;

      ShowInterrupt(1);
      SaveStatusStrings();
      for (cmd_ptr=gstWBInfo.last_cmd; cmd_ptr != NULL; cmd_ptr=prev_cmd,
            percent_start+=inc) {
         prev_cmd = cmd_ptr->prev;
         DeleteARedoRecord(cmd_ptr, percent_start,
               min(((double)percent_start+inc),((double)100.0)));
      }
      RestoreStatusStrings();
      HideInterrupt();
   }
   gstWBInfo.first_cmd = gstWBInfo.last_cmd = gstWBInfo.cur_cmd = NULL;
   gstWBInfo.first_stopped_cmd = NULL;
}

#ifdef _TGIF_WB2

#ifdef _TGIF_DBG_WB2
static
int CreateWBListenSocket()
{
   struct sockaddr_in server;

   gstWBInfo.listening = TRUE;

   sprintf(gstWBInfo.listen_fname, "%sTgifXXXXXX", TMP_DIR);
   mktemp(gstWBInfo.listen_fname);
   unlink(gstWBInfo.listen_fname);
   if ((gstWBInfo.listen_fp=fopen(gstWBInfo.listen_fname, "w+")) == NULL) {
      fprintf(stderr, "Fail to create '%s' for deserialization.\n",
            gstWBInfo.listen_fname);
      return FALSE;
   }
   gstWBInfo.listen_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   if (gstWBInfo.listen_socket == (-1)) {
      fprintf(stderr, "Fail to create a listening socket.\n");
      return FALSE;
   }
   SetSocketBlockingState(&gstWBInfo.listen_socket, FALSE);

   memset(&server, 0, sizeof(struct sockaddr_in));
   server.sin_family = AF_INET;
   server.sin_port = htons(gstWBInfo.port);
   server.sin_addr.s_addr = htonl(INADDR_ANY);

   if (bind(gstWBInfo.listen_socket, (struct sockaddr *)(&server),
         sizeof(server)) == (-1)) {
      fprintf(stderr, "Fail to bind a listening socket.\n");
      return FALSE;
   }
   listen(gstWBInfo.listen_socket, 1);
   return TRUE;
}
#endif /* _TGIF_DBG_WB2 */

#ifdef _TGIF_DBG_WB2
int AcceptWBSocket()
{
   int csd=(-1);

   for (;;) {
      struct sockaddr_in client;
      int clen=sizeof(struct sockaddr_in);

      if ((csd=accept(gstWBInfo.listen_socket, (struct sockaddr *)(&client),
            &clen)) == (-1)) {
         if (errno == EWOULDBLOCK) {
            break;
         } else if (errno == EINTR) {
#ifdef BSD
#ifndef NEXT
            int status;
#else /* NEXT */
            union wait status;
#endif /* ~NEXT */
            pid_t pid;

            while ((pid=wait3(&status, WNOHANG, NULL)) > 0) ;
#endif /* BSD */
            continue;
         }
         fprintf(stderr,
               "Socket error in AcceptWBSocket(): accept() failed.\n");
         continue;
      }
      break;
   }
   return csd;
}
#endif /* _TGIF_DBG_WB2 */

int SendWBData(pszWBData)
   char *pszWBData;
{
   int status=TG_REMOTE_STATUS_OK, n_socket=0;

   status = HttpDoConnect("localhost", gstWBInfo.port, &n_socket);
   if (status == TG_REMOTE_STATUS_INTR) {
      sprintf(gszMsgBox, "WhiteBoard: connection interrupted.");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      close(n_socket);
   } else if (status < 0) {
      /* n_socket already closed in TcpDoDonnect() */
      /*
       * May be nobody is listening!
       */
   } else if (status == TG_REMOTE_STATUS_OK) {
      /*
       * Serialize and send data.
       */
      status = TcpDoWrite(n_socket, pszWBData, (int)(strlen(pszWBData)+1));
      if (status == TG_REMOTE_STATUS_OK) {
      }
      close(n_socket);
   }
   return TRUE;
}

#endif /* _TGIF_WB2 */

/* =================== Cmd Related Functions =================== */

static
int SaveCmd(FP, CmdType, SubCmdPtr, TopSel, BotSel, NumObjs)
   FILE *FP;
   int CmdType, NumObjs;
   struct SubCmdRec *SubCmdPtr;
   struct SelRec *TopSel, *BotSel;
{
   struct SelRec *sel_ptr=NULL;
   int count=0;

   if (recordCmdIncludeTgifObj) {
      sprintf(gszMsgBox, "%s.\n%s %s.\n\n%s %s.",
            "Warning:  recordCmdIncludeTgifObj is TRUE in SaveCmd()",
            TOOL_NAME, "WhiteBoard may crash",
            "Please try to reproduce this error and",
            "send bug report to william@cs.ucla.edu");
      MsgBox(gszMsgBox, TOOL_NAME, STOP_MB);
      return FALSE;
   } else if (CmdType == CMD_GOTO_PAGE) {
      /*
       * GotoPage() is disabled for _TGIF_WB2.
       */
      return FALSE;
   } else if (CmdType == CMD_MOVE) {
      if (fprintf(FP, "cmd(%1d,%1d,%1d,%1d,%1d).\n",
            CmdType, SubCmdPtr->detail.mv->dx, SubCmdPtr->detail.mv->dy,
            recordCmdIncludeTgifObj, recordCmdUsesNewColormap) == EOF) {
         writeFileFailed = TRUE;
      }
   } else {
      if (fprintf(FP, "cmd(%1d,%1d,%1d).\n",
            CmdType, recordCmdIncludeTgifObj,
            recordCmdUsesNewColormap) == EOF) {
         writeFileFailed = TRUE;
      }
   }
   if (botSelBeforeInCmd != NULL) {
      if (fprintf(FP, "before_image(%1d,[\n", stackingCount) == EOF) {
         writeFileFailed = TRUE;
      }
      for (count=0, sel_ptr=topSelBeforeInCmd; sel_ptr != NULL;
            sel_ptr=sel_ptr->next, count++) {
         struct ObjRec *obj_ptr=sel_ptr->obj;
         char buf[MAXSTRING];

         if (obj_ptr->creator_full_id == NULL) {
            sprintf(buf, "%1d\t%s", obj_ptr->id, gszLocalPID);
         } else {
            strcpy(buf, obj_ptr->creator_full_id);
         }
         if (fprintf(FP, "\t\"") == EOF) writeFileFailed = TRUE;
         SaveString(FP, buf);
         if (sel_ptr->next == NULL) {
            if (fprintf(FP, "\"\n") == EOF) writeFileFailed = TRUE;
         } else {
            if (fprintf(FP, "\",\n") == EOF) writeFileFailed = TRUE;
         }
      }
      if (fprintf(FP, "]).\n") == EOF) writeFileFailed = TRUE;
   }
   if (BotSel != NULL) {
      struct SelRec *sel_ptr;
      int *stacking_pos=NULL, stacking_count=0;

      sel_ptr = PrepareStackingInfo(TopSel, BotSel, NumObjs, &stacking_pos,
            &stacking_count);
      if (sel_ptr == NULL && stacking_count == NumObjs) {
         if (fprintf(FP, "after_positions(%1d,[\n", NumObjs) == EOF) {
            writeFileFailed = TRUE;
         }
         for (count=0, sel_ptr=TopSel;
               count < stacking_count && sel_ptr != NULL;
               count++, sel_ptr=sel_ptr->next) {
            char buf[MAXSTRING];

            sprintf(buf, "%1d", stacking_pos[count]);
            if (fprintf(FP, "\t") == EOF) writeFileFailed = TRUE;
            SaveString(FP, buf);
            if (sel_ptr->next == NULL) {
               if (fprintf(FP, "\n") == EOF) writeFileFailed = TRUE;
            } else {
               if (fprintf(FP, ",\n") == EOF) writeFileFailed = TRUE;
            }
         }
         if (fprintf(FP, "]).\n") == EOF) writeFileFailed = TRUE;

         if (fprintf(FP, "after_image(%1d,[\n", NumObjs) == EOF) {
            writeFileFailed = TRUE;
         }
         for (count=0, sel_ptr=TopSel; sel_ptr != NULL;
               sel_ptr=sel_ptr->next, count++) {
            SaveObj(FP, sel_ptr->obj, 1);
            if (sel_ptr->next == NULL) {
               if (fprintf(FP, "\n") == EOF) writeFileFailed = TRUE;
            } else {
               if (fprintf(FP, ",\n") == EOF) writeFileFailed = TRUE;
            }
         }
         if (fprintf(FP, "]).\n") == EOF) writeFileFailed = TRUE;
      }
      if (stacking_pos != NULL) free(stacking_pos);
   }
   return TRUE;
}

char *SerializeCmd(CmdType, SubCmdPtr, TopSel, BotSel, NumObjs)
   int CmdType, NumObjs;
   struct SubCmdRec *SubCmdPtr;
   struct SelRec *TopSel, *BotSel;
{
   char *buf=NULL, header[MAXSTRING<<1];
   int content_sz=0, ok=TRUE, header_sz=0;
   time_t tloc;

   if (gstWBInfo.send_fp == NULL) {
      sprintf(gszMsgBox, "Cannot open '%s'.\n\nSerialization aborted.",
            gstWBInfo.send_fname);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return NULL;
   }
   rewind(gstWBInfo.send_fp);
   writeFileFailed = FALSE;

   time(&tloc);
   if (fprintf(gstWBInfo.send_fp, "%%TGWB begin - %ld\n", (long)tloc) == EOF) {
      writeFileFailed = TRUE;
   }
   Save(gstWBInfo.send_fp, NULL, 0, 1);
   ok = SaveCmd(gstWBInfo.send_fp, CmdType, SubCmdPtr, TopSel, BotSel, NumObjs);

   if (fprintf(gstWBInfo.send_fp, "%%TGWB end - %ld\n", (long)tloc) == EOF) {
      writeFileFailed = TRUE;
   }
   if (writeFileFailed) {
      sprintf(gszMsgBox, "Fail to write to '%s'.\n\nFile system may be full.",
            gstWBInfo.send_fname);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return NULL;
   }
   if (!ok) {
      return NULL;
   }
   content_sz = (int)ftell(gstWBInfo.send_fp);
   /*
    * From: <PID>\t<IP>\r\n
    * Content-Type: application/x-tgif-cmd\r\n
    * Content-Length: <LENGTH>\r\n
    * \r\n
    * <DATA of size LENGTH>
    */
   sprintf(header, "%s%s\r\n%s%s\r\n%s%1d\r\n\r\n",
         "From: ", gszLocalPID,
         "Content-Type: ", "application/x-tgif-cmd",
         "Content-Length: ", content_sz);
   header_sz = strlen(header);
   buf = (char*)malloc((header_sz+content_sz+1)*sizeof(char));
   if (buf == NULL) FailAllocMessage();
   strcpy(buf, header);
   rewind(gstWBInfo.send_fp);
   if (fread(&buf[header_sz], sizeof(char), content_sz, gstWBInfo.send_fp) !=
         content_sz) {
      sprintf(gszMsgBox, "Error in reading '%s'.\n\nSerialization aborted.",
            gstWBInfo.send_fname);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);

      UtilFree(buf);
      return NULL;
   }
   buf[header_sz+content_sz] = '\0';
#ifdef _TGIF_WB2
#ifdef _TGIF_DBG_WB2
fprintf(stderr, "%s\nSending %1d bytes!\n", buf, header_sz+content_sz+2);
#endif /* _TGIF_DBG_WB2 */
#endif /* _TGIF_WB2 */

   return buf;
}

void SerializeFreeBuf(pszWBData)
   char *pszWBData;
{
   UtilFree(pszWBData);
}

#define GETCMDVALUE(val,name) ScanValue("%d", &(val), name, "cmd")
#define GETBIMGVALUE(val,name) ScanValue("%d", &(val), name, "before_image")
#define GETPOSVALUE(val,name) ScanValue("%d", &(val), name, "after_positions")
#define GETAIMGVALUE(val,name) ScanValue("%d", &(val), name, "after_image")

static struct CmdRec gstDeserializeCmd;

static char **gppszBeforeImage=NULL;
static int gnBeforeImageCount=0;

static int *gpnAfterPositions=NULL;
static int gnAfterPositionsCount=0;

static
void CleanUpDeserializeCmd()
{
   memset(&gstDeserializeCmd, 0, sizeof(gstDeserializeCmd));
   gstDeserializeCmd.type = INVALID;
}

static
void CleanUpBeforeImage()
{
   if (gppszBeforeImage != NULL) {
      int i;

      for (i=0; i < gnBeforeImageCount; i++) {
         if (gppszBeforeImage[i] != NULL) {
            UtilFree(gppszBeforeImage[i]);
         } else {
            break;
         }
      }
      free(gppszBeforeImage);
   }
   gppszBeforeImage = NULL;
   gnBeforeImageCount = 0;
}

static
void CleanUpAfterPositions()
{
   if (gpnAfterPositions != NULL) free(gpnAfterPositions);
   gpnAfterPositions = NULL;
   gnAfterPositionsCount = 0;
}

static
void DeserializationCleanUp()
{
   CleanUpDeserializeCmd();
   CleanUpBeforeImage();
   CleanUpAfterPositions();
}

int ReadCmd(FP, Inbuf)
   FILE *FP;
   char *Inbuf;
{
   char *c_ptr;

   c_ptr = FindChar((int)'(', Inbuf);
   InitScan(c_ptr, "\t\n, ");
   if (GETCMDVALUE(gstDeserializeCmd.type, "cmd_type") == INVALID) {
      return FALSE;
   }
   if (gstDeserializeCmd.type == INVALID) {
      return FALSE;
   } else if (gstDeserializeCmd.type == CMD_GOTO_PAGE) {
      return FALSE;
   } else if (gstDeserializeCmd.type == CMD_MOVE) {
      if (GETCMDVALUE(gstDeserializeCmd.dx, "dx") == INVALID ||
          GETCMDVALUE(gstDeserializeCmd.dy, "dy") == INVALID) {
         return FALSE;
      }
   }
   if (GETCMDVALUE(recordCmdIncludeTgifObj, "include_tgif_obj") == INVALID ||
       GETCMDVALUE(recordCmdUsesNewColormap, "use_new_colormap") == INVALID) {
      return FALSE;
   }
   return TRUE;
}

int ReadBeforeImage(FP, Inbuf)
   FILE *FP;
   char *Inbuf;
{
   char *c_ptr, *line;
   int index=0, ok=TRUE;

   c_ptr = FindChar((int)'(', Inbuf);
   InitScan(c_ptr, "\t\n, ");
   if (GETBIMGVALUE(gnBeforeImageCount, "before_image_count") == INVALID) {
      return FALSE;
   }
   if (gnBeforeImageCount > 0) {
      gppszBeforeImage = (char**)malloc(gnBeforeImageCount*sizeof(char*));
      if (gppszBeforeImage == NULL) FailAllocMessage();
      memset(gppszBeforeImage, 0, gnBeforeImageCount*sizeof(char*));
   }
   index = 0;
   while ((line=UtilGetALine(FP)) != NULL) {
      char full_id[MAXSTRING];

      scanLineNum++;
      if (*line == ']') {
         free(line);
         break;
      }
      if (index >= gnBeforeImageCount) {
         fprintf(stderr, "Warning: too many lines in before_image.\n");
         free(line);
         break;
      }
      c_ptr = FindChar((int)'"', line);
      c_ptr = ParseStr(c_ptr, (int)'"', full_id, sizeof(full_id));
      if (c_ptr == NULL) {
         ok = FALSE;
         free(line);
         break;
      }
      gppszBeforeImage[index] = UtilStrDup(full_id);
      if (gppszBeforeImage[index] == NULL) FailAllocMessage();
      index++;
      free(line);
   }
   if (ok && index < gnBeforeImageCount) {
      fprintf(stderr, "Warning: too few lines in before_image.\n");
      ok = FALSE;
   }
   if (!ok) {
      CleanUpBeforeImage();
   }
   return ok;
}

int ReadAfterPositions(FP, Inbuf)
   FILE *FP;
   char *Inbuf;
{
   char *c_ptr, *line;
   int index=0, ok=TRUE;

   c_ptr = FindChar((int)'(', Inbuf);
   InitScan(c_ptr, "\t\n, ");
   if (GETPOSVALUE(gnAfterPositionsCount, "after_positions_count") == INVALID) {
      return FALSE;
   }
   if (gnAfterPositionsCount > 0) {
      gpnAfterPositions = (int*)malloc(gnAfterPositionsCount*sizeof(int));
      if (gpnAfterPositions == NULL) FailAllocMessage();
      memset(gpnAfterPositions, 0, gnAfterPositionsCount*sizeof(int));
   }
   index = 0;
   while ((line=UtilGetALine(FP)) != NULL) {
      int pos=(-1);

      scanLineNum++;
      if (*line == ']') {
         free(line);
         break;
      }
      if (index >= gnAfterPositionsCount) {
         fprintf(stderr, "Warning: too many lines in before_image.\n");
         free(line);
         break;
      }
      InitScan(line, "\t\n, ");
      if (GETPOSVALUE(pos, "after_positions_pos") == INVALID) {
         ok = FALSE;
         free(line);
         break;
      }
      gpnAfterPositions[index++] = pos;
      free(line);
   }
   if (ok && index < gnAfterPositionsCount) {
      fprintf(stderr, "Warning: too few lines in after_positions.\n");
      ok = FALSE;
   }
   if (!ok) {
      CleanUpBeforeImage();
   }
   return ok;
}

int ReadAfterImage(FP, Inbuf)
   FILE *FP;
   char *Inbuf;
{
   struct ObjRec *obj_ptr=NULL;
   char *c_ptr;
   int ok=TRUE, count_expected=(-1), count=0;

   c_ptr = FindChar((int)'(', Inbuf);
   InitScan(c_ptr, "\t\n, ");
   if (GETAIMGVALUE(count_expected, "after_image_count") == INVALID) {
      return FALSE;
   }
   count = 0;
   while (ReadObj(FP, &obj_ptr)) {
      if (obj_ptr == NULL) {
         ok = FALSE;
         break;
      }
      obj_ptr->next = NULL;
      obj_ptr->prev = botObj;
      if (botObj == NULL) {
         topObj = obj_ptr;
      } else {
         botObj->next = obj_ptr;
      }
      botObj = obj_ptr;
      obj_ptr = NULL;
      count++;
   }
   if (ok && count != count_expected) {
      fprintf(stderr, "Warning: object count mismatched in after_image.\n");
      ok = FALSE;
   }
   return ok;
}

static
void SelectObjects(top_obj, bot_obj)
   struct ObjRec *top_obj, *bot_obj;
{
   struct ObjRec *obj_ptr;

   for (obj_ptr=bot_obj; obj_ptr != NULL; obj_ptr=obj_ptr->prev) {
      struct SelRec *sel_ptr;

      obj_ptr->tmp_parent = NULL;
      sel_ptr = (struct SelRec *)malloc(sizeof(struct SelRec));
      if (sel_ptr == NULL) FailAllocMessage();
      sel_ptr->next = topSel;
      sel_ptr->obj = obj_ptr;
      sel_ptr->prev = NULL;
      if (topSel == NULL) {
         botSel = sel_ptr;
      } else {
         topSel->prev = sel_ptr;
      }
      topSel = sel_ptr;
   }
   UpdSelBBox();
}

static
struct CmdRec *DeserializeCmd(pszBuffer)
   char *pszBuffer;
{
   struct StkRec *stk_ptr=NULL;
   struct ObjRec *obj_ptr=NULL, *saved_top_obj=NULL, *saved_bot_obj=NULL;
   int read_status=TRUE, ok=TRUE;
   struct CmdRec *cmd_ptr=NULL;

   if (gstWBInfo.listen_fp == NULL) {
      sprintf(gszMsgBox, "Cannot open '%s'.\n\nDeserialization aborted.",
            gstWBInfo.listen_fname);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return NULL;
   }
   rewind(gstWBInfo.listen_fp);
   writeFileFailed = FALSE;
   if (fprintf(gstWBInfo.listen_fp, "%s", pszBuffer) == EOF) {
      writeFileFailed = TRUE;
   }
   if (writeFileFailed) {
      sprintf(gszMsgBox, "Fail to write to '%s'.\n\nFile system may be full.",
            gstWBInfo.listen_fname);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return NULL;
   }
   stk_ptr = SaveFileInfo();
   ResetFileInfo();

   recordCmdIncludeTgifObj = FALSE;
   recordCmdUsesNewColormap = FALSE;

   memset(&gstDeserializeCmd, 0, sizeof(gstDeserializeCmd));
   gstDeserializeCmd.type = INVALID;
   gppszBeforeImage = NULL;
   gpnAfterPositions = NULL;
   gnBeforeImageCount = gnAfterPositionsCount = 0;

   rewind(gstWBInfo.listen_fp);
   importingFile = TRUE;

   strcpy(scanFileName, gstWBInfo.listen_fname);
   scanLineNum = 0;
   readingPageNum = 0;
   loadedCurPageNum = 0;
   foundGoodStateObject = FALSE;
   while ((read_status=ReadObj(gstWBInfo.listen_fp, &obj_ptr)) == TRUE) {
      if (obj_ptr != NULL) {
         obj_ptr->tmp_parent = NULL;
         AdjForOldVersion(obj_ptr);
         AddObj(NULL, topObj, obj_ptr);
      }
   }
   importingFile = FALSE;

   if (read_status == INVALID) {
      sprintf(gszMsgBox, "File version too large (=%1d).  %s!",
            fileVersion, "Deserialization aborted");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      ok = FALSE;
   }
   if (!ok) {
      DelAllObj();
   } else {
      saved_top_obj = topObj;
      saved_bot_obj = botObj;
      topObj = botObj = NULL;
   }
   RestoreFileInfo(stk_ptr);
   ResetOnePageSize();
   free(stk_ptr);

   if (ok) {
      cmd_ptr = (struct CmdRec *)malloc(sizeof(struct CmdRec));
      if (cmd_ptr == NULL) FailAllocMessage();
      memset(cmd_ptr, 0, sizeof(struct CmdRec));
      cmd_ptr->serialized = TRUE; /* created from deserialization */
      cmd_ptr->top_before = cmd_ptr->bot_before = NULL;
      cmd_ptr->pos_before = (int*)gppszBeforeImage;
      cmd_ptr->count_before = gnBeforeImageCount;
      cmd_ptr->actual_pos_before = NULL;
      cmd_ptr->type = gstDeserializeCmd.type;
      cmd_ptr->undone = TRUE;
      cmd_ptr->include_tgif_obj = recordCmdIncludeTgifObj;
      cmd_ptr->new_colormap = recordCmdUsesNewColormap;
      cmd_ptr->dx = gstDeserializeCmd.dx;
      cmd_ptr->dy = gstDeserializeCmd.dy;

      if (saved_top_obj != NULL) {
         SelectObjects(saved_top_obj, saved_bot_obj);
         CopySel(topSel, numObjSelected,
               &(cmd_ptr->top_after), &(cmd_ptr->bot_after));
         cmd_ptr->pos_after = gpnAfterPositions;
         cmd_ptr->count_after = gnAfterPositionsCount;
         RemoveAllSel();
      } else {
         cmd_ptr->top_after = cmd_ptr->bot_after = NULL;
         cmd_ptr->pos_after = NULL;
         cmd_ptr->count_after = 0;
      }
      gppszBeforeImage = NULL;
      gnBeforeImageCount = 0;
      gpnAfterPositions = NULL;
      gnAfterPositionsCount = 0;
   }
   DeserializationCleanUp();

   return cmd_ptr;
}

void ProcessRemoteCmd(psz_buf, buf_sz)
   char *psz_buf;
   int buf_sz;
{
   struct SelRec *saved_top_sel=topSel, *saved_bot_sel=botSel;
   int highlighted_before=FALSE;

   if (topSel != NULL) {
      highlighted_before = TRUE;
      HighLightReverse();
      topSel = botSel = NULL;
   }
   gstWBInfo.cur_cmd = DeserializeCmd(psz_buf);

   if (gstWBInfo.cur_cmd != NULL) {
      int beyond_first_stopped_cmd=FALSE;

      /*
       * Should look at the timestamp of this message and decides where
       *     to insert it.  At this time, it's always appended.
       * If the insertion point is after first_stopped_cmd,
       *     beyond_first_stopped_cmd should be set to TRUE.
       */
      InsertCmd(gstWBInfo.last_cmd, NULL, gstWBInfo.cur_cmd,
            &gstWBInfo.first_cmd, &gstWBInfo.last_cmd);
      gstWBInfo.cur_cmd = gstWBInfo.last_cmd;

      if (!beyond_first_stopped_cmd) {
         /*
          * If this command is inserted beyond the command pointed to by
          *     first_stopped_cmd, it should not even be attempted for
          *     execution.
          */
         gstWBInfo.first_redo_after_deserialize = TRUE;
         if (RedoACmd(gstWBInfo.cur_cmd, FALSE)) {
            /*
             * Success!
             */
            gstWBInfo.first_redo_after_deserialize = FALSE;
            if (topSel != NULL) RemoveAllSel();
            UndoACmd(gstWBInfo.cur_cmd, FALSE);
            if (topSel != NULL) RemoveAllSel();
            RedoACmd(gstWBInfo.cur_cmd, FALSE);
         } else {
            /*
             * For some reason, can't do it!
             */
            gstWBInfo.first_redo_after_deserialize = FALSE;
            /*
             * We should use first_stopped_cmd to point to
             *     gstWBInfo.cur_cmd here!
             */
         }
         if (topSel != NULL) RemoveAllSel();
      }
   }
   topSel = saved_top_sel;
   botSel = saved_bot_sel;
   UpdSelBBox();
   if (highlighted_before) HighLightForward();
}

/* =================== Init and CleanUp Functions =================== */

void CleanUpWhiteBoard()
{
   CleanUpWBSockets();
   CleanUpWBCmds();
   memset(&gstWBInfo, 0, sizeof(struct WhiteBoardRec));
}

int InitWhiteBoard()
{
   memset(&gstWBInfo, 0, sizeof(struct WhiteBoardRec));

   gstWBInfo.port = 8009;

#ifdef _TGIF_WB2
   gstWBInfo.do_whiteboard = TRUE;
#endif /* _TGIF_WB2 */
   if (cmdLineNoWhiteBoard) gstWBInfo.do_whiteboard = FALSE;

   *gstWBInfo.send_fname = '\0';
   gstWBInfo.send_fp = NULL;
   gstWBInfo.send_socket = (-1);

   gstWBInfo.listening = FALSE;

   *gstWBInfo.listen_fname = '\0';
   gstWBInfo.listen_fp = NULL;
   gstWBInfo.listen_socket = (-1);

#ifdef _TGIF_WB2
#ifdef _TGIF_DBG_WB2
   if (gstWBInfo.do_whiteboard && cmdLineWhiteBoardListenOnly) {
      if (!CreateWBListenSocket()) {
         CleanUpWBSockets();
         exit(-1);
      }
   }
#endif /* _TGIF_DBG_WB2 */
   if (gstWBInfo.do_whiteboard) {
      sprintf(gstWBInfo.send_fname, "%sTgifXXXXXX", TMP_DIR);
      mktemp(gstWBInfo.send_fname);
      unlink(gstWBInfo.send_fname);
      if ((gstWBInfo.send_fp=fopen(gstWBInfo.send_fname, "w+")) == NULL) {
         fprintf(stderr, "Fail to create '%s' for serialization.\n",
               gstWBInfo.send_fname);
         exit(-1);
      }
   }
#endif /* _TGIF_WB2 */
   return TRUE;
}

